/*
 * Copyright (C) 2008-2009 Federal University of Pernambuco and 
 * University of Central Florida
 *
 * This file is part of AJML
 *
 * AJML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * AJML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AJML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: AspectUtil.java,v 1.0 2009/03/27 16:48:06 henriquerebelo Exp $
 */

package org.aspectjml.util;

import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;
import org.aspectjml.ajmlrac.RacConstants;
import org.aspectjml.ajmlrac.TransUtils;
import org.aspectjml.checker.JmlAbstractVisitor;
import org.aspectjml.checker.JmlBehaviorSpec;
import org.aspectjml.checker.JmlExceptionalBehaviorSpec;
import org.aspectjml.checker.JmlFormalParameter;
import org.aspectjml.checker.JmlMethodDeclaration;
import org.aspectjml.checker.JmlNormalBehaviorSpec;
import org.aspectjml.checker.JmlSpecCase;
import org.aspectjml.checker.JmlTypeDeclaration;
import org.multijava.mjc.CMethod;
import org.multijava.mjc.CSpecializedType;
import org.multijava.mjc.CType;
import org.multijava.mjc.JFormalParameter;
import org.multijava.mjc.JMethodDeclarationType;

import com.thoughtworks.qdox.JavaDocBuilder;
import com.thoughtworks.qdox.model.Annotation;
import com.thoughtworks.qdox.model.JavaClass;
import com.thoughtworks.qdox.model.JavaField;
import com.thoughtworks.qdox.model.JavaMethod;
import com.thoughtworks.qdox.model.JavaParameter;
import com.thoughtworks.qdox.model.Type;

public class AspectUtil {

	private static AspectUtil instance = null;

	private List surrogateTypeFiles = null;
	private List generatedTypeFiles = null;
	private StringBuffer preconditionPointcuts = null;
	private List interfaceStaticModelFields = null;
	private List defaultSignalsClauseTokenRefereces = null;
	private List defaultEnsuresClauseTokenRefereces = null;
	private List defaultRequiresClauseTokenRefereces = null;
	private List preconditionNewMethodsAdvice = null;
	private List postconditionNewMethodsAfterAdvice = null;
	private List postconditionNewMethodsAroundAdvice = null;
	private List preconditionNewMethodsAdviceXCS = null;
	private List postconditionNewMethodsAfterAdviceXCS = null;
	private List postconditionNewMethodsAroundAdviceXCS = null;
	private List quantifierInnerClasses = null;
	private List quantifierUniqueID = null;
	private List crosscutSpecMeths = null;
	private List aspectAdvice = null;
	private List exceptionsInSignalsClausesInCompilationUnit = null;
	private boolean aroundAdvice = false;
	private int uniqueVarGenForQuantifier = 0;
	private boolean currentCompilationUnitInvariants = false;
	private boolean currentCompilationUnitConstraints = false;
	public boolean isCompilationUnitRacable = false;

	// for multiple spec case checking
	public HashMap preconditionSpecCases = null;
	public HashMap nPostconditionSpecCases = null;
	public HashMap publicPreconditionSpecCases = null;
	public HashMap protectedPreconditionSpecCases = null;
	public HashMap defaultPreconditionSpecCases = null;
	public HashMap privatePreconditionSpecCases = null;
	public HashMap publicNPostconditionSpecCases = null;
	public HashMap protectedNPostconditionSpecCases = null;
	public HashMap defaultNPostconditionSpecCases = null;
	public HashMap privateNPostconditionSpecCases = null;

	public HashMap nPostconditionForContextSpecCases = null;
	public HashMap xPostconditionForContextSpecCases = null;
	public HashMap publicNPostconditionForContextSpecCases = null;
	public HashMap publicXPostconditionForContextSpecCases = null;
	public HashMap protectedNPostconditionForContextSpecCases = null;
	public HashMap protectedXPostconditionForContextSpecCases = null;
	public HashMap defaultNPostconditionForContextSpecCases = null;
	public HashMap defaultXPostconditionForContextSpecCases = null;
	public HashMap privateNPostconditionForContextSpecCases = null;
	public HashMap privateXPostconditionForContextSpecCases = null;

	public HashMap nPostTokenReferenceSpecCases = null;
	public HashMap xPostTokenReferenceSpecCases = null;
	public HashMap publicNPostTokenReferenceSpecCases = null;
	public HashMap publicXPostTokenReferenceSpecCases = null;
	public HashMap protectedNPostTokenReferenceSpecCases = null;
	public HashMap protectedXPostTokenReferenceSpecCases = null;
	public HashMap defaultNPostTokenReferenceSpecCases = null;
	public HashMap defaultXPostTokenReferenceSpecCases = null;
	public HashMap privateNPostTokenReferenceSpecCases = null;
	public HashMap privateXPostTokenReferenceSpecCases = null;

	// for crosscutting contract specification
	private JavaDocBuilder qDoxFile = null;
	private List typeAnnotations = null;
	private List partialDecls = null;
	public boolean hasThisRef = false;
	public boolean isStaticPC = false;
	
	// for open class / open implementation / class slice / hiper slice 
	// aspect inter-type declaration / aspect introduction / aop --- [[[hemr]]]
	private List allJavaTypeDecls = null;
	private List allJavaMethDecls = null;
	private List jmlFileFieldDecl = null;
	private List itdFieldDecl = null;
	private List fieldDecl = null;
	private List itdMethDecl = null;
	private List itdClassDecl = null;
	private List supplyMethDecl = null;
	private List adviceDeclJavaFile = null;
	private List adviceDeclForXActivation = null;
	private List declareMethodAnnoForAdvice = null;
	private List declareParentDecl = null;
	private List declareAnnotationDecl = null;
	private List annotedTypeDecls = null;
	private List annotatedMethodDecls = null;
	private List annotatedFieldDecls = null;
	private List declareSoftMethodDecl = null;
	private List declareSoftMethodDeclStmt = null;
	private List processedDRTypes = null;
	
	// for crossref mapping
	public HashMap fileMap = null;
	public HashMap fileMapFullPath = null;
	
	// for better compilation in Java 1.8 and handle potential 
	// not found method while getting a compilation error --- [[[hemr]]]
	public List potentialNotFoundMeths = null;
	
	// for aspect code that should run just after rac (PostRac) --- [[[hemr]]]
	public StringBuffer postRacCode = null;
	
	// for visibility checking and avoiding duplicating error messages in this regard --- [[[hemr]]]
	private List typeSpecsPrivacyViolationErrors = null;

	public void initializeMethodSpecCaseCheckingContent(){
		this.preconditionSpecCases = new HashMap();
		this.nPostconditionSpecCases = new HashMap();
		this.publicPreconditionSpecCases = new HashMap();
		this.protectedPreconditionSpecCases = new HashMap();
		this.defaultPreconditionSpecCases = new HashMap();
		this.privatePreconditionSpecCases = new HashMap();
		this.publicNPostconditionSpecCases = new HashMap();
		this.protectedNPostconditionSpecCases = new HashMap();
		this.defaultNPostconditionSpecCases = new HashMap();
		this.privateNPostconditionSpecCases = new HashMap();
		this.nPostconditionForContextSpecCases = new HashMap();
		this.xPostconditionForContextSpecCases = new HashMap();
		this.publicNPostconditionForContextSpecCases = new HashMap();
		this.publicXPostconditionForContextSpecCases = new HashMap();
		this.protectedNPostconditionForContextSpecCases = new HashMap();
		this.protectedXPostconditionForContextSpecCases = new HashMap();
		this.defaultNPostconditionForContextSpecCases = new HashMap();
		this.defaultXPostconditionForContextSpecCases = new HashMap();
		this.privateNPostconditionForContextSpecCases = new HashMap();
		this.privateXPostconditionForContextSpecCases = new HashMap();
		this.nPostTokenReferenceSpecCases = new HashMap();
		this.xPostTokenReferenceSpecCases = new HashMap();
		this.publicNPostTokenReferenceSpecCases = new HashMap();
		this.publicXPostTokenReferenceSpecCases = new HashMap();
		this.protectedNPostTokenReferenceSpecCases = new HashMap();
		this.protectedXPostTokenReferenceSpecCases = new HashMap();
		this.defaultNPostTokenReferenceSpecCases = new HashMap();
		this.defaultXPostTokenReferenceSpecCases = new HashMap();
		this.privateNPostTokenReferenceSpecCases = new HashMap();
		this.privateXPostTokenReferenceSpecCases = new HashMap();
	}

	private AspectUtil(){
		this.surrogateTypeFiles = new ArrayList();
		this.generatedTypeFiles = new ArrayList();
		this.interfaceStaticModelFields = new ArrayList();
		this.preconditionPointcuts = new StringBuffer();
		this.defaultSignalsClauseTokenRefereces = new ArrayList();
		this.defaultEnsuresClauseTokenRefereces = new ArrayList();
		this.defaultRequiresClauseTokenRefereces = new ArrayList();
		this.preconditionNewMethodsAdvice = new ArrayList();
		this.postconditionNewMethodsAfterAdvice = new ArrayList();
		this.postconditionNewMethodsAroundAdvice = new ArrayList();
		this.preconditionNewMethodsAdviceXCS = new ArrayList();
		this.postconditionNewMethodsAfterAdviceXCS = new ArrayList();
		this.postconditionNewMethodsAroundAdviceXCS = new ArrayList();
		this.quantifierInnerClasses = new ArrayList();
		this.quantifierUniqueID = new ArrayList();
		this.crosscutSpecMeths = new ArrayList();
		this.aspectAdvice = new ArrayList();
		this.allJavaTypeDecls = new ArrayList();
		this.allJavaMethDecls = new ArrayList();
		this.itdFieldDecl = new ArrayList();
		this.jmlFileFieldDecl = new ArrayList();
		this.fieldDecl = new ArrayList();
		this.itdMethDecl = new ArrayList();
		this.itdClassDecl = new ArrayList();
		this.supplyMethDecl = new ArrayList();
		this.adviceDeclJavaFile = new ArrayList();
		this.adviceDeclForXActivation = new ArrayList();
		this.declareMethodAnnoForAdvice = new ArrayList();
		this.declareParentDecl = new ArrayList();
		this.declareAnnotationDecl = new ArrayList();
		this.declareSoftMethodDecl = new ArrayList();
		this.declareSoftMethodDeclStmt = new ArrayList();
		this.annotedTypeDecls = new ArrayList();
		this.annotatedMethodDecls = new ArrayList();
		this.annotatedFieldDecls = new ArrayList();
		this.qDoxFile = new JavaDocBuilder();
		this.typeAnnotations = new ArrayList();
		this.partialDecls = new ArrayList();
		this.processedDRTypes = new ArrayList();
		this.exceptionsInSignalsClausesInCompilationUnit = new ArrayList();
		this.fileMap = new HashMap();
		this.fileMapFullPath = new HashMap();
		this.potentialNotFoundMeths = new ArrayList();
		this.postRacCode = new StringBuffer("");
		this.typeSpecsPrivacyViolationErrors = new ArrayList();
	}

	public static AspectUtil getInstance(){
		if(instance == null){
			instance = new AspectUtil();
		}
		return instance;
	}

	public void emptyInterfaceStaticModelFields() {
		this.interfaceStaticModelFields.clear();
	}

	public void emptyPreconditionNewMethodsAdvice() {
		this.preconditionNewMethodsAdvice.clear();
	}

	public void emptyPostconditionNewMethodsAfterAdvice() {
		this.postconditionNewMethodsAfterAdvice.clear();
	}

	public void emptyPostconditionNewMethodsAroundAdvice() {
		this.postconditionNewMethodsAroundAdvice.clear();
	}
	
	public void emptyPreconditionNewMethodsAdviceXCS() {
		this.preconditionNewMethodsAdviceXCS.clear();
	}

	public void emptyPostconditionNewMethodsAfterAdviceXCS() {
		this.postconditionNewMethodsAfterAdviceXCS.clear();
	}

	public void emptyPostconditionNewMethodsAroundAdviceXCS() {
		this.postconditionNewMethodsAroundAdviceXCS.clear();
	}
	
//	public void emptyCrossrefs() {
//		this.crossrefs.clear();
//	}

	public void emptyQuantifierInnerClasses() {
		this.quantifierInnerClasses.clear();
	}

	public void emptyQuantifierUniqueID(){
		this.quantifierUniqueID.clear();
	}
	
	public void emptyExceptionsInSignalsClausesInCompilationUnit() {
		this.exceptionsInSignalsClausesInCompilationUnit.clear();
	}
	
	public void emptyAllJavaTypeDeclarations(){
		this.allJavaTypeDecls.clear();
	}
	
	public void emptyAllJavaMethDeclarations(){
		this.allJavaMethDecls.clear();
	}
	
	public void emptyITDFieldDeclaration(){
		this.itdFieldDecl.clear();
	}
	
	public void emptyFieldDeclarations(){
		this.fieldDecl.clear();
	}
	
	public void emptyITDMethodDeclaration(){
		this.itdMethDecl.clear();
	}
	
	public void emptyITDClassDeclaration(){
		this.itdClassDecl.clear();
	}
	
	public void emptyAdviceDeclJavaFile(){
		this.adviceDeclJavaFile.clear();
	}
	
	public void emptyDeclareSoftMethJavaFile(){
		this.annotatedMethodDecls.clear();
	}
	
	public void emptyDeclareSoftTypeJavaFile(){
		this.annotedTypeDecls.clear();
	}
	
	public void emptyTypeAnnotations(){
		this.typeAnnotations.clear();
	}
	
	public void emptyAdviceDeclForXActivation(){
		this.adviceDeclForXActivation.clear();
	}
	
	public void emptyDeclareMethodAnnoForAdvice(){
		this.declareMethodAnnoForAdvice.clear();
	}
	
	public void emptyDeclareParentDecl(){
		this.declareParentDecl.clear();
	}
	
	public void emptyDeclareAnnotationDecl(){
		this.declareAnnotationDecl.clear();
	}
	
	public void emptyDeclareSoftMethodDecl(){
		this.declareSoftMethodDecl.clear();
	}
	
	public void emptyDeclareSoftMethodDeclStmt(){
		this.declareSoftMethodDeclStmt.clear();
	}
	
	public void emptyTypeSpecsPrivacyViolationErrors(){
		this.typeSpecsPrivacyViolationErrors.clear();
	}

	public boolean isAroundAdvice(){
		return this.aroundAdvice;
	}

	public int getUniqueVarGenForQuantifier(){
		return this.uniqueVarGenForQuantifier++;
	}

	public boolean hasPreconditionPointcuts(){
		return this.preconditionPointcuts.length() > 0;
	}

	public void setAroundAdvice(boolean b){
		this.aroundAdvice = b;
	}
	
	public List getSurrogateTypeFiles(){
		return this.surrogateTypeFiles;
	}
	
	public List getGeneratedTypeFiles(){
		return this.generatedTypeFiles;
	}

	public List getInterfaceStaticModelFields() {
		return this.interfaceStaticModelFields;
	}

	public List getPreconditionNewMethodsAdvice() {
		return this.preconditionNewMethodsAdvice;
	}

	public List getPostconditionNewMethodsAfterAdvice() {
		return this.postconditionNewMethodsAfterAdvice;
	}

	public List getPostconditionNewMethodsAroundAdvice() {
		return this.postconditionNewMethodsAroundAdvice;
	}
	
	public List getPreconditionNewMethodsAdviceXCS() {
		return this.preconditionNewMethodsAdviceXCS;
	}

	public List getPostconditionNewMethodsAfterAdviceXCS() {
		return this.postconditionNewMethodsAfterAdviceXCS;
	}

	public List getPostconditionNewMethodsAroundAdviceXCS() {
		return this.postconditionNewMethodsAroundAdviceXCS;
	}

//	public List getCrossrefs() {
//		return this.crossrefs;
//	}

	public List getQuantifierUniqueIDs(){
		return this.quantifierUniqueID;
	}

	public List getQuantifierInnerClasses(){
		return this.quantifierInnerClasses;
	}
	
	public List getExceptionsInSignalsClausesInCompilationUnit(){
		return this.exceptionsInSignalsClausesInCompilationUnit;
	}
	
	public boolean isGeneratedTempFile(File file){
		boolean ret = false;
		for (Iterator iterator = this.generatedTypeFiles.iterator(); iterator.hasNext();) {
			File currentTemp = (File) iterator.next();
			if(currentTemp.getName().equals(file.getName())){
				if(currentTemp.getAbsolutePath().endsWith(".java") && file.getAbsolutePath().endsWith(".java")){
					ret = true;
				}
			}
		}
		return ret;
	}

	public String getQuantifierInnerClassByID(String quantifierID){
		String result = "";
		for (Iterator iterator = quantifierInnerClasses.iterator(); iterator.hasNext();) {
			String currentQtf = (String) iterator.next();
			if(currentQtf.contains(quantifierID)){
				result = currentQtf;
				break;
			}
		}
		return result;
	}

	public String getQuantifierInnerClassesForTrans(String pred){
		StringBuffer result = new StringBuffer();
		result.append("");
		if(pred.contains(RacConstants.CN_RAC_QUANTIFIER_ID)){
			List ids = AspectUtil.getInstance().getQuantifierUniqueIDs();
			for (Iterator iterator = ids.iterator(); iterator.hasNext();) {
				String currentID = (String) iterator.next();
				if(pred.contains(currentID)){
					String qcode = AspectUtil.getInstance().getQuantifierInnerClassByID(currentID);
					result.append(qcode+"\n");

				}
			}
		}
		return result.toString();
	}
	
	public void appendSurrogateTypeFiles(File surrogateFile){
		this.surrogateTypeFiles.add(surrogateFile);
	}
	
	public void appendGeneratedTypeFiles(File generatedFile){
		this.generatedTypeFiles.add(generatedFile);
	}

	public void appendInterfaceStaticModelFields(String modelField) {
		this.interfaceStaticModelFields.add(modelField);
	}

	public void appendPreconditionNewMethodsAdvice(JMethodDeclarationType meth){
		this.isCompilationUnitRacable = true;
		this.preconditionNewMethodsAdvice.add(meth);
	}

	public void appendPostconditionNewMethodsAfterAdvice(JMethodDeclarationType meth){
		this.isCompilationUnitRacable = true;
		this.postconditionNewMethodsAfterAdvice.add(meth);
	}

	public void appendPostconditionNewMethodsAroundAdvice(JMethodDeclarationType meth){
		this.isCompilationUnitRacable = true;
		this.postconditionNewMethodsAroundAdvice.add(meth);
	}
	
	public void appendPreconditionNewMethodsAdviceXCS(JMethodDeclarationType meth){
		this.isCompilationUnitRacable = true;
		this.preconditionNewMethodsAdviceXCS.add(meth);
	}

	public void appendPostconditionNewMethodsAfterAdviceXCS(JMethodDeclarationType meth){
		this.isCompilationUnitRacable = true;
		this.postconditionNewMethodsAfterAdviceXCS.add(meth);
	}

	public void appendPostconditionNewMethodsAroundAdviceXCS(JMethodDeclarationType meth){
		this.isCompilationUnitRacable = true;
		this.postconditionNewMethodsAroundAdviceXCS.add(meth);
	}
	
//	public void appendCrossrefPointcut(String crossrefPC){
//		this.crossrefs.add(crossrefPC);
//	}

	public void appendQuantifierInnerClasses(String quantifierInnerClass){
		this.quantifierInnerClasses.add(quantifierInnerClass);
	}

	public void appendQuantifierUniqueID(String quantifierID){
		this.quantifierUniqueID.add(quantifierID);
	}

	public void appendCrosscutSpecMeth(JavaMethod crosscutSpecMeth){
		this.crosscutSpecMeths.add(crosscutSpecMeth);
	}

	public void appendAspectAdvice(JavaMethod adviceOrPC){

	}
	
	public void appendAllJavaTypeDecl(JavaClass javaType){
		this.allJavaTypeDecls.add(javaType);
	}
	
	public void appendAllJavaMethDecl(JavaMethod javaMethod){
		boolean canAppend = true;
		for (Iterator iterator = this.allJavaMethDecls.iterator(); iterator.hasNext();) {
			JavaMethod currentMethod = (JavaMethod) iterator.next();
			if(currentMethod.getSource().getFile().getAbsolutePath().equals(javaMethod.getSource().getFile().getAbsolutePath())){
				if(currentMethod.getLineNumber() == javaMethod.getLineNumber()){
					canAppend = false;
				}
			}
		}
		if(canAppend){
			this.allJavaMethDecls.add(javaMethod);
		}
	}
	
	public void appendJMLFileFieldDecl(JavaField jmlFileFieldDecl){
		this.jmlFileFieldDecl.add(jmlFileFieldDecl);
	}
	
	public void appendItdFieldDecl(JavaField itdFieldDecl){
		this.itdFieldDecl.add(itdFieldDecl);
	}
	
	public void appendFieldDecls(List fieldDecls){
		this.fieldDecl.addAll(fieldDecls);
	}
	
	public void appendItdMethDecl(JavaMethod itdMethDecl){
		this.itdMethDecl.add(itdMethDecl);
	}
	
	public void appendItdClassDecl(JavaClass itdClassDecl){
		this.itdClassDecl.add(itdClassDecl);
	}
	
	public void appendSupplyMethDecl(JavaMethod supplyMethDecl){
		this.supplyMethDecl.add(supplyMethDecl);
	}
	
	public void appendAdviceDeclJavaFile(JavaMethod adviceDeclJavaFile){
		this.adviceDeclJavaFile.add(adviceDeclJavaFile);
	}
	
	public void appendProcessedDRType(String drType){
		this.processedDRTypes.add(drType); 
	}
	
	public void appendAdviceDeclForXActivation(String adviceForXActivation){
		this.adviceDeclForXActivation.add(adviceForXActivation);
	}
	
	public void appendDeclareMethodAnnoForAdvice(String declareMethodAnnoForAdvice){
		this.declareMethodAnnoForAdvice.add(declareMethodAnnoForAdvice);
	}
	
	public void appendDeclareParentDecl(String declareParentDecl){
		this.declareParentDecl.add(declareParentDecl);
	}
	
	public void appendAnnotationDecl(String declareAnnotationDecl){
		this.declareAnnotationDecl.add(declareAnnotationDecl);
	}
	
	public void appendDeclareSoftMethodDecl(JavaMethod declareSoftMethodDecl){
		this.declareSoftMethodDecl.add(declareSoftMethodDecl);
	}
	
	public void appendDeclareSoftMethodDeclStmt(String declareSoftMethodDeclStmt){
		this.declareSoftMethodDeclStmt.add(declareSoftMethodDeclStmt);
	}
	
	public void appendAnnotatedJavaFieldDecl(JavaField annotatedJavaField){
		this.annotatedFieldDecls.add(annotatedJavaField);
	}
	
	public void appendAnnotatedJavaMethodDecl(JavaMethod annotatedJavaMethod){
		this.annotatedMethodDecls.add(annotatedJavaMethod);
	}
	
	public void appendAnnotatedJavaTypeDecl(JavaClass annotedTypeDecl){
		this.annotedTypeDecls.add(annotedTypeDecl);
	}
	
	public void appendExceptionInSignalsClauseInCompilationUnit(CType exception){
		if(!exception.toString().startsWith("java.lang")){
			boolean can = true;
			for (Iterator<CType> iterator = 
					this.exceptionsInSignalsClausesInCompilationUnit.iterator(); iterator
					.hasNext();) {
				CType cType = iterator.next();
				if(cType.toString().equals(exception.toString())){
					can = false;
					break;
				}
			}
			if(can){
				this.exceptionsInSignalsClausesInCompilationUnit.add(exception);
			}
		}
	}
	
	public void appendTypeSpecsPrivacyViolationErrors(String typeSpecPrivacyViolationError){
		this.typeSpecsPrivacyViolationErrors.add(typeSpecPrivacyViolationError);
	}

	public void emptyPreconditionPointcuts(){
		this.preconditionPointcuts = new StringBuffer();
	}

	public void appendPreconditionPointcut(String pc){
		if(this.preconditionPointcuts.length() == 0)
			this.preconditionPointcuts.append(pc);
		else
			this.preconditionPointcuts.append("\n").
			append("         ").append(" || ").append(pc);
	}

	public String getPreconditionPointcut(){
		String pc = this.preconditionPointcuts.toString();
//		pc = pc.replaceAll("&&(\\s)*this\\((\\s)*object\\$rac(\\s)*\\)", "").
//				replaceAll("&&(\\s)*target\\((\\s)*object\\$rac(\\s)*\\)", "").
//				replaceAll("&&(\\s)*args\\([^\\)]+\\)", "");
		return AspectUtil.processMethSig("("+pc+")");
	}

	public void currentCompilationUnitHasInvariants(){
		this.isCompilationUnitRacable = true;
		this.currentCompilationUnitInvariants = true;
	}

	public void resetCurrentCompilationUnitHasInvariants(){
		this.currentCompilationUnitInvariants = false;
	}

	public void currentCompilationUnitHasConstraints(){
		this.isCompilationUnitRacable = true;
		this.currentCompilationUnitConstraints = true;
	}

	public void resetCurrentCompilationUnitHasConstraints(){
		this.currentCompilationUnitConstraints = false;
	}

	public boolean hasInvariantsInCurrentCompilationUnit(){
		return this.currentCompilationUnitInvariants;
	}

	public boolean hasConstraintsInCurrentCompilationUnit(){
		return this.currentCompilationUnitConstraints;
	}

	public void appendDefaultSignalsClauseTokenRefereces(String tokenReference){
		this.defaultSignalsClauseTokenRefereces.add(tokenReference);
	}

	public List getDefaultSignalsClauseTokenRefereces(){
		return this.defaultSignalsClauseTokenRefereces;
	}

	public void appendDefaultEnsuresClauseTokenRefereces(String tokenReference){
		this.defaultEnsuresClauseTokenRefereces.add(tokenReference);
	}

	public List getDefaultEnsuresClauseTokenRefereces(){
		return this.defaultEnsuresClauseTokenRefereces;
	}

	public void appendDefaultRequiresClauseTokenRefereces(String tokenReference){
		this.defaultRequiresClauseTokenRefereces.add(tokenReference);
	}

	public List getDefaultRequiresClauseTokenRefereces(){
		return this.defaultRequiresClauseTokenRefereces;
	}

	public List getCrosscutSpecMeths(){
		return this.crosscutSpecMeths;
	}

	public List getAspectAdviceOrPCs(){
		return this.aspectAdvice;
	}
	
	public List getAllJavaTypeDeclarations(){
		return this.allJavaTypeDecls;
	}
	
	public JavaClass [] getAllJavaTypeDeclarationsAsArray(){
		JavaClass [] result = new JavaClass[this.allJavaTypeDecls.size()];
		int index = 0;
		for (Iterator iterator = this.allJavaTypeDecls.iterator(); iterator.hasNext();) {
			JavaClass currentType = (JavaClass) iterator.next();
			result[index] = currentType;
			index++;
		}
		return result;
	}
	
	public List getAllJavaMethodDeclarations(){
		return this.allJavaMethDecls;
	}
	
	public List getJMLFileFieldDecls(){
		return this.jmlFileFieldDecl;
	}
	
	public List getItdFieldDecls(){
		return this.itdFieldDecl;
	}
	
	public List getFieldDecls(){
		return this.fieldDecl;
	}
	
	public List getItdMethDecls(){
		return this.itdMethDecl;
	}
	
	public List getItdClassDecls(){
		return this.itdClassDecl;
	}
	
	public List getSupplyMethDecls(){
		return this.supplyMethDecl;
	}
	
	public List getAdviceDeclJavaFile(){
		return this.adviceDeclJavaFile;
	}
	
	public List getTypeAnnotations(){
		return this.typeAnnotations;
	}
	
	public List getPartialDecls(){
		return this.partialDecls;
	}
	
	public List getProcessedDRTypes(){
		return this.processedDRTypes;
	}
	
	public List getAdviceDeclForXActivation(){
		return this.adviceDeclForXActivation;
	}
	
	public List getDeclareMethodAnnoForAdvice(){
		return this.declareMethodAnnoForAdvice;
	}
	
	public List getDeclareParentDecls(){
		return this.declareParentDecl;
	}
	
	public List getDeclareAnnotationDecls(){
		return this.declareAnnotationDecl;
	}
	
	public List getDeclareSoftMethodDecls(){
		return this.declareSoftMethodDecl;
	}
	
	public List getDeclareSoftMethodStmtDecls(){
		return this.declareSoftMethodDeclStmt;
	}
	
	public List getAnnotatedFieldDecls(){
		return this.annotatedFieldDecls;
	}
	
	public List getAnnotatedMethodDecls(){
		return this.annotatedMethodDecls;
	}
	
	public List getAnnotatedTypeDecls(){
		return this.annotedTypeDecls;
	}
	
	public List getTypeSpecsPrivacyViolationErrors(){
		return this.typeSpecsPrivacyViolationErrors;
	}
	
	public boolean containsTypeSpecPrivacyViolationError(String typeSpecError) {
		boolean result = false;
		for (Iterator iterator = typeSpecsPrivacyViolationErrors.iterator(); iterator.hasNext();) {
			String currentTypeSpecError = (String) iterator.next();
			if(currentTypeSpecError.equals(typeSpecError)) {
				result = true;
				break;
			}
		}
		
		return result;
	}

	public static String changeThisOrSuperRefToAdviceRef(String pred, JmlTypeDeclaration typeDecl){
		// handling common replacements
//		System.out.println("pred = "+pred);
		pred = pred.replace("this.", "object$rac.");
		
		try {
			pred = pred.replace(typeDecl.getCClass().getJavaName()+".object$rac.", "object$rac.");
			pred = pred.replace(typeDecl.getCClass().ident()+".object$rac.", "object$rac.");
			pred = pred.replace("super.", "("+"("+typeDecl.getCClass().getSuperClass().getJavaName()+")"+"object$rac).");
			// handling inner types
			for (Iterator iterator = typeDecl.inners().iterator(); iterator.hasNext();) {
				JmlTypeDeclaration currentInnerType = (JmlTypeDeclaration) iterator.next();
				pred = pred.replace("super.", "("+"("+currentInnerType.getCClass().getSuperClass().getJavaName()+")"+"object$rac).");
			}
			pred = pred.replace("delegee_"+typeDecl.ident()+"().", "");
			pred = pred.replaceAll("\\bthis\\b", Matcher.quoteReplacement("object$rac"));
		} catch (Exception e) {}
		return pred;
	}

	public static String removeThisSuperOrConstructorRefToAdvicePC(String methodName, JmlTypeDeclaration typeDecl){
		methodName = methodName.replace("this.", "");
		methodName = methodName.replace("super.", "");
		int index = methodName.indexOf(typeDecl.ident());
		if(!(index == -1))
			methodName = methodName.substring(index).replace(typeDecl.ident(), "new");
		return methodName;
	}

	public static String replaceDollaSymbol(String member){
		return member.replace("$", ".");
	}

	public static String processMethSig(String sig){
		return sig.replace("\\bigint", "java.math.BigInteger");
	}

	public static boolean hasAssertion(String pred){
		if(pred == null){
			return false;
		}
		boolean result = ((!pred.equals("")) && (!pred.equals(" ")) && (!pred.equals("true")) 
				&& (!pred.equals("(true)")) && (!pred.equals("null")));
//		boolean result = ((!pred.equals("")) && (!pred.equals(" ")) && (!pred.equals("null")));
		return result;
	}
	
	public static boolean hasPrecondition(String pred){
		if(pred == null){
			return false;
		}
//		boolean result = ((!pred.equals("")) && (!pred.equals(" ")) && (!pred.equals("true")) 
//				&& (!pred.equals("(true)")) && (!pred.equals("null")));
		boolean result = ((!pred.equals("")) && (!pred.equals(" ")) && (!pred.equals("null")));
		return result;
	}

	public void addJavaFileAsString(String strFile, File file){
		qDoxFile.addSource(new StringReader(strFile));
		qDoxFile.getSources()[qDoxFile.getSources().length-1].setFile(file);
		// renaming types with appropriate package names if needed --- [[[hemr]]]
		// this happens when using multiple types with different package names 
		QDoxUtil.processJavaTypesWithRightFullQualifiedNames(qDoxFile);
	}
	
	public boolean isFieldDeclaredInAnyValidAspectJMLFile(String fieldIdent, String typeFullQualifiedName){
		boolean result = false;
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();
		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(currentJavaDeclType.getFullyQualifiedName().equals(typeFullQualifiedName)){
				JavaField foundField = currentJavaDeclType.getFieldByName(fieldIdent);
				if(foundField != null){
					result = true;
				}
			}
		}
		return result;
	}
	
	public boolean isJavaField(String fieldIdent, String typeFullQualifiedName){
		boolean result = false;
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();
		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(currentJavaDeclType.getFullyQualifiedName().equals(typeFullQualifiedName)){
				try {
					if(currentJavaDeclType.getSource().getFile().getCanonicalPath().endsWith(".java")) {
						JavaField foundField = currentJavaDeclType.getFieldByName(fieldIdent);
						if(foundField != null){
							result = true;
						}
					}
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
		return result;
	}

	public List<JavaMethod> getAllDeclaredJavaMethodsFromATypeInFile(String typeFullQualifiedName){
		List<JavaMethod> result = new ArrayList<JavaMethod>();
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(currentJavaDeclType.getFullyQualifiedName().equals(typeFullQualifiedName)){
				JavaMethod [] javaDeclMethods = currentJavaDeclType.getMethods();
				for (int j = 0; j < javaDeclMethods.length; j++) {
					JavaMethod currentJavaDeclMethod = javaDeclMethods[j];
					result.add(currentJavaDeclMethod);
				}
			}
		}
		Comparator<JavaMethod> COMPARE_BY_LINE_NUMBER = new Comparator<JavaMethod>() {
			public int compare(JavaMethod one, JavaMethod other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	public List<JavaField> getAllDeclaredJavaFieldsFromATypeInFile(String typeFullQualifiedName, boolean drType){
		List<JavaField> result = new ArrayList<JavaField>();
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			JavaClass topOwner = QDoxUtil.getTopOfTopOwner(currentJavaDeclType);
			if(topOwner.getFullyQualifiedName().equals(typeFullQualifiedName)){
				if(drType){
					if(isTypeDeclDR(topOwner)){
						JavaField [] javaDeclFields = currentJavaDeclType.getFields();
						for (int j = 0; j < javaDeclFields.length; j++) {
							JavaField currentJavaDeclField = javaDeclFields[j];
							result.add(currentJavaDeclField);
						}
					}
				}
				else if(!isTypeDeclDR(topOwner)){
					JavaField [] javaDeclFields = currentJavaDeclType.getFields();
					for (int j = 0; j < javaDeclFields.length; j++) {
						JavaField currentJavaDeclField = javaDeclFields[j];
						if(!QDoxUtil.isFieldDeclPartial(currentJavaDeclField)){
							result.add(currentJavaDeclField);
						}
					}
				}
			}
		}
		Comparator<JavaField> COMPARE_BY_LINE_NUMBER = new Comparator<JavaField>() {
			public int compare(JavaField one, JavaField other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	public List<JavaMethod> getAllDeclaredJavaMethodsFromATypeInFile(String typeFullQualifiedName, boolean drType){
		List<JavaMethod> result = new ArrayList<JavaMethod>();
		List allDeclaredMethods = this.getAllJavaMethodDeclarations();
		
		for (Iterator iterator = allDeclaredMethods.iterator(); iterator.hasNext();) {
			JavaMethod currentJavaDeclMethod = (JavaMethod) iterator.next();
			JavaClass topOwner = QDoxUtil.getTopOfTopOwner(currentJavaDeclMethod.getParentClass());

			if(topOwner.getFullyQualifiedName().equals(typeFullQualifiedName)){
				if(drType){
					if(isTypeDeclDR(topOwner)){
						result.add(currentJavaDeclMethod);
					}
				}
				else{
					if(!isTypeDeclDR(topOwner)){
						result.add(currentJavaDeclMethod);
					}
				}
			}
		}
		
		Comparator<JavaMethod> COMPARE_BY_LINE_NUMBER = new Comparator<JavaMethod>() {
			public int compare(JavaMethod one, JavaMethod other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	public List<JavaClass> getAllImplementingInterfaceTypesFromATypeInFile(String typeFullQualifiedName, boolean drType){
		List<JavaClass> result = new ArrayList<JavaClass>();
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			JavaClass topOwner = QDoxUtil.getTopOfTopOwner(currentJavaDeclType);
			if(topOwner.getFullyQualifiedName().equals(typeFullQualifiedName)){
				if(drType){
					if(isTypeDeclDR(topOwner)){
						for (int j = 0; j < currentJavaDeclType.getImplementedInterfaces().length; j++) {
							result.add(currentJavaDeclType.getImplementedInterfaces()[j]);
						}
					}
				}
				else if(!isTypeDeclDR(topOwner)){
					for (int j = 0; j < currentJavaDeclType.getImplementedInterfaces().length; j++) {
						result.add(currentJavaDeclType.getImplementedInterfaces()[j]);
					}

				}
			}
		}
		return result;
	}
	
	public List<JavaClass> getSuperClassFromATypeInFile(String typeFullQualifiedName, boolean drType){
		List<JavaClass> result = new ArrayList<>();
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			JavaClass topOwner = QDoxUtil.getTopOfTopOwner(currentJavaDeclType);
			if(topOwner.getFullyQualifiedName().equals(typeFullQualifiedName)){
				if(drType){
					if(isTypeDeclDR(topOwner)){
						if(!currentJavaDeclType.getSuperJavaClass().getFullyQualifiedName().equals("java.lang.Object")){
							result.add(currentJavaDeclType.getSuperJavaClass());
						}
					}
				}
				else if(!isTypeDeclDR(topOwner)){
					if(!currentJavaDeclType.getSuperJavaClass().getFullyQualifiedName().equals("java.lang.Object")){
						result.add(currentJavaDeclType.getSuperJavaClass());
					}
				}
			}
		}
		return result;
	}
	
	public List<JavaClass> getDRInterfaceTypeImplementingInterfaceTypeInFile(String typeFullQualifiedName, String interfaceSig){
		List<JavaClass> result = new ArrayList<>();
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			JavaClass topOwner = QDoxUtil.getTopOfTopOwner(currentJavaDeclType);
			if(topOwner.getFullyQualifiedName().equals(typeFullQualifiedName)){
				if(isTypeDeclDR(topOwner)){
					for (int j = 0; j < currentJavaDeclType.getImplementedInterfaces().length; j++) {
						if(currentJavaDeclType.getImplementedInterfaces()[j].getFullyQualifiedName().equals(interfaceSig)){
							result.add(currentJavaDeclType);
						}
					}
				}
			}
		}
		return result;
	}
	
	public List<JavaClass> getDRTypesExtendingATypeFromInFile(String typeFullQualifiedName, String superClassSig){
		List<JavaClass> result = new ArrayList<>();
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			JavaClass topOwner = QDoxUtil.getTopOfTopOwner(currentJavaDeclType);
			if(topOwner.getFullyQualifiedName().equals(typeFullQualifiedName)){
				if(isTypeDeclDR(topOwner)){
					if(currentJavaDeclType.getSuperJavaClass().getFullyQualifiedName().equals(superClassSig)){
						result.add(currentJavaDeclType);
					}
				}
			}
		}
		return result;
	}
	
	// processing the AspectJML precedence rules --- [[[hemr]]]
	public String getDeclarePrecedenceProcessed(String declarePrecedence, JavaClass javaClass){
		StringBuffer result = new StringBuffer("");
		//JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();

		String [] declarePrecedenceSplit = declarePrecedence.split(",");
		
		for (int i = 0; i < declarePrecedenceSplit.length; i++) {
			String currentPrecedenceType = declarePrecedenceSplit[i].trim();
			// annotation types remain the same... [[[hemr]]]
			if(currentPrecedenceType.contains("@")){
				if(result.length() > 0)
					result.append(", ");
				result.append(currentPrecedenceType);
			}
			// if the type pattern is exactly equals to "*"
			else if(currentPrecedenceType.equals("*")){
				if(result.length() > 0)
					result.append(", ");
				result.append(currentPrecedenceType);
			}
			else{
				// if it's not an annotation type or plain *, we need to handle some rules
				// Is it a compound type pattern?
				if(currentPrecedenceType.contains("||")||
						currentPrecedenceType.contains("&&")){
//					System.out.println("currentPrecedenceType<init> = "+currentPrecedenceType);
					String [] currentPrecedenceTypeSplitOr = currentPrecedenceType.replace("||", "#").replace("&&", "#").split("#");
					for (int j = 0; j < currentPrecedenceTypeSplitOr.length; j++) {
						String currentPrecedenceTypeOr = currentPrecedenceTypeSplitOr[j].replace("(", "").replace(")", "").trim();
//						System.out.println("--->"+currentPrecedenceTypeSplitOr[j]);
//						System.out.println("currentPrecedenceTypeOr = "+currentPrecedenceTypeOr);
//						System.out.println("currentPrecedenceTypeOr<surrogate> = "+getCurrentPrecedenceTypeProcessed(currentPrecedenceTypeOr));
//						System.out.println("currentPrecedenceType<temp:before> = "+currentPrecedenceType);
						currentPrecedenceType = StringUtils.replaceOnce(currentPrecedenceType, currentPrecedenceTypeOr, getCurrentPrecedenceTypeProcessed(currentPrecedenceTypeOr));
//						System.out.println("currentPrecedenceType<temp:after> = "+currentPrecedenceType);
					}
					if(result.length() > 0)
						result.append(", ");
//					System.out.println("currentPrecedenceType<after> = "+currentPrecedenceType);
					result.append(currentPrecedenceType);
				}
				else{
					if(result.length() > 0)
						result.append(", ");
					result.append(getCurrentPrecedenceTypeProcessed(currentPrecedenceType.replace("(", "").replace(")", "").trim()));
				}
			}
		}
//		System.out.println("<declare precedence: >"+result.toString());
		return result.toString();
	}
	
	private String getCurrentPrecedenceTypeProcessed(String currentPrecedenceType){
		String prefixNot = "";
		String currentPrecedenceTypeProcessed = "";
		if(currentPrecedenceType.startsWith("{") ||
				currentPrecedenceType.startsWith("[")){
			currentPrecedenceTypeProcessed = currentPrecedenceType.replace("{", "").replace("}", "").replace("[", "").replace("]", "");
		}
		else{
			if(currentPrecedenceType.startsWith("!")){
				currentPrecedenceType = currentPrecedenceType.replace("!", "");
				prefixNot = "!";
			}
			// new simple way after some tests 12/1/2017 --- [[[hemr]]]
			currentPrecedenceTypeProcessed = prefixNot+"#AspectJMLInnerCC_"+currentPrecedenceType.replace("*..", "*").replace("..*", "*").replace("..", "*").replace(".", "_");
			
			// old
//			if(currentPrecedenceType.replace("*..", "*").replace("..*", "*").replace("..", "*").contains(".")){
//				currentPrecedenceTypeProcessed = "("+prefixNot+"#AspectJMLRac_*.AspectJMLInnerCC_"+currentPrecedenceType.replace("*..", "*").replace("..*", "*").replace("..", "*").replace(".", "_") 
//						+ " || "+"(("+prefixNot+currentPrecedenceType+")"+" && "+"(!#AspectJMLRac_*.AspectJMLInnerCC_*))"+")";
//			}
//			else{
//				currentPrecedenceTypeProcessed = "("+prefixNot+"#AspectJMLRac_*.AspectJMLInnerCC_"+currentPrecedenceType.replace("*..", "*").replace("..*", "*").replace("..", "*")
//						+ " || "+"(("+prefixNot+currentPrecedenceType+")"+" && "+"(!#AspectJMLRac_*.AspectJMLInnerCC_*))"+")";
//			}
			
//			System.out.println("<declare precedence<return processing>: >"+currentPrecedenceTypeProcessed.replace("#", ""));
		}
		return currentPrecedenceTypeProcessed.replace("#", "*..");
	}
	
	private boolean isTypeDeclDR(JavaClass javaClass){
		boolean result = false;
		
		if(javaClass.isInner()){
			while(javaClass.getParentClass() != null){
				javaClass = javaClass.getParentClass();
			}
		}
		
		for (int i = 0; i < javaClass.getAnnotations().length; i++) {
			Annotation currentAnno = javaClass.getAnnotations()[i];
			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.DR"))){
				result = true;
				break;
			}
		}
		return result;
	}
	
	public boolean isDRInterfaceImplemented(List<JavaClass> interfaces, String drInterface){
		boolean result = false;
		for (Iterator iterator = interfaces.iterator(); iterator.hasNext();) {
			JavaClass currentInterface = (JavaClass) iterator.next();
			if(currentInterface.getFullyQualifiedName().equals(drInterface)){
				result = true;
			}
		}
		return result;
	}
	
	public boolean isDRSuperClassImplemented(List<JavaClass> superClasses, String drSuperClass){
		boolean result = false;
		for (Iterator iterator = superClasses.iterator(); iterator.hasNext();) {
			JavaClass currentInterface = (JavaClass) iterator.next();
			if(currentInterface.getFullyQualifiedName().equals(drSuperClass)){
				result = true;
			}
		}
		return result;
	}
	
	public boolean isDRMethodImplemented(List<JavaMethod> implMethods, String drMethSig){
		boolean result = false;
		for (Iterator iterator = implMethods.iterator(); iterator.hasNext();) {
			JavaMethod implMeth = (JavaMethod) iterator.next();
			// checking if the method was really implemented --- [[[hemr]]]
			if(!(implMeth.getCodeBlock().charAt(implMeth.getCodeBlock().length()-2) == ';')){
				String methSig = AspectUtil.getInstance().getFullyJavaMethodSignature(implMeth, false, true);
				if(methSig.equals(drMethSig)){
					result = true;
				}
			}
		}
		return result;
	}
	
	public boolean isDRFieldImplemented(List<JavaField> implFields, String drFieldSig){
		boolean result = false;
		for (Iterator iterator = implFields.iterator(); iterator.hasNext();) {
			JavaField implField = (JavaField) iterator.next();
			String fieldSig = AspectUtil.getInstance().getFullyJavaFieldSignature(implField);
			if(fieldSig.equals(drFieldSig)){
				result = true;
			}
		}
		return result;
	}
	
	public String getFullyJavaFieldSignature(JavaField javaField){
		StringBuffer result = new StringBuffer("");
		// modifiers
		String [] modifiers = javaField.getModifiers();
		for (int i = 0; i < modifiers.length; i++) {
			String currentModifier = modifiers[i];
			result.append(currentModifier).append(" ");
		}
		result.append(javaField.getType().getFullyQualifiedName()).append(" ");
		String fieldName = javaField.getName();
		result.append(fieldName);
		return result.toString();
	}
	
	public String getFullyJavaMethodSignature(JavaMethod javaMethod, boolean paramNames, boolean exception){
		StringBuffer result = new StringBuffer("");
		// modifiers
		String [] modifiers = javaMethod.getModifiers();
		for (int i = 0; i < modifiers.length; i++) {
			String currentModifier = modifiers[i];
			result.append(currentModifier).append(" ");
		}
		if(!javaMethod.isConstructor()){
			if(javaMethod.getReturnType() != null){
				result.append(javaMethod.getReturnType().getFullyQualifiedName()).append(" ");
			}
			else {
				result.append("void").append(" ");
			}
		}
		String methName = javaMethod.getName();
		if(javaMethod.isConstructor()){
			methName = "new";
		}
		if(isAspectJMLReviseMethod(javaMethod)){
			result.append(javaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+"."+methName);
		}
		else{
			result.append(methName);
		}
		String parameters = QDoxUtil.getSignatureErasure(AspectUtil.generateMethodParameters(javaMethod.getParameters(), false).toString(), javaMethod);
		result.append(parameters);
		if(exception){
			Type[] exceptions = javaMethod.getExceptions();
			if(exceptions.length > 0){
				result.append(" throws ");
				for (int i = 0; i < exceptions.length; i++) {
					if(i > 0){
						result.append(", ");
					}
					result.append(exceptions[i].getFullyQualifiedName());
				}
			}
		}
		return result.toString();
	}
	
	public String getFullyJavaMethodSignatureWithOwnerName(JavaMethod javaMethod){
		StringBuffer result = new StringBuffer("");
		// modifiers
		String [] modifiers = javaMethod.getModifiers();
		for (int i = 0; i < modifiers.length; i++) {
			String currentModifier = modifiers[i];
			result.append(currentModifier).append(" ");
		}
		if(!javaMethod.isConstructor()){
			if(javaMethod.getReturnType() != null){
				result.append(javaMethod.getReturnType().getFullyQualifiedName()).append(" ");
			}
			else {
				result.append("void").append(" ");
			}
		}
		else{
			result.append("void").append(" ");
		}
		String methName = javaMethod.getName();
		if(this.isAspectJMLReviseMethod(javaMethod)){
			if(javaMethod.isConstructor()){
				methName = "new$Revise";
			}
			else{
				methName += "$Revise";
			}
			methName += "$"+javaMethod.getSource().getFile().getName().replace(".", "_");
		}
		else if(this.isAspectJMLAppendLikeMethod(javaMethod)){
			if(this.isAspectJMLAppendBefore(javaMethod)){
				methName += "$AppendBefore";
			}
			else if(this.isAspectJMLAppendAfter(javaMethod)){
				methName += "$AppendAfter";
			}
			else if(this.isAspectJMLAppendAfterReturning(javaMethod)){
				methName += "$AppendAfterReturning";
			}
			else if(this.isAspectJMLAppendAfterThrowing(javaMethod)){
				methName += "$AppendAfterThrowing";
			}
			methName += "$"+javaMethod.getSource().getFile().getName().replace(".", "_");
		}
		
		String plus = javaMethod.isStatic()? "":"+";
		result.append(javaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+plus+"."+methName);
		String parameters = QDoxUtil.getSignatureErasure(AspectUtil.generateMethodParameters(javaMethod.getParameters(), false).toString(), javaMethod);
		result.append(parameters);
		Type[] exceptions = javaMethod.getExceptions();
		if(exceptions.length > 0){
			result.append(" throws ");
			for (int i = 0; i < exceptions.length; i++) {
				if(i > 0){
					result.append(", ");
				}
				result.append(exceptions[i].getFullyQualifiedName());
			}
		}
		
		return result.toString();
	}
	
	public JavaMethod getCorrespondingJavaMethodThroughJMLMethod(String typeFullQualifiedName, JmlMethodDeclaration methodDecl){
		JavaMethod javaMethod = null;
		List<JavaMethod> javaMeths = AspectUtil.getInstance().getAllDeclaredJavaMethodsFromATypeInFile(typeFullQualifiedName);
		String jmlMeth = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false);
		boolean needsImportCompletion = AspectUtil.needsImportCompletion(methodDecl.parameters());
//		System.out.println("************************************************************************");
		for (Iterator<JavaMethod> iterator = javaMeths.iterator(); iterator.hasNext();) {
			JavaMethod currentJavaMeth = iterator.next();
			needsImportCompletion = AspectUtil.needsImportCompletion(currentJavaMeth.getParameters());
			String currentMethSig = currentJavaMeth.getName()+QDoxUtil.getSignatureErasure(AspectUtil.generateMethodParameters(currentJavaMeth.getParameters(), false).toString(), currentJavaMeth);
//			System.out.println("currentMethSig = "+currentMethSig);
//			System.out.println("jmlMeth = "+jmlMeth);
			if(jmlMeth.equals(currentMethSig)){
				if(methodDecl.getTokenReference().file().getAbsolutePath().equals(currentJavaMeth.getParentClass().getSource().getFile().getAbsolutePath())){
					javaMethod = currentJavaMeth;
					break;
				}
			}
			else if(needsImportCompletion){
				String methodNameWithParameterTypesCompletion = methodDecl.ident()+AspectUtil.generatePossibleMethodParametersMatch(methodDecl.parameters(), currentJavaMeth.getParameters()).toString();
//				System.out.println("methodNameWithParameterTypesCompletion = "+methodNameWithParameterTypesCompletion);
				if(methodNameWithParameterTypesCompletion.equals(currentMethSig)){
					if(methodDecl.getTokenReference().file().getAbsolutePath().equals(currentJavaMeth.getParentClass().getSource().getFile().getAbsolutePath())){
						javaMethod = currentJavaMeth;
						break;
					}
				}
			}
		}
//		System.out.println("************************************************************************");
//		System.out.println("");
		return javaMethod;
	}
	
//	public List<JavaMethod> getCorrespondingITDJavaMethodThroughJMLMethod(JmlMethodDeclaration methodDecl){
//		List<JavaMethod> javaMethods = new ArrayList();
//		List<JavaMethod> javaMeths = QDoxUtil.getAllDeclaredJavaMethodsInFile2(this.getAllJavaTypeDeclarationsAsArray());
//		String jmlMeth = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false);
////		boolean needsImportCompletion = AspectUtil.needsImportCompletion(methodDecl.parameters());
//		boolean needsImportCompletion = true; // for now... better to check!! --- [[[hemr]]]
//		for (Iterator<JavaMethod> iterator = javaMeths.iterator(); iterator.hasNext();) {
//			JavaMethod currentJavaMeth = iterator.next();
//			String currentMethSig = currentJavaMeth.getName()+QDoxUtil.getSignatureErasure(AspectUtil.generateMethodParameters(currentJavaMeth.getParameters(), false).toString(), currentJavaMeth);
//			if(jmlMeth.equals(currentMethSig)){
//				javaMethods.add(currentJavaMeth);
////				break;
//			}
//			else if(needsImportCompletion){
//				String methodNameWithParameterTypesCompletion = methodDecl.ident()+AspectUtil.generatePossibleMethodParametersMatch2(methodDecl.parameters(), currentJavaMeth.getParameters()).toString();
//				if(methodNameWithParameterTypesCompletion.equals(currentMethSig)){
//					javaMethods.add(currentJavaMeth);
////					break;
//				}
//			}
//		}
//		
//		return javaMethods;
//	}

	public static StringBuffer generateMethodParametersForAdvicePointcut(JFormalParameter[] parameters) {
		final StringBuffer code = new StringBuffer("");
		if (parameters != null) {
			for (int i = 0; i < parameters.length; i++) {
				if (i != 0)
					code.append(", ");
				parameters[i].accept( new JmlAbstractVisitor() {
					public void	visitJmlFormalParameter(
							JmlFormalParameter self) {
						CType actualType = self.getType();
						code.append(actualType.toString());
					}
				});
			}
		}
		return code;
	}

	public static String generateMethodParameters(CSpecializedType[] parameters) {
		final StringBuffer code = new StringBuffer("");
		code.append("(");
		if (parameters != null) {
			for (int i = 0; i < parameters.length; i++) {
				if (i != 0) {
					code.append(", ");	
				}
				CType actualType = parameters[i].staticType();
				code.append(TransUtils.toString(actualType));
			}
		}
		code.append(")");
		return code.toString();
	}

	public static StringBuffer generateMethodParameters(JFormalParameter[] parameters, final boolean includeIdentName) {
		final StringBuffer code = new StringBuffer("");
		code.append("(");
		if (parameters != null) {
			for (int i = 0; i < parameters.length; i++) {
				if (i != 0) {
					code.append(", ");	
				}
				parameters[i].accept( new JmlAbstractVisitor() {
					public void	visitJmlFormalParameter(
							JmlFormalParameter self) {
						CType actualType = self.getType();
						String ident = self.ident();
						code.append(TransUtils.toString(actualType));
						if(includeIdentName){
							code.append(" ");
							code.append(ident);
						}
					}
				});	
			}
		}
		code.append(")");
		return code;
	}
	
	public static StringBuffer generateMethodParametersOnlyParamNames(JFormalParameter[] parameters) {
		final StringBuffer code = new StringBuffer("");
		code.append("(");
		if (parameters != null) {
			for (int i = 0; i < parameters.length; i++) {
				if (i != 0) {
					code.append(", ");	
				}
				parameters[i].accept( new JmlAbstractVisitor() {
					public void	visitJmlFormalParameter(
							JmlFormalParameter self) {
						CType actualType = self.getType();
						String ident = self.ident();
						code.append(ident);
					}
				});	
			}
		}
		code.append(")");
		return code;
	}

	public static StringBuffer generateMethodParameters(JavaParameter[] parameters, boolean includeIdentName) {
		final StringBuffer code = new StringBuffer("");
		code.append("(");
		if (parameters != null) {
			for (int i = 0; i < parameters.length; i++) {
				if (i != 0){
					code.append(", ");	
				}
				code.append(parameters[i].getType()+"".replace("$", "."));
				if(includeIdentName){
					code.append(" ");
					code.append(parameters[i].getName());
				}
			}
		}
		code.append(")");
		return code;
	}
	
	public static StringBuffer generateMethodParametersOnlyParamNames(JavaParameter[] parameters) {
		final StringBuffer code = new StringBuffer("");
		code.append("(");
		if (parameters != null) {
			for (int i = 0; i < parameters.length; i++) {
				if (i != 0){
					code.append(", ");	
				}
				code.append(parameters[i].getName());
			}
		}
		code.append(")");
		return code;
	}
	
	public static boolean needsImportCompletion(JavaParameter[] parameters) {
		boolean result = false;
		if (parameters != null) {
			for (int i = 0; i < parameters.length; i++) {
				String type = parameters[i].getType()+"".replace("$", ".");
				if(!type.contains(".")){
					result = true;
				}
			}
		}
		
		return result;
	}
	
	public static boolean needsImportCompletion(JFormalParameter[] parameters) {
		boolean result = false;
		if (parameters != null) {
			for (int i = 0; i < parameters.length; i++) {
				CType actualType = parameters[i].getType();
				String type = TransUtils.toString(actualType);
				if(!type.contains(".")){
					result = true;
				}
			}
		}
		
		return result;
	}
	
	public static boolean needsImportCompletionArgumentType(String parameter) {
		boolean result = false;
		if(!parameter.contains(".")){
			result = true;
		}
		
		return result;
	}
	
	public static StringBuffer generatePossibleMethodParametersMatch(JFormalParameter[] parameters, JavaParameter[] parametersCompletion) {
		final StringBuffer code = new StringBuffer("");
		code.append("(");
		if (parameters != null) {
			for (int i = 0; i < parameters.length; i++) {
				if (i != 0) {
					code.append(", ");	
				}
				CType actualType = parameters[i].getType();
				if((parameters.length == parametersCompletion.length) && (AspectUtil.needsImportCompletionArgumentType(TransUtils.toString(actualType)) || 
						AspectUtil.needsImportCompletionArgumentType(parametersCompletion[i].getType().getFullyQualifiedName())) ){
					String parameterCompletion = parametersCompletion[i].getType()+"".replace("$", ".");
					// [[[bug]]] need to look at some point if needed!! --- [[[hemr]]]
//					if(parameterCompletion.endsWith(TransUtils.toString(actualType))){
//						code.append(parameterCompletion);
//					}
					if(TransUtils.toString(actualType).endsWith(parameterCompletion)){
						code.append(parameterCompletion);
					}
					else{
						code.append(TransUtils.toString(actualType));
					}
				}
				else{
					code.append(TransUtils.toString(actualType));
				}
			}
		}
		code.append(")");
		return code;
	}
	
	public static StringBuffer generatePossibleMethodParametersMatch2(JFormalParameter[] parameters, JavaParameter[] parametersCompletion) {
		final StringBuffer code = new StringBuffer("");
		code.append("(");
		if (parameters != null) {
			if(parameters.length == parametersCompletion.length){
				for (int i = 0; i < parameters.length; i++) {
					if (i != 0) {
						code.append(", ");	
					}
					CType actualType = parameters[i].getType();
					String surrogate = parameters[i].getType().toString().substring(parameters[i].getType().toString().lastIndexOf(".")+1, 
							parameters[i].getType().toString().length());
					if((AspectUtil.needsImportCompletionArgumentType(TransUtils.toString(actualType)) || 
							AspectUtil.needsImportCompletionArgumentType(parametersCompletion[i].getType().getFullyQualifiedName())) 
							&& (parameters.length == parametersCompletion.length)){
						String parameterCompletion = parametersCompletion[i].getType()+"".replace("$", ".");
						if(parameterCompletion.endsWith(TransUtils.toString(actualType))){
							code.append(parameterCompletion);
						}
						else{
							code.append(surrogate);
						}
					}
					else{
						code.append(TransUtils.toString(actualType));
					}
				}
			}
		}
		code.append(")");
		return code;
	}
	
	private boolean isParameterEquivalent(JFormalParameter[] parameters, JavaParameter[] parameters2){
		boolean result = true;
		
		if(parameters.length == parameters2.length){
			for (int i = 0; i < parameters.length; i++) {
				CType actualType = parameters[i].getType();
				String paramType = TransUtils.toString(actualType);
				String paramType2 = parameters2[i].getType()+"".replace("$", ".");
				if(!(paramType.equals(paramType2))){
					if(paramType.lastIndexOf(".") != -1){
						paramType = paramType.substring((paramType.lastIndexOf(".")+1), (paramType.length()));
						if(!(paramType.equals(paramType2))){
							result = false;
							break;
						}
					}
					else {
						result = false;
						break;
					}
				}
			}
		}
		else {
			result = false;
		}
		
		return result;
	}

	public boolean isCrosscutSpecChecking(JmlMethodDeclaration methodDecl){
		boolean isCrosscutSpecMeth = false;
		String classQualifiedName = methodDecl.getMethod().owner().getJavaName();
		String methodNameWithParameterTypes = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false).toString();
		boolean needsImportCompletion = AspectUtil.needsImportCompletion(methodDecl.parameters());
		List javaMethCrosscutSpecChecking = this.getCrosscutSpecMeths();
		for (Iterator iterator = javaMethCrosscutSpecChecking.iterator(); iterator
				.hasNext();) {
			JavaMethod currentJavaMethod = (JavaMethod) iterator.next();
			if(classQualifiedName.equals(currentJavaMethod.getParentClass().getFullyQualifiedName())){
				if(methodDecl.getTokenReference().file().getAbsolutePath().equals(currentJavaMethod.getParentClass().getSource().getFile().getAbsolutePath())){
					String currentJavaMethNameWithParameterTypes = currentJavaMethod.getName()+
							QDoxUtil.getSignatureErasure(AspectUtil.generateMethodParameters(currentJavaMethod.getParameters(), false).toString(), currentJavaMethod);
					if(methodNameWithParameterTypes.equals(currentJavaMethNameWithParameterTypes)){
						isCrosscutSpecMeth = true;
						break;
					}
					else if(methodDecl.ident().equals(currentJavaMethod.getName())){
						if(isParameterEquivalent(methodDecl.parameters(), currentJavaMethod.getParameters())){
							isCrosscutSpecMeth = true;
							break;
						}
					}
					else if(needsImportCompletion){
						String methodNameWithParameterTypesCompletion = methodDecl.ident()+AspectUtil.generatePossibleMethodParametersMatch(methodDecl.parameters(), currentJavaMethod.getParameters()).toString();
						if(methodNameWithParameterTypesCompletion.equals(currentJavaMethNameWithParameterTypes)){
							isCrosscutSpecMeth = true;
							break;
						}
					}
				}
			}
		}
		// check if besides pc declaration it really has JML specs
		if(isCrosscutSpecMeth){
			// if does not have JML specs, it is only considered an ordinary pointcut declaration for AspectJ... [[[hemr]]]
			if(!(methodDecl.hasSpecification() && methodDecl.methodSpecification().hasSpecCases())){
				isCrosscutSpecMeth = false;
			}
		}
		return isCrosscutSpecMeth;
	}
	
	public boolean isXCSFlexible(JavaMethod jm){
		boolean result = false;
		for (int i = 0; i < jm.getAnnotations().length; i++) {
    		Annotation currentAnno = jm.getAnnotations()[i];
    		if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Flexible")) || 
    				(currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Global"))){
    			result = true;
    		}
    	}
		return result;
	}
	
	public boolean isAspectJMLAdviceMethod(JavaMethod jm){
		boolean result = false;
		for (int i = 0; i < jm.getAnnotations().length; i++) {
    		Annotation currentAnno = jm.getAnnotations()[i];
    		if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Before")) || 
    				(currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.After")) || 
    				  (currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Around"))){
    			result = true;
    			break;
    		}
    	}
		return result;
	}
	
	public boolean isAspectJMLReviseMethod(JavaMethod jm){
		boolean result = false;
		for (int i = 0; i < jm.getAnnotations().length; i++) {
    		Annotation currentAnno = jm.getAnnotations()[i];
    		if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Revise")) ){
    			result = true;
    			break;
    		}
    	}
		return result;
	}
	
	public boolean isAspectJMLAppendLikeMethod(JavaMethod jm){
		boolean result = false;
		for (int i = 0; i < jm.getAnnotations().length; i++) {
    		Annotation currentAnno = jm.getAnnotations()[i];
    		if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Append"))||
    				(currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Add"))){
    			result = true;
    			break;
    		}
    	}
		return result;
	}
	
	public boolean isAspectJMLAppendBefore(JavaMethod jm){
		boolean result = false;
		for (int i = 0; i < jm.getAnnotations().length; i++) {
    		Annotation currentAnno = jm.getAnnotations()[i];
    		if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.AppendBefore"))||
    				(currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.AddBefore"))){
    			result = true;
    			break;
    		}
    	}
		return result;
	}
	
	public boolean isAspectJMLAppendAfter(JavaMethod jm){
		boolean result = false;
		for (int i = 0; i < jm.getAnnotations().length; i++) {
    		Annotation currentAnno = jm.getAnnotations()[i];
    		if((currentAnno.toString().equals("@org.aspectjml.lang.annotation.AppendAfter()"))||
    				(currentAnno.toString().equals("@org.aspectjml.lang.annotation.AddAfter()"))){
    			result = true;
    			break;
    		}
    	}
		return result;
	}
	
	public boolean isAspectJMLAppendAfterReturning(JavaMethod jm){
		boolean result = false;
		for (int i = 0; i < jm.getAnnotations().length; i++) {
    		Annotation currentAnno = jm.getAnnotations()[i];
    		if((currentAnno.toString().equals("@org.aspectjml.lang.annotation.AppendAfterReturning()"))||
    				(currentAnno.toString().equals("@org.aspectjml.lang.annotation.AddAfterReturning()"))){
    			result = true;
    			break;
    		}
    	}
		return result;
	}
	
	public boolean isAspectJMLAppendAfterThrowing(JavaMethod jm){
		boolean result = false;
		for (int i = 0; i < jm.getAnnotations().length; i++) {
    		Annotation currentAnno = jm.getAnnotations()[i];
    		if((currentAnno.toString().equals("@org.aspectjml.lang.annotation.AppendAfterThrowing()"))||
    				(currentAnno.toString().equals("@org.aspectjml.lang.annotation.AddAfterThrowing()"))){
    			result = true;
    			break;
    		}
    	}
		return result;
	}
	
	public boolean isAspectJAdviceMethod(JavaMethod jm){
		boolean result = false;
		for (int i = 0; i < jm.getAnnotations().length; i++) {
    		Annotation currentAnno = jm.getAnnotations()[i];
    		if((currentAnno.toString().startsWith("@org.aspectj.lang.annotation.Before")) || 
    				(currentAnno.toString().startsWith("@org.aspectj.lang.annotation.After")) || 
    				  (currentAnno.toString().startsWith("@org.aspectj.lang.annotation.Around"))){
    			result = true;
    			break;
    		}
    	}
		return result;
	}
	
	public boolean isHashCodeMeth(JmlMethodDeclaration methodDecl){
		boolean isHashCodeMeth = false;
		
		if(methodDecl.ident().equals("hashCode")){
			if((methodDecl.returnType().toString().equals("int")) && (methodDecl.parameters().length == 0)){
				isHashCodeMeth = true;
			}
		}
		return isHashCodeMeth;
	}

	public JavaMethod getJavaMethCrosscutSpecChecking(JmlMethodDeclaration methodDecl){
		JavaMethod result = null;
		String classQualifiedName = methodDecl.getMethod().owner().getJavaName();
		String methodNameWithParameterTypes = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false).toString();
		boolean needsImportCompletion = AspectUtil.needsImportCompletion(methodDecl.parameters());
		List javaMethXCS = this.getCrosscutSpecMeths();
		for (Iterator iterator = javaMethXCS.iterator(); iterator
				.hasNext();) {
			JavaMethod currentJavaMethod = (JavaMethod) iterator.next();
			if(classQualifiedName.equals(currentJavaMethod.getParentClass().getFullyQualifiedName())){
				if(methodDecl.getTokenReference().file().getAbsolutePath().equals(currentJavaMethod.getParentClass().getSource().getFile().getAbsolutePath())){
					String currentJavaMethNameWithParameterTypes = currentJavaMethod.getName()+AspectUtil.generateMethodParameters(currentJavaMethod.getParameters(), false).toString();
					if(methodNameWithParameterTypes.equals(currentJavaMethNameWithParameterTypes)){
						result = currentJavaMethod;
						break;
					}
					else if(methodDecl.ident().equals(currentJavaMethod.getName())){
						if(isParameterEquivalent(methodDecl.parameters(), currentJavaMethod.getParameters())){
							result = currentJavaMethod;
							break;
						}
					}
					else if(needsImportCompletion){
						String methodNameWithParameterTypesCompletion = methodDecl.ident()+AspectUtil.generatePossibleMethodParametersMatch(methodDecl.parameters(), currentJavaMethod.getParameters()).toString();
						if(methodNameWithParameterTypesCompletion.equals(currentJavaMethNameWithParameterTypes)){
							result = currentJavaMethod;
							break;
						}
					}
				}
			}
		}
		return result;
	}
	
	public JavaMethod getJavaMethAdviceCrosscutting(JmlMethodDeclaration methodDecl){
		JavaMethod result = null;
		String classQualifiedName = methodDecl.getMethod().owner().getJavaName();
		String methodNameWithParameterTypes = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false).toString();
		boolean needsImportCompletion = AspectUtil.needsImportCompletion(methodDecl.parameters());
		List javaMethXCS = this.getAdviceDeclJavaFile();
		for (Iterator iterator = javaMethXCS.iterator(); iterator
				.hasNext();) {
			JavaMethod currentJavaMethod = (JavaMethod) iterator.next();
			if(classQualifiedName.equals(currentJavaMethod.getParentClass().getFullyQualifiedName())){
				if(methodDecl.getTokenReference().file().getAbsolutePath().equals(currentJavaMethod.getParentClass().getSource().getFile().getAbsolutePath())){
					String currentJavaMethNameWithParameterTypes = currentJavaMethod.getName()+QDoxUtil.getSignatureErasure(AspectUtil.generateMethodParameters(currentJavaMethod.getParameters(), false).toString(), currentJavaMethod);
					if(methodNameWithParameterTypes.equals(currentJavaMethNameWithParameterTypes)){
						result = currentJavaMethod;
						break;
					}
					else if(methodDecl.ident().equals(currentJavaMethod.getName())){
						if(isParameterEquivalent(methodDecl.parameters(), currentJavaMethod.getParameters())){
							result = currentJavaMethod;
							break;
						}
					}
					else if(needsImportCompletion){
						String methodNameWithParameterTypesCompletion = methodDecl.ident()+AspectUtil.generatePossibleMethodParametersMatch(methodDecl.parameters(), currentJavaMethod.getParameters()).toString();
						if(methodNameWithParameterTypesCompletion.equals(currentJavaMethNameWithParameterTypes)){
							result = currentJavaMethod;
							break;
						}
					}
				}
			}
		}
		return result;
	}
	
	public boolean isJavaMethDeclSealed(JmlMethodDeclaration methodDecl){
		boolean isSealed = false;
		JavaMethod javaMeth = this.getQDoxJavaMethFromJML(methodDecl);
		if(javaMeth != null){
			if(this.isDeclSealed(javaMeth)){
				isSealed = true;
			}
		}
		else{
			String methodNameWithParameterTypes = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false).toString();
			try {
				if(methodDecl.isConstructor()){
					Constructor<?>[] constructors = Class.forName(methodDecl.getMethod().owner().getJavaName()).getDeclaredConstructors();
					for (int i = 0; i < constructors.length; i++) {
						Constructor<?> currentConstructor = constructors[i];
						String fullSignature = ReflectUtil.getConstructorSignatureWithLongTypeNames(currentConstructor);
						if(methodNameWithParameterTypes.equals(fullSignature)){
							if(this.isDeclSealed(currentConstructor)){
								isSealed = true;
								break;
							}
						}
					}
				}
				else{
					Method [] methods = Class.forName(methodDecl.getMethod().owner().getJavaName()).getDeclaredMethods();
					for (int i = 0; i < methods.length; i++) {
						Method currentMeth = methods[i];
						String fullSignature = ReflectUtil.getMethodSignatureWithLongTypeNames(currentMeth);
						if(methodNameWithParameterTypes.equals(fullSignature)){
							if(this.isDeclSealed(currentMeth)){
								isSealed = true;
								break;
							}
						}
					}
				}
				
			} catch (ClassNotFoundException e) {
				
			}
		}
		return isSealed;
	}
	
	public JavaMethod getQDoxJavaMethFromJML(JmlMethodDeclaration methodDecl){
		JavaMethod result = null;
		String classQualifiedName = methodDecl.getMethod().owner().getJavaName();
		String methodNameWithParameterTypes = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false).toString();
		boolean needsImportCompletion = AspectUtil.needsImportCompletion(methodDecl.parameters());
		List javaMeths = QDoxUtil.getAllDeclaredJavaMethodsInFile2(this.getAllJavaTypeDeclarationsAsArray());
		for (Iterator iterator = javaMeths.iterator(); iterator
				.hasNext();) {
			JavaMethod currentJavaMethod = (JavaMethod) iterator.next();
			if(classQualifiedName.equals(currentJavaMethod.getParentClass().getFullyQualifiedName())){
				if(currentJavaMethod.getParentClass().getSource().getFile().getAbsolutePath().endsWith(".java")){
					String currentJavaMethNameWithParameterTypes = currentJavaMethod.getName()+QDoxUtil.getSignatureErasure(AspectUtil.generateMethodParameters(currentJavaMethod.getParameters(), false).toString(), currentJavaMethod);
					if(methodNameWithParameterTypes.equals(currentJavaMethNameWithParameterTypes)){
						result = currentJavaMethod;
						break;
					}
					else if(methodDecl.ident().equals(currentJavaMethod.getName())){
						if(isParameterEquivalent(methodDecl.parameters(), currentJavaMethod.getParameters())){
							result = currentJavaMethod;
							break;
						}
					}
					else if(needsImportCompletion){
						String methodNameWithParameterTypesCompletion = methodDecl.ident()+AspectUtil.generatePossibleMethodParametersMatch(methodDecl.parameters(), currentJavaMethod.getParameters()).toString();
						if(methodNameWithParameterTypesCompletion.equals(currentJavaMethNameWithParameterTypes)){
							result = currentJavaMethod;
							break;
						}
					}
				}
			}
		}
		return result;
	}
	
	public List<String> getAllImplementAndExtendClauseAsDeclareParentsInFile(String fileAbstractPath){
		List<String> result = new ArrayList<String>();
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(QDoxUtil.isTypeDeclOpen(currentJavaDeclType) || currentJavaDeclType.isInner()){
				if(currentJavaDeclType.getSource().getFile().getAbsolutePath().equals(fileAbstractPath)){
					// handling super-class
					if (currentJavaDeclType.getSuperJavaClass() != null)
					if(!currentJavaDeclType.getSuperJavaClass().getFullyQualifiedName().equals("java.lang.Object")){
						result.add("declare parents: "+currentJavaDeclType.getFullyQualifiedName().replace("$",".")+
								" extends "+currentJavaDeclType.getSuperJavaClass().getFullyQualifiedName().replace("$",".")+";");
					}
					// handling implemented interfaces
					JavaClass [] interfaces = currentJavaDeclType.getImplementedInterfaces();
					for (int j = 0; j < interfaces.length; j++) {
						result.add("declare parents: "+currentJavaDeclType.getFullyQualifiedName().replace("$",".")+
								" implements "+interfaces[j].getFullyQualifiedName().replace("$",".")+";");
					}
				}
			}
		}
		
		return result;
	}
	
	public JavaMethod getJavaMethDeclareSoftCrosscutting(JmlMethodDeclaration methodDecl){
		JavaMethod result = null;
		String classQualifiedName = methodDecl.getMethod().owner().getJavaName();
		String methodNameWithParameterTypes = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false).toString();
		boolean needsImportCompletion = AspectUtil.needsImportCompletion(methodDecl.parameters());
		List javaMethXCS = this.getDeclareSoftMethodDecls();
		for (Iterator iterator = javaMethXCS.iterator(); iterator
				.hasNext();) {
			JavaMethod currentJavaMethod = (JavaMethod) iterator.next();
			if(classQualifiedName.equals(currentJavaMethod.getParentClass().getFullyQualifiedName())){
				if(methodDecl.getTokenReference().file().getAbsolutePath().equals(currentJavaMethod.getSource().getFile().getAbsolutePath())){
					String currentJavaMethNameWithParameterTypes = currentJavaMethod.getName()+QDoxUtil.getSignatureErasure(AspectUtil.generateMethodParameters(currentJavaMethod.getParameters(), false).toString(), currentJavaMethod);
					if(methodNameWithParameterTypes.equals(currentJavaMethNameWithParameterTypes)){
						result = currentJavaMethod;
						break;
					}
					else if(methodDecl.ident().equals(currentJavaMethod.getName())){
						if(isParameterEquivalent(methodDecl.parameters(), currentJavaMethod.getParameters())){
							result = currentJavaMethod;
							break;
						}
					}
					else if(needsImportCompletion){
						String methodNameWithParameterTypesCompletion = methodDecl.ident()+AspectUtil.generatePossibleMethodParametersMatch(methodDecl.parameters(), currentJavaMethod.getParameters()).toString();
						if(methodNameWithParameterTypesCompletion.equals(currentJavaMethNameWithParameterTypes)){
							result = currentJavaMethod;
							break;
						}
					}
				}
			}
		}
		return result;
	}
	
	public boolean isCrosscuttingContractSpecificationForInterface(JavaMethod methodDecl){
		boolean isCrosscutSpecForInterface = false;
		if(methodDecl.getParentClass().getParentClass() != null){
			if(methodDecl.getParentClass().getParentClass().isInterface()){
				for (int i = 0; i < methodDecl.getParentClass().getAnnotations().length; i++) {
					if(methodDecl.getParentClass().getAnnotations()[i].toString().startsWith("@org.aspectjml.lang.annotation.InterfaceXCS()")){
						isCrosscutSpecForInterface = true;
					}
				}
			}
		}
		return isCrosscutSpecForInterface;
	}

	public boolean isAspectAdvice(JmlMethodDeclaration methodDecl){
		boolean isAspectMeth = false;
		String classQualifiedName = methodDecl.getMethod().owner().getJavaName();
		String methodNameWithParameterTypes = methodDecl.ident()+AspectUtil.generateMethodParameters(methodDecl.parameters(), false).toString();
		List javaAspectMeth = this.getAspectAdviceOrPCs();
		for (Iterator iterator = javaAspectMeth.iterator(); iterator
				.hasNext();) {
			JavaMethod currentJavaMethod = (JavaMethod) iterator.next();
			if(classQualifiedName.equals(currentJavaMethod.getParentClass().getFullyQualifiedName())){
				String currentJavaMethNameWithParameterTypes = currentJavaMethod.getName()+
						AspectUtil.generateMethodParameters(currentJavaMethod.getParameters(), false).toString();
				if(methodNameWithParameterTypes.equals(currentJavaMethNameWithParameterTypes)){
					isAspectMeth = true;
					break;
				}
				else if(methodDecl.ident().equals(currentJavaMethod.getName())){
					if(isParameterEquivalent(methodDecl.parameters(), currentJavaMethod.getParameters())){
						isAspectMeth = true;
						break;
					}
				}
			}
		}
		return isAspectMeth;
	}
	
	public boolean isAspectTypeHasMixedAdviceForms(JmlTypeDeclaration typeDecl){
		boolean result = false;
		boolean hasAspectJMLAdvice = false;
		boolean hasAspectJAdvice = false;
		JavaClass aspectType = getAspectTypeAsJavaClass(typeDecl);
		
		if(aspectType != null){
			JavaMethod [] jms = aspectType.getMethods();
			for (int i = 0; i < jms.length; i++) {
				JavaMethod currentJavaMethod = jms[i];
				if(isAspectJMLAdviceMethod(currentJavaMethod)){
					hasAspectJMLAdvice = true;
				}
				if(isAspectJAdviceMethod(currentJavaMethod)){
					hasAspectJAdvice = true;
				}
				if(hasAspectJMLAdvice && hasAspectJAdvice){
					break;
				}	
			}
		}
		result = hasAspectJMLAdvice && hasAspectJAdvice;
		return result;
	}
	
	public boolean isAspectTypeHasAnyAspectJMLAdvice(JmlTypeDeclaration typeDecl){
		boolean result = false;
		boolean hasAspectJMLAdvice = false;
//		boolean hasAspectJAdvice = false;
		JavaClass aspectType = getAspectTypeAsJavaClass(typeDecl);
		
		if(aspectType != null){
			JavaMethod [] jms = aspectType.getMethods();
			for (int i = 0; i < jms.length; i++) {
				JavaMethod currentJavaMethod = jms[i];
				if(isAspectJMLAdviceMethod(currentJavaMethod)){
					hasAspectJMLAdvice = true;
					break;
				}
			}
		}
		result = hasAspectJMLAdvice;
		return result;
	}
	
	public JavaClass getQDoxJavaClassFromJMLType(JmlTypeDeclaration typeDecl){
		JavaClass result = null;
		String classQualifiedName = AspectUtil.replaceDollaSymbol(typeDecl.getCClass().getJavaName());
		
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();
		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentType = javaDeclTypes[i];
			if(AspectUtil.replaceDollaSymbol(currentType.getFullyQualifiedName()).equals(classQualifiedName)){
				if(currentType.getSource().getFile().getAbsolutePath().equals(typeDecl.getTokenReference().file().getAbsolutePath())){
					result = currentType;
					break;
				}
			}
		}
		return result;
	}
	
	public boolean isDominantJavaTypeDeclFound(JmlTypeDeclaration typeDecl){
		boolean found = false;
		String classQualifiedName = AspectUtil.replaceDollaSymbol(typeDecl.getCClass().getJavaName());
		
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();
		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentType = javaDeclTypes[i];
			if(AspectUtil.replaceDollaSymbol(currentType.getFullyQualifiedName()).equals(classQualifiedName)){
				if(currentType.getSource().getFile().getAbsolutePath().endsWith(".java")){
					found = true;
					break;
				}
			}
		}
		// looking at the classpath
		if(!found){
			try {
				Class.forName(classQualifiedName);
				found = true;
			} catch (ClassNotFoundException e) {
				found = false;
			}
		}
		return found;
	}
	
	public boolean isDominantJavaTypeDeclSealed(JmlTypeDeclaration typeDecl){
		boolean found = false;
		boolean sealed = false;
		String classQualifiedName = AspectUtil.replaceDollaSymbol(typeDecl.getCClass().getJavaName());
		
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();
		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentType = javaDeclTypes[i];
			if(AspectUtil.replaceDollaSymbol(currentType.getFullyQualifiedName()).equals(classQualifiedName)){
				if(currentType.getSource().getFile().getAbsolutePath().endsWith(".java")){
					found = true;
					if(isDeclSealed(currentType)){
						sealed = true;
						break;
					}
				}
			}
		}
		// looking at the classpath
		if(!found){
			try {
				sealed = isDeclSealed(Class.forName(classQualifiedName));
			} catch (ClassNotFoundException e) {
				// not gonna happen... catch this scenario before
			}
		}
		return sealed;
	}

	public boolean isTypeDeclAnAspect(JmlTypeDeclaration typeDecl){
		boolean isAspectDecl = false;
		String classQualifiedName = AspectUtil.replaceDollaSymbol(typeDecl.getCClass().getJavaName());
		
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();
		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentType = javaDeclTypes[i];
			if(AspectUtil.replaceDollaSymbol(currentType.getFullyQualifiedName()).equals(classQualifiedName)){
				if(typeDecl.getTokenReference().file().getAbsolutePath().equals(currentType.getSource().getFile().getAbsolutePath())){
					if(isTypeDeclAspectDecl(currentType)){
						isAspectDecl = true;
						break;
					}
				}
			}
		}
		return isAspectDecl;
	}
	
	public JavaClass getAspectTypeAsJavaClass(JmlTypeDeclaration typeDecl){
		JavaClass result = null;
		String classQualifiedName = AspectUtil.replaceDollaSymbol(typeDecl.getCClass().getJavaName());
		
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();
		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentType = javaDeclTypes[i];
			if(AspectUtil.replaceDollaSymbol(currentType.getFullyQualifiedName()).equals(classQualifiedName)){
				if(typeDecl.getTokenReference().file().getAbsolutePath().equals(currentType.getSource().getFile().getAbsolutePath())){
					if(isTypeDeclAspectDecl(currentType)){
						result = currentType;
						break;
					}
				}
			}
		}
		return result;
	}
	
	public boolean isTypeDeclAspectDecl(JavaClass type) {
		boolean isAspect = false;
		for (int i = 0; i < type.getAnnotations().length; i++) {
			if(type.getAnnotations()[i].toString().startsWith("@org.aspectj.lang.annotation.Aspect")){
				isAspect = true;
				break;
			}
		}
		return isAspect;
	}
	
	public boolean isDeclSealed(JavaClass type) {
		boolean isSealed = false;

		for (int i = 0; i < type.getAnnotations().length; i++) {
			if(type.getAnnotations()[i].toString().startsWith("@org.aspectjml.lang.annotation.Sealed")){
				isSealed = true;
				break;
			}
		}
		return isSealed;
	}
	
	public boolean isDeclSealed(Class<?> type) {
		boolean isSealed = false;

		java.lang.annotation.Annotation[] typeAnnos = type.getAnnotations();
		for (int i = 0; i < typeAnnos.length; i++) {
			if(typeAnnos[i].annotationType().getName().equals("org.aspectjml.lang.annotation.Sealed")){
				isSealed = true;
				break;
			}
		}
		
		return isSealed;
	}
	
	public boolean isDeclSealed(Field field) {
		boolean isSealed = false;

		java.lang.annotation.Annotation[] typeAnnos = field.getAnnotations();
		for (int i = 0; i < typeAnnos.length; i++) {
			if(typeAnnos[i].annotationType().getName().equals("org.aspectjml.lang.annotation.Sealed")){
				isSealed = true;
				break;
			}
		}
		
		return isSealed;
	}
	
	public boolean isDeclSealed(Method method) {
		boolean isSealed = false;

		java.lang.annotation.Annotation[] typeAnnos = method.getAnnotations();
		for (int i = 0; i < typeAnnos.length; i++) {
			if(typeAnnos[i].annotationType().getName().equals("org.aspectjml.lang.annotation.Sealed")){
				isSealed = true;
				break;
			}
		}
		
		return isSealed;
	}
	
	public boolean isDeclSealed(Constructor<?> constructor) {
		boolean isSealed = false;

		java.lang.annotation.Annotation[] typeAnnos = constructor.getAnnotations();
		for (int i = 0; i < typeAnnos.length; i++) {
			if(typeAnnos[i].annotationType().getName().equals("org.aspectjml.lang.annotation.Sealed")){
				isSealed = true;
				break;
			}
		}
		
		return isSealed;
	}
	
	public boolean isDeclSealed(JavaMethod javaMeth) {
		boolean isSealed = false;

		for (int i = 0; i < javaMeth.getAnnotations().length; i++) {
			if(javaMeth.getAnnotations()[i].toString().startsWith("@org.aspectjml.lang.annotation.Sealed")){
				isSealed = true;
				break;
			}
		}
		return isSealed;
	}
	
	public boolean isMethDeclHelper(Constructor method) {
		boolean isHelper = false;

		java.lang.annotation.Annotation[] typeAnnos = method.getAnnotations();
		for (int i = 0; i < typeAnnos.length; i++) {
			if(typeAnnos[i].annotationType().getName().equals("org.aspectjml.lang.annotation.Helper")){
				isHelper = true;
				break;
			}
		}
		
		return isHelper;
	}
	
	public boolean isMethDeclHelper(Method method) {
		boolean isHelper = false;

		java.lang.annotation.Annotation[] typeAnnos = method.getAnnotations();
		for (int i = 0; i < typeAnnos.length; i++) {
			if(typeAnnos[i].annotationType().getName().equals("org.aspectjml.lang.annotation.Helper")){
				isHelper = true;
				break;
			}
		}
		
		return isHelper;
	}
	
//	public boolean isTypeDeclWrongOpenSpec(JmlTypeDeclaration typeDecl){
//		boolean result = false;
//		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();
//		String classQualifiedName = AspectUtil.replaceDollaSymbol(typeDecl.getCClass().getJavaName());
//		
//		for (int i = 0; i < javaDeclTypes.length; i++) {
//			JavaClass currentJavaDeclType = javaDeclTypes[i];
//			if(AspectUtil.replaceDollaSymbol(currentJavaDeclType.getFullyQualifiedName()).equals(classQualifiedName)){
//				if(typeDecl.getTokenReference().file().getAbsolutePath().equals(currentJavaDeclType.getSource().getFile().getAbsolutePath())){
//					if(QDoxUtil.isTypeDeclWrongOpenSpec(currentJavaDeclType)){
//						result = true;		
//						break;
//					}
//				}
//			}
//		}
//		return result;
//	}
	
	public boolean isTypeDeclSealed(JmlTypeDeclaration typeDecl){
		boolean result = false;
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();
		String classQualifiedName = AspectUtil.replaceDollaSymbol(typeDecl.getCClass().getJavaName());
		
		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(AspectUtil.replaceDollaSymbol(currentJavaDeclType.getFullyQualifiedName()).equals(classQualifiedName)){
				if(typeDecl.getTokenReference().file().getAbsolutePath().equals(currentJavaDeclType.getSource().getFile().getAbsolutePath())){
					if(isDeclSealed(currentJavaDeclType)){
						result = true;		
						break;
					}
				}
			}
		}
		return result;
	}
	
	public boolean isTypeDeclReadOnly(JmlTypeDeclaration typeDecl){
		boolean result = false;
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();
		String classQualifiedName = AspectUtil.replaceDollaSymbol(typeDecl.getCClass().getJavaName());
		
		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(AspectUtil.replaceDollaSymbol(currentJavaDeclType.getFullyQualifiedName()).equals(classQualifiedName)){
				if(typeDecl.getTokenReference().file().getAbsolutePath().equals(currentJavaDeclType.getSource().getFile().getAbsolutePath())){
					if(QDoxUtil.isTypeDeclReadOnly(currentJavaDeclType)){
						result = true;		
						break;
					}
				}
			}
		}
		return result;
	}
	
	public boolean isTypeDeclOpenSpecOrSpecsOnly(JmlTypeDeclaration typeDecl){
		boolean result = false;
		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();
		String classQualifiedName = AspectUtil.replaceDollaSymbol(typeDecl.getCClass().getJavaName());
		
		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(AspectUtil.replaceDollaSymbol(currentJavaDeclType.getFullyQualifiedName()).equals(classQualifiedName)){
				if(typeDecl.getTokenReference().file().getAbsolutePath().equals(currentJavaDeclType.getSource().getFile().getAbsolutePath())){
					if(QDoxUtil.isTypeDeclOpenSpecOrSpecsOnly(currentJavaDeclType)){
						result = true;		
						break;
					}
				}
			}
		}
		return result;
	}
	
//	public boolean isTypeDeclPartialOrOpen(JmlTypeDeclaration typeDecl){
//		boolean result = false;
//		JavaClass [] javaDeclTypes  = this.getAllJavaTypeDeclarationsAsArray();
//		String classQualifiedName = AspectUtil.replaceDollaSymbol(typeDecl.getCClass().getJavaName());
//		
//		for (int i = 0; i < javaDeclTypes.length; i++) {
//			JavaClass currentJavaDeclType = javaDeclTypes[i];
////			JavaClass currentJavaDeclOwnerTopType = QDoxUtil.getTopOfTopOwner(currentJavaDeclType);
//			if(AspectUtil.replaceDollaSymbol(currentJavaDeclType.getFullyQualifiedName()).equals(classQualifiedName)){
//				if(typeDecl.getTokenReference().file().getAbsolutePath().equals(currentJavaDeclType.getSource().getFile().getAbsolutePath())){
//					if(QDoxUtil.isTypeDeclPartialOrOpen(currentJavaDeclType)){
//						result = true;		
//						break;
//					}
//				}
//			}
//		}
//		return result;
//	}
	
	public String extractPCValueForDeclWarningOrErrorPC(String pcExpr, JmlTypeDeclaration typeDecl) {
		if(!(pcExpr.contains("("))){
			for (Iterator iterator = AspectUtil.getInstance().getFieldDecls().iterator(); iterator.hasNext();) {
				JavaField javaField = (JavaField) iterator.next();
				if(pcExpr.contains(".")){
					String field = javaField.getParentClass().getFullyQualifiedName().replace("$", ".")+"."+javaField.getName();
					if(field.equals(pcExpr)){
						pcExpr = javaField.getInitializationExpression().replace("\"", "");
					}
				}
				else{
					if(javaField.getName().equals(pcExpr)){
						pcExpr = javaField.getInitializationExpression().replace("\"", "");
					}
				}
			}
		}
		else {
			// means that it is defined by a pointcut declaration
			List javaMethXCS = AspectUtil.getInstance().getCrosscutSpecMeths();
			for (Iterator iterator = javaMethXCS.iterator(); iterator
					.hasNext();) {
				JavaMethod currentJavaMethod = (JavaMethod) iterator.next();
				Pattern thisPattern = Pattern.compile("\\.\\b"+Pattern.quote(currentJavaMethod.getName())+"\\b");
				Matcher thisMatcher = thisPattern.matcher(pcExpr);
				if(!thisMatcher.find()){
					Pattern thisPatternAux = Pattern.compile("\\b"+Pattern.quote(currentJavaMethod.getName())+"\\b");
					Matcher thisMatcherAux = thisPatternAux.matcher(pcExpr);
					if(thisMatcherAux.find()){
						//testing if it is local
						if(currentJavaMethod.getParentClass().getFullyQualifiedName().equals(typeDecl.getCClass().getJavaName())){
				        	for (int i = 0; i < currentJavaMethod.getAnnotations().length; i++) {
				        		Annotation currentAnno = currentJavaMethod.getAnnotations()[i];
				        		if((currentAnno.toString().startsWith("@org.aspectj.lang.annotation.Pointcut"))){
				        			pcExpr = pcExpr.replace(thisMatcherAux.group(), typeDecl.getCClass().getJavaName()+"."+thisMatcherAux.group());
//				        			pcAnno = currentAnno;
				        			break;
				        		}
				        	}
						}
						else {
							ArrayList allOverriddenMethodsFromSuperClasses = new ArrayList();
							typeDecl.getCClass().getAllMethods(allOverriddenMethodsFromSuperClasses);
							for (Iterator iterator2 = allOverriddenMethodsFromSuperClasses
									.iterator(); iterator2.hasNext();) {
								CMethod currentInheritedMeth = (CMethod) iterator2.next();
								if(currentJavaMethod.getParentClass().getFullyQualifiedName().equals(currentInheritedMeth.owner().getJavaName())){
						        	for (int i = 0; i < currentJavaMethod.getAnnotations().length; i++) {
						        		Annotation currentAnno = currentJavaMethod.getAnnotations()[i];
						        		if((currentAnno.toString().startsWith("@org.aspectj.lang.annotation.Pointcut"))){
						        			pcExpr = pcExpr.replace(thisMatcherAux.group(), currentInheritedMeth.owner().getJavaName()+"."+thisMatcherAux.group());
//						        			pcAnno = currentAnno;
						        			break;
						        		}
						        	}
								}
							}
						}
					}
				}
			}
		}
		return pcExpr;
	}
	
	public boolean isOldVarReferencedWithinPrecondition(HashMap preconditions, String oldVarIdent){
		boolean result = false;
		
		for (Iterator iterator = preconditions.keySet().iterator(); iterator.hasNext();) {
			int index = (int) iterator.next();
			String precondition = (String)preconditions.get(index);
			if(precondition.contains(oldVarIdent)){
				result = true;
				break;
			}
		}
		
		return result;
	}
	
	public boolean isOldVarReferencedWithinPreExpr(List prexpList, String oldVarIdent){
		boolean result = false;
		
		for (Iterator iterator = prexpList.iterator(); iterator.hasNext();) {
			String prexp = (String)iterator.next();
			if(prexp.contains(oldVarIdent)){
				result = true;
				break;
			}
		}
		
		return result;
	}
	
	public boolean hasElementsStoredOldExpressions(HashMap oldExprs){
		boolean result = false;
		for (Iterator iterator = oldExprs.keySet().iterator(); iterator.hasNext();) {
			int index = (int) iterator.next();
			List currentOldList = (List) oldExprs.get(index);
			if(currentOldList != null){
				if(currentOldList.size() > 0){
					result = true;
					break;
				}
			}
			
		}
		return result;
	}
	
	public boolean hasBehavioralSpec(JmlMethodDeclaration methodDecl){
		boolean hasBehavioralKindSpec = false;
		if(methodDecl.methodSpecification().hasSpecCases()){
			JmlSpecCase[] specCases = methodDecl.methodSpecification().specCases();
			for (int i = 0; i < specCases.length; i++) {
				if(specCases[i] instanceof JmlBehaviorSpec){
					hasBehavioralKindSpec = true;
				}
				else if(specCases[i] instanceof JmlNormalBehaviorSpec){
					hasBehavioralKindSpec = true;
				}
				else if(specCases[i] instanceof JmlExceptionalBehaviorSpec){
					hasBehavioralKindSpec = true;
				}
			}
		}
		return hasBehavioralKindSpec;
	}
}
