/*
 * Copyright (C) 2008-2013 Universidade Federal de Pernambuco and 
 * University of Central Florida
 *
 * This file is part of AJML
 *
 * AJML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * AJML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AJML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: QDoxUtil.java,v 1.0 2013/05/17 13:00:38 henriquerebelo Exp $
 */

package org.aspectjml.util;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Stack;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;
import org.aspectjml.ajmlrac.Main;
import org.aspectjml.checker.JmlCommonOptions;
import org.aspectjml.checker.JmlTypeDeclaration;
import org.multijava.mjc.MjcCommonOptions;
import org.multijava.util.compiler.PositionedError;

import com.thoughtworks.qdox.JavaDocBuilder;
import com.thoughtworks.qdox.model.Annotation;
import com.thoughtworks.qdox.model.DocletTag;
import com.thoughtworks.qdox.model.JavaClass;
import com.thoughtworks.qdox.model.JavaField;
import com.thoughtworks.qdox.model.JavaMethod;
import com.thoughtworks.qdox.model.TypeVariable;

public class QDoxUtil {

	public static String getTypeErasureForTypeDeclsInFile(File file, 
			boolean isGenericSource, MjcCommonOptions options) throws FileNotFoundException,  IOException, PositionedError {
		BufferedReader buffer = null;
		StringBuffer bufferedFile = null;
		String fileAsString = "";
		String line = "";
		
		// issue to parse default methods in interfaces in Java 8 --- [[[hemr]]]
		boolean java8OrLater = Float.parseFloat(((JmlCommonOptions) options).source()) >= 1.8f;

		bufferedFile = new StringBuffer("");
		buffer = new BufferedReader(new FileReader( file ));
		line = buffer.readLine();
		line = QDoxUtil.getMapEntryImport(java8OrLater, line);
		while (line != null) {
			bufferedFile.append(QDoxUtil.getMapEntryImport(java8OrLater, line));
			bufferedFile.append("\n");
			line = buffer.readLine();
		}
		buffer.close();
		fileAsString = bufferedFile.toString();
		fileAsString = fileAsString.replace("{}", "{ }");
		if(fileAsString.contains("$CPSurrogate$AJML File")){
			return fileAsString;
		}

//		handling issues about default methods --- [[[hemr]]]
		if(java8OrLater){
			if(fileAsString.contains("default")){
				Pattern pattern = Pattern.compile("(\\b)default(\\b)(.)+\\((.)*\\)(.)*((\\{)|;)");
				Matcher matcher = pattern.matcher(fileAsString);
				while(matcher.find()){
					Pattern pattern2 = Pattern.compile("(\\b)default(\\b)(\\s)*\\:");
					Matcher matcher2 = pattern2.matcher(matcher.group());
					if(matcher2.find()){
						continue;
					}
					if(matcher.group().endsWith(";") && 
							!matcher.group().contains("\"")){
						fileAsString = fileAsString.replace(matcher.group(), matcher.group().replace("default", "").replace(";", "{/*##defaultmethod##*/}"));
					}
					else{
						fileAsString = fileAsString.replace(matcher.group(), matcher.group().replace("default", ""));
					}	
				}
			}
		}
		
//        if(((JmlCommonOptions) options).javadocBasedJML()){
        	// for multi-line predicates using nested @ that are non-top level --- [[[hemr]]]
//    		fileAsString = fileAsString.replaceAll("@old(\\b)", "#old").replaceAll("@forall(\\b)", "#forall").replaceAll("@exists(\\b)", "#exists");
//        }
		AspectUtil.getInstance().addJavaFileAsString(fileAsString, file);
		JavaDocBuilder qDoxFile = new JavaDocBuilder();
		qDoxFile.addSource(new StringReader(fileAsString));
		qDoxFile.getSources()[qDoxFile.getSources().length-1].setFile(file);
		// renaming types with appropriate package names if needed --- [[[hemr]]]
		// this happens when using multiple types with different package names 
		QDoxUtil.processJavaTypesWithRightFullQualifiedNames(qDoxFile);
		
		// handling JavaDocTags in File
//		this should be sure at some point whether should be enabled or not... [[[hemr]]]
//		if(((JmlCommonOptions) options).javadocBasedJML()){
			List<DocletTag> javaMethsWithDocletTagsFile = QDoxUtil.getAllJavaDocTagsInFile(qDoxFile);
			if(javaMethsWithDocletTagsFile.size() > 0){
				if(fileAsString.contains("@old") || fileAsString.contains("@forall") || fileAsString.contains("@exists")){
					String fileAsStringProcessed = fileAsString;
					if(fileAsString.contains("@old")){
						fileAsStringProcessed = fileAsStringProcessed.replaceAll("@old(\\b)", "#old");
					}
					if(fileAsString.contains("@forall")){
						fileAsStringProcessed = fileAsStringProcessed.replaceAll("@forall(\\b)", "#forall");
					}
					if(fileAsString.contains("@exists")){
						fileAsStringProcessed = fileAsStringProcessed.replaceAll("@exists(\\b)", "#exists");
					}
					buffer = new BufferedReader(new StringReader(fileAsStringProcessed));
				}
				else{
					buffer = new BufferedReader(new StringReader(fileAsString));
				}

				StringBuffer fileAsStringJavaDocProcessed = new StringBuffer("");
				line = buffer.readLine();
				int fileLineNumber = 1;
				while (line != null) {
					// if any
					String jmlClause = QDoxUtil.getJavaDocTagAsJMLClause(javaMethsWithDocletTagsFile, fileLineNumber, options);
					fileAsStringJavaDocProcessed.append(line).append(jmlClause);
					fileAsStringJavaDocProcessed.append("\n");
					line = buffer.readLine();
					fileLineNumber++;
				}
				buffer.close();
				fileAsString = StringUtils.replaceOnce(fileAsString, fileAsString, fileAsStringJavaDocProcessed.toString());
			}

			// handling javadoc tags in Java types that should be shifted
			List<JavaClass> javaDeclTypeWithJavadocTags = QDoxUtil.getAllDeclaredJavaTypesWithJavaDocTagsInFile(qDoxFile);
			for (Iterator<JavaClass> iterator = javaDeclTypeWithJavadocTags.iterator(); iterator
					.hasNext();) {
				JavaClass javaTypeWithJavadoc = iterator.next();
				String jmlClausesToShift = QDoxUtil.getJavaDocTagAsJMLClauseForTypeToShift(QDoxUtil.getAllJavaDocTagsInAJavaTypedDecl(javaTypeWithJavadoc));
				if(jmlClausesToShift.equals("")){
					continue;
				}
				buffer = new BufferedReader(new StringReader(fileAsString));
				int lineStart = javaTypeWithJavadoc.getLineNumber();
				int fileLineNumber = 1;
				StringBuffer TypeDeclUtilOpenBrace = new StringBuffer("");
				line = buffer.readLine();
				fileLineNumber = 1;
				while (line != null) {
					if(fileLineNumber >= lineStart){
						if(line.contains("{")){
							int indexOpenBrace = line.indexOf('{');
							String lineTmp = line.substring(0, (indexOpenBrace+1));
							TypeDeclUtilOpenBrace.append(lineTmp);
							break;
						}
						else{
							TypeDeclUtilOpenBrace.append(line);
							TypeDeclUtilOpenBrace.append("\n");
						}
					}
					line = buffer.readLine();
					fileLineNumber++;
				}
				buffer.close();
				String TypeDeclUtilOpenBraceStr = TypeDeclUtilOpenBrace.toString().trim();
				// processing java field tags
				// placing them where the JML compiler can understand - [[[hemr]]]
				fileAsString = StringUtils.replaceOnce(fileAsString, TypeDeclUtilOpenBraceStr, TypeDeclUtilOpenBraceStr+jmlClausesToShift);
			}// end for

			// handling javadoc tags in Java fields that should be shifted
			List<JavaField> javaDeclFieldsWithJavadocTags = QDoxUtil.getAllDeclaredJavaFieldsWithJavaDocTagsInFile(qDoxFile);
			for (Iterator<JavaField> iterator = javaDeclFieldsWithJavadocTags.iterator(); iterator
					.hasNext();) {
				JavaField javaFieldWithJavadoc = iterator.next();
				String jmlClausesToShift = QDoxUtil.getJavaDocTagAsJMLClauseForFieldToShift(QDoxUtil.getAllJavaDocTagsInAJavaFieldDecl(javaFieldWithJavadoc));
				if(jmlClausesToShift.equals("")){
					continue;
				}
				buffer = new BufferedReader(new StringReader(fileAsString));
				int lineStart = javaFieldWithJavadoc.getLineNumber();
				int fileLineNumber = 1;
				StringBuffer fieldDecl = new StringBuffer("");
				line = buffer.readLine();
				fileLineNumber = 1;
				while (line != null) {
					if(fileLineNumber >= lineStart){
						if(line.contains(";")){
							int indexSemiColon = line.lastIndexOf(';');
							fieldDecl.append(line.substring(0, indexSemiColon+1));
							break;
						}
						else{
							fieldDecl.append(line);
							fieldDecl.append("\n");
						}
					}
					line = buffer.readLine();
					fileLineNumber++;
				}
				buffer.close();
				String fieldDeclStr = fieldDecl.toString().trim();
				// processing java field tags
				// placing them where the JML compiler can understand - [[[hemr]]]
				fileAsString = StringUtils.replaceOnce(fileAsString, fieldDeclStr, fieldDeclStr+jmlClausesToShift);
			}// end for
//		}
		
		// Generic Source or any Java 5+ features (hopefully :-)
		if(isGenericSource){
			String actualFileAsString = fileAsString;
			bufferedFile.delete(0, (bufferedFile.length()-1)); // reset for later use

			// handling enum types
			List<JavaClass> javaDeclEnumTypes = QDoxUtil.getAllDeclaredJavaEnumTypesInFile(qDoxFile);
			if(javaDeclEnumTypes.size() > 0){
				fileAsString = QDoxUtil.getFileEnumTypeErasureProcessingAsString(bufferedFile, actualFileAsString, javaDeclEnumTypes);
			}
			
			// collecting all methods that lexically occur within a file
			bufferedFile.delete(0, (bufferedFile.length()-1)); // reset for later use
			List<JavaMethod> javaDeclMeths = QDoxUtil.getAllDeclaredJavaMethodsInFile(qDoxFile);
			List<JavaMethod> adviceDeclJavaFile = null;
			List<JavaMethod> supplyMethJavaFile = null;
			
			if(file.getCanonicalPath().endsWith(".jml") || file.getCanonicalPath().endsWith(".ajml")){
				actualFileAsString = QDoxUtil.handleConstructorDeclInJMLFileMode(bufferedFile, actualFileAsString, javaDeclMeths);
			}
			
			if(file.getCanonicalPath().endsWith(".jml")){
				QDoxUtil.getAllDeclaredJavaFieldsInJMLFile(qDoxFile);
			}
			
			if(file.getCanonicalPath().endsWith(".ajml")){
				if(Main.aRacOptions != null && !Main.aRacOptions.nocrosscutting()){
					// inter-type declarations are really processed in practice when declared as module slices in a .ajml file.
					QDoxUtil.getAllDeclaredAnnotatedITDJavaFieldsInFile(qDoxFile);
					QDoxUtil.getAllDeclaredAnnotatedITDJavaMethodsInFile(qDoxFile);
					QDoxUtil.getAllDeclaredAnnotatedITDJavaClassesInFile(qDoxFile);

					// adding supplying methods --- [[[hemr]]]
					supplyMethJavaFile = QDoxUtil.getAllDeclaredAnnotatedSuppliedJavaMethodsInFile(qDoxFile);
					for (Iterator iterator = supplyMethJavaFile.iterator(); iterator
							.hasNext();) {
						JavaMethod supplyMethod = (JavaMethod) iterator.next();
						AspectUtil.getInstance().appendSupplyMethDecl(supplyMethod);
					}
				}
			}
			// collecting advising code in java-like method advice --- [[[hemr]]]
			if(Main.aRacOptions != null && !Main.aRacOptions.nocrosscutting()){
				adviceDeclJavaFile = QDoxUtil.getAllDeclaredAnnotatedAdviceJavaMethodsInFile(qDoxFile, 
						(file.getCanonicalPath().endsWith(".ajml")));
				for (Iterator iterator = adviceDeclJavaFile.iterator(); iterator
						.hasNext();) {
					JavaMethod adviceDecl = (JavaMethod) iterator.next();
					AspectUtil.getInstance().appendAdviceDeclJavaFile(adviceDecl);
				}
				
				// adding annotated members for supplying annotations in AspectJ-style --- [[[hemr]]]
				AspectUtil.getInstance().appendFieldDecls(getAllDeclaredJavaFieldsInFile(qDoxFile));
				// adding annotated fields for @field
				List<JavaField> annotatedJavaFields = getAllDeclaredAnnotatedJavaFieldsInFile(qDoxFile);
				for (Iterator iterator = annotatedJavaFields.iterator(); iterator
						.hasNext();) {
					JavaField javaField = (JavaField) iterator.next();
					AspectUtil.getInstance().appendAnnotatedJavaFieldDecl(javaField);
				}

				// adding annotated methods for @methods
				List<JavaMethod> annotatedJavaMethods = getAllDeclaredAnnotatedJavaMethodsInFile(qDoxFile);
				for (Iterator iterator = annotatedJavaMethods.iterator(); iterator
						.hasNext();) {
					JavaMethod javaMethod = (JavaMethod) iterator.next();
					AspectUtil.getInstance().appendAnnotatedJavaMethodDecl(javaMethod);
				}

				// adding annotated types for @type
				List<JavaClass> annotatedTypes = getAllDeclaredAnnotatedJavaTypesInFile(qDoxFile);
				for (Iterator iterator = annotatedTypes.iterator(); iterator
						.hasNext();) {
					JavaClass javaClass = (JavaClass) iterator.next();
					AspectUtil.getInstance().appendAnnotatedJavaTypeDecl(javaClass);
				}
				
				// adding the annotatedDeclareSoftMethods
				List<JavaMethod> javaDeclMethDeclareSoft = getAllDeclaredAnnotatedDeclareSoftJavaMethodsInFile(qDoxFile);
				for (Iterator iterator = javaDeclMethDeclareSoft.iterator(); iterator
						.hasNext();) {
					JavaMethod javaMethod = (JavaMethod) iterator.next();
					AspectUtil.getInstance().appendDeclareSoftMethodDecl(javaMethod);
				}
			}

			List<com.github.antlrjavaparser.api.body.BodyDeclaration> members = QDoxUtil.getAllDeclaredJavaMethodsInFile(actualFileAsString);
			List<String> fileMeths = QDoxUtil.getAllJavaMethodDeclLexicallyInFile(bufferedFile, actualFileAsString, javaDeclMeths, members);
			
			if(fileMeths.size() != javaDeclMeths.size()){
				System.err.println("file = "+file.getCanonicalPath());
				System.err.println("processed ---> "+fileMeths.size());
				System.err.println("really contains = "+javaDeclMeths.size());
			}
			
//			for (Iterator<String> iterator = fileMeths.iterator(); iterator.hasNext();) {
//				String currentMeth = iterator.next();
//				System.out.println("matchedMeth = "+currentMeth);
//			}
			
			fileAsString = QDoxUtil.stripMethBodies(fileAsString, javaDeclMeths, fileMeths); // method bodies stripped... [[[hemr]]]
			
//			eliminating the pattern --> default {return null;};
			if(fileAsString.contains("default") && fileAsString.contains("return")){
				fileAsString = fileAsString.replaceAll("(\\b)default(\\b)(\\s)*\\{(\\s)*return [\\w;]+(\\s)*\\};", ";");
			}
			
			// handling annotated Java fields
			String fileAsStringProcessed = fileAsString;
			List<JavaField> javaDeclAnnotatedFields = QDoxUtil.getAllDeclaredAnnotatedJavaFieldsInFile(qDoxFile);
			List<com.github.antlrjavaparser.api.body.FieldDeclaration> javaDeclAnnotatedFields2 = QDoxUtil.getAllDeclaredAnnotatedJavaFieldsInFile(actualFileAsString);
			try {
				for (int i = 0; i < javaDeclAnnotatedFields.size(); i++) {
					JavaField annotatedJavaField = javaDeclAnnotatedFields.get(i);
					com.github.antlrjavaparser.api.body.FieldDeclaration annotatedJavaField2 = javaDeclAnnotatedFields2.get(i);
					
					buffer = new BufferedReader(new StringReader(fileAsStringProcessed));
					StringBuffer annotationArea = new StringBuffer("");
					StringBuffer annotationAreaCommented = new StringBuffer("");
					int lineStart = annotatedJavaField2.getAnnotations().get(0).getBeginLine();
					int lineEnd = annotatedJavaField2.getAnnotations().get(annotatedJavaField2.getAnnotations().size() - 1).getEndLine();
					line = buffer.readLine();
					int fileLineNumber = 1;
					while (line != null) {
						if(fileLineNumber >= lineStart){
							if(fileLineNumber == lineEnd){
								// checking if annotation and field declaration appears in the same line
								Pattern fieldNamePattern = Pattern.compile("(\\b)"+annotatedJavaField.getName()+"(\\b)");
								Matcher fieldNameMatcher = fieldNamePattern.matcher(line);
								if(fieldNameMatcher.find()){
									char [] lineChars = line.toCharArray();
									StringBuffer annotationSB = new StringBuffer("");
									StringBuffer fieldDeclSB = new StringBuffer("");
									boolean findAnnotation = false;
									int annoQtd = QDoxUtil.getAnnotationDeclPerLine(line);
									
									for (int j = 0; j < lineChars.length; j++) {
										if(QDoxUtil.getAnnotationDeclPerLine(annotationSB.toString()) == annoQtd){
											if(findAnnotation){
												fieldDeclSB.append(lineChars[j]);
											}
											if(!Character.isWhitespace(lineChars[j]) && !findAnnotation){
												annotationSB.append(lineChars[j]);
											}
											if(annotationSB.toString().contains("(")){
												if(QDoxUtil.balancedParenthensies(annotationSB.toString())){
													findAnnotation = true;
												}
											}
											else{
												if(Character.isWhitespace(lineChars[j]) && !findAnnotation){
													findAnnotation = true;
												}
											}
										}
										else{
											annotationSB.append(lineChars[j]);
										}
							    	}
								    annotationArea.append(line);
								    String lineSurrogate = "/*"+fileLineNumber+"*/"+"/* "+annotationSB.toString().replace("@", "#")+"*/"+fieldDeclSB.toString();
								    String fieldDeclAnnotations = QDoxUtil.getFieldDeclAnnotationsAsJML(annotationArea, annotatedJavaField);
								    String fieldInProcessedAnno = QDoxUtil.getFieldJMLInAnnotationsProcessedAsStandardJML(
											annotatedJavaField);
								    lineSurrogate = fieldDeclAnnotations+lineSurrogate+fieldInProcessedAnno;		
								    fileAsStringProcessed = changeFileLine(fileAsStringProcessed, fileLineNumber,lineSurrogate);
									break;
								}
								else{
									annotationArea.append(line);
									String lineSurrogate = "/*"+fileLineNumber+"*/"+"/* "+line.replace("@", "#")+"*/";
									 String fieldDeclAnnotations = QDoxUtil.getFieldDeclAnnotationsAsJML(annotationArea, annotatedJavaField);
									    String fieldInProcessedAnno = QDoxUtil.getFieldJMLInAnnotationsProcessedAsStandardJML(
												annotatedJavaField);
									    lineSurrogate = fieldDeclAnnotations+lineSurrogate;	
									fileAsStringProcessed = changeFileLine(fileAsStringProcessed, fileLineNumber, lineSurrogate);
									fileAsStringProcessed = appendFileLine(fileAsStringProcessed, lineEnd+1, fieldInProcessedAnno);
									break;
								}
							}
							else{
								annotationArea.append(line);
								annotationArea.append("\n");
								annotationAreaCommented.append("/*"+fileLineNumber+"*/"+"/* "+line.replace("@", "#")+"*/");
								fileAsStringProcessed = changeFileLine(fileAsStringProcessed, fileLineNumber, "/*"+fileLineNumber+"*/"+"/* "+line.replace("@", "#")+"*/");
								annotationAreaCommented.append("\n");
							}
						}				
						line = buffer.readLine(); // next line
						fileLineNumber++;
					} // end while
					buffer.close();
				}// end for
			} catch (Exception e) {}
			// adding the processed annotated fields
			fileAsString = fileAsStringProcessed;
			
			// handling generics in Java fields
			List<JavaField> javaDeclFields = QDoxUtil.getAllDeclaredJavaFieldsInFile(qDoxFile);
			List<com.github.antlrjavaparser.api.body.FieldDeclaration> javaDeclFields2 = QDoxUtil.getAllDeclaredJavaFieldsInFile(actualFileAsString);
			for (int i = 0; i < javaDeclFields.size(); i++) {
				JavaField javaField = javaDeclFields.get(i);
				com.github.antlrjavaparser.api.body.FieldDeclaration javaField2 = null;
				try {
					javaField2 = javaDeclFields2.get(i);
				} catch (java.lang.IndexOutOfBoundsException e) {
					continue; // igore the generic field processing... [[[hemr]]]
				}
				
				buffer = new BufferedReader(new StringReader(fileAsString));
				StringBuffer fieldDeclArea = new StringBuffer("");
				int lineStart = javaField2.getBeginLine();
				int lineEnd = javaField2.getEndLine();
				line = buffer.readLine();
				int fileLineNumber = 1;
				while (line != null) {
					if(fileLineNumber >= lineStart){
						if(fileLineNumber == lineEnd){
							fieldDeclArea.append(line);
							break;
						}
						else{
							fieldDeclArea.append(line);
							fieldDeclArea.append("\n");
						}
					}				
					line = buffer.readLine(); // next line
					fileLineNumber++;
				} // end while
				buffer.close();
				// handling ITD fields to have their orignalDecl resolved --- [[[hemr]]]
			    QDoxUtil.resolveOriginalITDFieldDeclWithNameProcessed(javaField, fieldDeclArea.toString());
				//handling generics part
				HashMap<String, String> genericAndErasureTypeField = QDoxUtil.getGenericAndErasureTypeParametersForField(javaField.getParentClass(), javaField);
				Set<String> genericTypes = genericAndErasureTypeField.keySet();
				String fieldDeclAreaStr = fieldDeclArea.toString();
				for (Iterator<String> iterator2 = genericTypes.iterator(); iterator2
						.hasNext();) {
					String genericType = iterator2.next();
					String erasureType = genericAndErasureTypeField.get(genericType);
//					System.out.println("genericType[i] = "+genericType);
//					System.out.println("erasureType[i] = "+erasureType);
//					System.out.println("fieldDeclAreaStr = "+fieldDeclAreaStr);
					fieldDeclAreaStr = fieldDeclAreaStr.replaceAll("(\\b)"+genericType+"(\\b)", erasureType);
				}
//				System.out.println("fieldDeclArea.toString() = "+fieldDeclArea.toString());
//				System.out.println("fieldDeclAreaStr = "+fieldDeclAreaStr);
//				if(javaField.getParentClass().getTypeParameters().length == 0){
					fileAsString = StringUtils.replaceOnce(fileAsString, fieldDeclArea.toString(), fieldDeclAreaStr);	
//				}
					
				
				// handling the diamond part in fields
				String fieldDeclStrDiamond = "";
//				String fieldDeclStrDiamond = fieldDeclAreaStr;
//				String fieldDeclStrDiamond = fieldDeclArea.toString();
				fieldDeclStrDiamond = fieldDeclAreaStr;
//				if(javaField.getParentClass().getTypeParameters().length == 0){
//					fieldDeclStrDiamond = fieldDeclAreaStr;
//				}
//				else{
//					fieldDeclStrDiamond = fieldDeclArea.toString();
//				}
				
				
//				avoiding problem with diamonds within quotes --- [[[hemr]]]
//				fieldDeclStrDiamond = fieldDeclStrDiamond.replaceAll("\"(.)*[\\<\\>]+(.)*\"", "\"\""); 
//				fieldDeclStrDiamond = fieldDeclStrDiamond.replaceAll("\'[\\<\\>]?\'", "\'\'"); 
				Pattern fieldIssuePattern = Pattern.compile("\"(.)*[\\<\\>]+(.)*\"");
				Matcher fieldIssueMatcher = fieldIssuePattern.matcher(fieldDeclStrDiamond);
				while(fieldIssueMatcher.find()){
					fieldDeclStrDiamond = fieldDeclStrDiamond.replace(fieldIssueMatcher.group(), "\"\"");
				}
				fieldIssueMatcher.reset();
				fieldIssuePattern = Pattern.compile("\'[\\<\\>]?\'");
				fieldIssueMatcher = fieldIssuePattern.matcher(fieldDeclStrDiamond);
				while(fieldIssueMatcher.find()){
					if(!fieldIssueMatcher.group().equals("''")){
						fieldDeclStrDiamond = fieldDeclStrDiamond.replace(fieldIssueMatcher.group(), "' '");
					}
				}
				
				while(fieldDeclStrDiamond.contains("<") && fieldDeclStrDiamond.contains(">")){
					char [] currentFieldChars = fieldDeclStrDiamond.toCharArray();
					StringBuffer currentFieldSb = new StringBuffer(""); 
					boolean canAdd = false;
					for (int j = 0; j < currentFieldChars.length; j++) {
						if(currentFieldChars[j] == '<'){
							canAdd = true;
						}
						if(canAdd){
							currentFieldSb.append(currentFieldChars[j]);
						}
						if(currentFieldSb.length() > 0 && balancedDiamond(currentFieldSb.toString())){
							break;
						}
					}
					// checking the dot/genercis part in field initialization
					int newLineCount = QDoxUtil.getLineNumbersQtd(currentFieldSb.toString());
					String lines = QDoxUtil.getNewLinesCaracter(newLineCount);
//					if(fieldDeclStrDiamond.contains("."+currentFieldSb.toString()+" ")){
//						fieldDeclStrDiamond = fieldDeclStrDiamond.replaceFirst("\\."+currentFieldSb.toString()+"[\\W]*", "");
//					}
					if(fieldDeclStrDiamond.contains("."+currentFieldSb.toString())){
						fieldDeclStrDiamond = fieldDeclStrDiamond.replaceFirst("."+currentFieldSb.toString()+"[\\W]*", "");
					}
					else{
						fieldDeclStrDiamond = fieldDeclStrDiamond.replace(currentFieldSb.toString(), "");
					}
				}
				// updating file with removed diamnod in fields
				fileAsString = StringUtils.replaceOnce(fileAsString, fieldDeclAreaStr, fieldDeclStrDiamond);
//				if(javaField.getParentClass().getTypeParameters().length == 0){
//					fileAsString = StringUtils.replaceOnce(fileAsString, fieldDeclAreaStr, fieldDeclStrDiamond);
//				}
//				else{
//					fileAsString = StringUtils.replaceOnce(fileAsString, fieldDeclArea.toString(), fieldDeclStrDiamond);
//				}
			}
			
			// handling annotated Java Methods
			List<String> javaDeclMethsAnnotationArea = new ArrayList<String>();
			List<JavaMethod> javaDeclAnnotatedMeths = QDoxUtil.getAllDeclaredAnnotatedJavaMethodsInFile(qDoxFile);
			List<com.github.antlrjavaparser.api.body.BodyDeclaration> javaDeclAnnotatedMeths2 = QDoxUtil.getAllDeclaredAnnotatedJavaMethodsInFile(actualFileAsString);
			for (int i = 0; i < javaDeclAnnotatedMeths.size(); i++) {
				JavaMethod annotatedJavaMethod = javaDeclAnnotatedMeths.get(i);
				com.github.antlrjavaparser.api.body.BodyDeclaration annotatedJavaMethod2 = javaDeclAnnotatedMeths2.get(i);
				
				buffer = new BufferedReader(new StringReader(fileAsString));
				StringBuffer annotationArea = new StringBuffer("");
				StringBuffer annotationAreaCommented = new StringBuffer("");
				int lineStart = annotatedJavaMethod2.getAnnotations().get(0).getBeginLine();
				int lineEnd = annotatedJavaMethod2.getAnnotations().get(annotatedJavaMethod2.getAnnotations().size() - 1).getEndLine();
				line = buffer.readLine();
				int fileLineNumber = 1;
				while (line != null) {
					if(fileLineNumber >= lineStart){
						if(fileLineNumber == lineEnd){
							annotationArea.append(line);
							annotationArea.append("\n");
							annotationAreaCommented.append("/*"+fileLineNumber+"*/"+"/* "+line+"*/");
							annotationAreaCommented.append("\n");
							break;
						}
						else{
							annotationArea.append(line);
							annotationArea.append("\n");
							annotationAreaCommented.append("/*"+fileLineNumber+"*/"+"/* "+line+"*/");
							annotationAreaCommented.append("\n");
						}
					}
					line = buffer.readLine();
					fileLineNumber++;
				} // end while
				// processing java method annotations
				String annotationAreaCommentedStr = QDoxUtil.getMethodAnnotationAreaCommentedProcessedWithJML(
						annotatedJavaMethod, annotationAreaCommented);
				
				// updating the current meth annotations in the file
				buffer = new BufferedReader(new StringReader(fileAsString));
				StringBuffer fileStartCurrentMeth = new StringBuffer("");
				
				lineStart = annotatedJavaMethod.getLineNumber();
				lineEnd = annotatedJavaMethod2.getEndLine();
				line = buffer.readLine();
				fileLineNumber = 1;
				while (line != null) {
					if(fileLineNumber >= lineStart){
						if(fileLineNumber == lineEnd){
							fileStartCurrentMeth.append(line);
							fileStartCurrentMeth.append("\n");
							break;
						}
						else{
							fileStartCurrentMeth.append(line);
							fileStartCurrentMeth.append("\n");
						}
						
					}
					line = buffer.readLine();
					fileLineNumber++;
				}
				String fileAsStringTmp = StringUtils.replaceOnce(fileStartCurrentMeth.toString(), annotationArea.toString(), annotationAreaCommentedStr.toString());
				fileAsString = StringUtils.replaceOnce(fileAsString, fileStartCurrentMeth.toString(), fileAsStringTmp);
				javaDeclMethsAnnotationArea.add(annotationAreaCommentedStr);
			}// end for
			
			// resolving ITD Java Types
			List<JavaClass> javaDeclTypes = QDoxUtil.getAllDeclaredJavaTypesInFile(qDoxFile);
			List<com.github.antlrjavaparser.api.body.TypeDeclaration> javaDeclTypes2 = QDoxUtil.getAllDeclaredJavaTypesInFile(actualFileAsString);
			for (int i = 0; i < javaDeclTypes.size(); i++) {
				JavaClass javaType = javaDeclTypes.get(i);
				com.github.antlrjavaparser.api.body.TypeDeclaration javaType2 = javaDeclTypes2.get(i);
				buffer = new BufferedReader(new StringReader(actualFileAsString));
				StringBuffer javaTypeArea = new StringBuffer("");
				int lineNumberStart = 0;
				int lineNumberEnd = 0;
				if(javaType2.getAnnotations()!= null && javaType2.getAnnotations().size() > 0){
					if((javaType2.getAnnotations().get(javaType2.getAnnotations().size()-1).getEndLine() + 1) < javaType2.getBeginLine()){
						lineNumberStart = javaType2.getBeginLine();
					}
					else{
						lineNumberStart = javaType2.getAnnotations().get(javaType2.getAnnotations().size()-1).getEndLine() + 1;
					}
					if(javaType2.getBeginComments() != null){
						lineNumberStart += javaType2.getBeginComments().size();
					}
					lineNumberEnd = javaType2.getEndLine();
				}
				else{
					lineNumberStart = javaType2.getBeginLine();
					lineNumberEnd = javaType2.getEndLine();
				}
				line = buffer.readLine();
				int fileLineNumber = 1;
				while (line != null) {
					if(fileLineNumber >= lineNumberStart){
						if(fileLineNumber == lineNumberEnd){
							javaTypeArea.append(line);
							javaTypeArea.append("\n");
							break;
						}
						else{
							javaTypeArea.append(line);
							javaTypeArea.append("\n");
						}
					}
					line = buffer.readLine();
					fileLineNumber++;
				} // end while
				QDoxUtil.resolveOriginalITDTypeDeclWithNameProcessed(javaType, javaTypeArea.toString());
			}

			// handling annotated Java Types
			List<JavaClass> javaDeclAnnotatedTypes = QDoxUtil.getAllDeclaredAnnotatedJavaTypesInFile(qDoxFile);
			List<com.github.antlrjavaparser.api.body.TypeDeclaration> javaDeclAnnotatedTypes2 = QDoxUtil.getAllDeclaredAnnotatedJavaTypesInFile(actualFileAsString);
			for (int i = 0; i < javaDeclAnnotatedTypes.size(); i++) {
				JavaClass annotatedJavaType = javaDeclAnnotatedTypes.get(i);
				com.github.antlrjavaparser.api.body.TypeDeclaration annotatedJavaType2 = javaDeclAnnotatedTypes2.get(i);
				buffer = new BufferedReader(new StringReader(fileAsString));
				StringBuffer annotationArea = new StringBuffer("");
				StringBuffer annotationAreaCommented = new StringBuffer("");
				int lineStart = annotatedJavaType2.getAnnotations().get(0).getBeginLine();
				int lineEnd = annotatedJavaType2.getAnnotations().get(annotatedJavaType2.getAnnotations().size() - 1).getEndLine();
//				System.out.println("lineStart = "+lineStart);
//				System.out.println("lineEnd = "+lineEnd);
				line = buffer.readLine();
				int fileLineNumber = 1;
				while (line != null) {
					if(fileLineNumber >= lineStart){
						if(fileLineNumber == lineEnd){
							annotationArea.append(line);
							annotationArea.append("\n");
							annotationAreaCommented.append("/*"+fileLineNumber+"*/"+"/* "+line.replace("@", "#")+"*/");
							annotationAreaCommented.append("\n");
							break;
						}
						else{
							annotationArea.append(line);
							annotationArea.append("\n");
							annotationAreaCommented.append("/*"+fileLineNumber+"*/"+"/* "+line.replace("@", "#")+"*/");
							annotationAreaCommented.append("\n");
						}
					}
					line = buffer.readLine();
					fileLineNumber++;
				} // end while
				
				// selecting the entire type decl
				buffer = new BufferedReader(new StringReader(fileAsString));
				StringBuffer typeDecl = new StringBuffer("");
                int newEnd = annotatedJavaType2.getEndLine();
				lineStart = lineEnd + 1;
				lineEnd = newEnd;
//				System.out.println("lineStart<new> = "+lineStart);
//				System.out.println("lineEnd<new> = "+lineEnd);
				line = buffer.readLine();
				fileLineNumber = 1;
				while (line != null) {
					if(fileLineNumber >= lineStart){
						if(fileLineNumber == lineEnd){
							typeDecl.append(line);
							typeDecl.append("\n");
						}
						else{
							typeDecl.append(line);
							typeDecl.append("\n");
						}
					}
					line = buffer.readLine();
					fileLineNumber++;
				}
				
				String typeJMLAnno = "";
				if(annotationArea.toString().contains("@NonNullByDefault")){
					typeJMLAnno = "/*@ non_null_by_default @*/";
				}
				else if(annotationArea.toString().contains("@NullableByDefault")){
					typeJMLAnno = "/*@ nullable_by_default @*/";		
				}
				
				if(annotationArea.toString().contains("@SpecPublic")){
					typeJMLAnno = "/*@ spec_public @*/" + typeJMLAnno;
				}
				else if(annotationArea.toString().contains("@SpecProtected")){
					typeJMLAnno = "/*@ spec_protected @*/" + typeJMLAnno;		
				}
				
				// processing java type annotations
				String annotationAreaCommentedStr = QDoxUtil.getTypeAnnotationAreaCommentedProcessedWithJML(
						annotatedJavaType, annotationAreaCommented.toString());
				
				// placing them where the JML compiler can understand - [[[hemr]]]
				String typeDeclStr = StringUtils.replaceOnce(typeDecl.toString(), "{", "{"+annotationAreaCommentedStr);
				
				// updating the current type annotations in the file
				// removing type annotations
				fileAsString = QDoxUtil.appendAtBeginFileLine(fileAsString, annotatedJavaType2.getAnnotations().get(0).getBeginLine(), "##$AJML$##-->");
				fileAsString = StringUtils.replaceOnce(fileAsString, "##$AJML$##-->"+annotationArea.toString(), typeJMLAnno);
				fileAsString = StringUtils.replaceOnce(fileAsString, typeDecl.toString(), typeDeclStr);
			}// end for
			
			// handling the diamond part in types
			List<com.github.antlrjavaparser.api.body.TypeDeclaration> javaTypeDecls = QDoxUtil.getAllDeclaredJavaTypesInFile(actualFileAsString);
			JavaDocBuilder qDoxJavaTypeParser = new JavaDocBuilder();
			for (Iterator iterator = javaTypeDecls.iterator(); iterator.hasNext();) {
				com.github.antlrjavaparser.api.body.TypeDeclaration javaTypeDecl = (com.github.antlrjavaparser.api.body.TypeDeclaration) iterator.next();
				buffer = new BufferedReader(new StringReader(fileAsString));
				StringBuffer javaTypeDeclArea = new StringBuffer("");
				int lineStart = javaTypeDecl.getBeginLine();
				line = buffer.readLine();
				int fileLineNumber = 1;
				while (line != null) {
					if(fileLineNumber >= lineStart){
						javaTypeDeclArea.append(line);
						try {
							qDoxJavaTypeParser.addSource(new StringReader("public class JMethodParser{\n"+javaTypeDeclArea.toString()+"\n}\n}"));
							break;
						}catch(Exception e){
							// not yet...parsed
							javaTypeDeclArea.append("\n");
						}
					}
					line = buffer.readLine();
					fileLineNumber++;
				} // end while
				
				String javaTypeDeclAreaStr = javaTypeDeclArea.toString();
				if(javaTypeDeclAreaStr.contains("@*/")){
					javaTypeDeclAreaStr = javaTypeDeclAreaStr.replaceAll("\\{(\\s)*/\\*@(.)+\\@*/", "{"); 
				}
				String typeDeclStrDiamond = javaTypeDeclAreaStr;
				if(typeDeclStrDiamond.contains("<") && typeDeclStrDiamond.contains(">")){
					while(typeDeclStrDiamond.contains("<") && typeDeclStrDiamond.contains(">")){
						char [] currentTypeChars = typeDeclStrDiamond.toCharArray();
						StringBuffer currentTypedSb = new StringBuffer(""); 
						boolean canAdd = false;
						for (int j = 0; j < currentTypeChars.length; j++) {
							if(currentTypeChars[j] == '<'){
								canAdd = true;
							}
							if(canAdd){
								currentTypedSb.append(currentTypeChars[j]);
							}
							if(currentTypedSb.length() > 0 && balancedDiamond(currentTypedSb.toString())){
								break;
							}
						}
						int newLineCount = QDoxUtil.getLineNumbersQtd(currentTypedSb.toString());
						String lines = QDoxUtil.getNewLinesCaracter(newLineCount);
						typeDeclStrDiamond = typeDeclStrDiamond.replace(currentTypedSb.toString(), ""+lines);
					}
					// updating file with removed diamnod in type
					fileAsString = StringUtils.replaceOnce(fileAsString, javaTypeDeclAreaStr, typeDeclStrDiamond);
				}
				
			}// end for
			
//			// collecting all methods that lexically occur within a file
			bufferedFile.delete(0, (bufferedFile.length()-1)); // reset for later use
			javaDeclMeths = QDoxUtil.getAllDeclaredJavaMethodsInFile(qDoxFile);
			fileMeths = QDoxUtil.getAllJavaMethodDeclLexicallyInFile(bufferedFile, fileAsString, javaDeclMeths, members);
//            System.out.println("fileMeths = "+fileMeths.size());
//			for (Iterator<String> iterator = fileMeths.iterator(); iterator.hasNext();) {
//				String currentMeth = iterator.next();
//				System.out.println("matchedMeth = "+currentMeth);
//			}
			
			fileAsString = QDoxUtil.getFileMethDeclsProcessed(fileAsString, javaDeclMeths, fileMeths, file);
			
			// final issues about JML type annotations
			Pattern jmlAnnoPattern = null;
			Matcher jmlAnnoMatcher = null;
			if(fileAsString.contains("Pure")){
				jmlAnnoPattern =	Pattern.compile("@(\\s)*Pure(\\b)(\\s)*(\\((\\s)*\\))?");
				jmlAnnoMatcher = jmlAnnoPattern.matcher(fileAsString);
				while(jmlAnnoMatcher.find()){
					int numberOfNewLines = QDoxUtil.getLineNumbersQtd(jmlAnnoMatcher.group());
					fileAsString = StringUtils.replaceOnce(fileAsString, jmlAnnoMatcher.group(), "/*@ pure @*/"+QDoxUtil.getNewLinesCaracter(numberOfNewLines));
				}
				jmlAnnoMatcher.reset();
			}
			
			if(fileAsString.contains("Helper")){
				jmlAnnoPattern = Pattern.compile("@(\\s)*Helper(\\b)(\\s)*(\\((\\s)*\\))?");
				jmlAnnoMatcher = jmlAnnoPattern.matcher(fileAsString);
				while(jmlAnnoMatcher.find()){
					int numberOfNewLines = QDoxUtil.getLineNumbersQtd(jmlAnnoMatcher.group());
					fileAsString = StringUtils.replaceOnce(fileAsString, jmlAnnoMatcher.group(), "/*@ helper @*/"+QDoxUtil.getNewLinesCaracter(numberOfNewLines));
				}
				jmlAnnoMatcher.reset();
			}
			
//			jmlAnnoPattern = Pattern.compile("@(\\s)*Nullable(\\b)(\\s)*(\\((\\s)*\\))?");
//			jmlAnnoMatcher = jmlAnnoPattern.matcher(fileAsString);
//			while(jmlAnnoMatcher.find()){
//				int numberOfNewLines = QDoxUtil.getLineNumbersQtd(jmlAnnoMatcher.group());
//				fileAsString = StringUtils.replaceOnce(fileAsString, jmlAnnoMatcher.group(), "/*@ nullable @*/"+QDoxUtil.getNewLinesCaracter(numberOfNewLines));
//			}
//			jmlAnnoMatcher.reset();
//			jmlAnnoPattern = Pattern.compile("@(\\s)*NonNull(\\b)(\\s)*(\\((\\s)*\\))?");
//			jmlAnnoMatcher = jmlAnnoPattern.matcher(fileAsString);
//			while(jmlAnnoMatcher.find()){
//				int numberOfNewLines = QDoxUtil.getLineNumbersQtd(jmlAnnoMatcher.group());
//				fileAsString = StringUtils.replaceOnce(fileAsString, jmlAnnoMatcher.group(), "/*@ non_null @*/"+QDoxUtil.getNewLinesCaracter(numberOfNewLines));
//			}
//			jmlAnnoMatcher.reset();
//			jmlAnnoPattern = Pattern.compile("@(\\s)*SpecPublic(\\b)(\\s)*(\\((\\s)*\\))?");
//			jmlAnnoMatcher = jmlAnnoPattern.matcher(fileAsString);
//			while(jmlAnnoMatcher.find()){
//				int numberOfNewLines = QDoxUtil.getLineNumbersQtd(jmlAnnoMatcher.group());
//				fileAsString = StringUtils.replaceOnce(fileAsString, jmlAnnoMatcher.group(), "/*@ spec_public @*/"+QDoxUtil.getNewLinesCaracter(numberOfNewLines));
//			}
//			jmlAnnoMatcher.reset();
//			jmlAnnoPattern = Pattern.compile("@(\\s)*SpecProtected(\\b)(\\s)*(\\((\\s)*\\))?");
//			jmlAnnoMatcher = jmlAnnoPattern.matcher(fileAsString);
//			while(jmlAnnoMatcher.find()){
//				int numberOfNewLines = QDoxUtil.getLineNumbersQtd(jmlAnnoMatcher.group());
//				fileAsString = StringUtils.replaceOnce(fileAsString, jmlAnnoMatcher.group(), "/*@ spec_protected @*/"+QDoxUtil.getNewLinesCaracter(numberOfNewLines));
//			}
//			jmlAnnoMatcher.reset();
			jmlAnnoPattern = Pattern.compile("[-+]?([0-9]+((\\.|\\_)[0-9]+)[FfLl]?)");
			jmlAnnoMatcher = jmlAnnoPattern.matcher(fileAsString);
			while(jmlAnnoMatcher.find()){
//				System.out.println(jmlAnnoMatcher.group());
				if(jmlAnnoMatcher.group().contains("_")){
					fileAsString = StringUtils.replaceOnce(fileAsString, jmlAnnoMatcher.group(), jmlAnnoMatcher.group().replace("_", ""));	
				}
			}

			fileAsString = fileAsString.replaceAll("/\\*(\\s)*/\\*@", "/*@");
			fileAsString = fileAsString.replaceAll("@\\*/(\\s)*\\*/", "@*/");
			// final issues due to annotation types support
			fileAsString = fileAsString.replace("@interface", "interface");
			
			Pattern defaultTextPattern = Pattern.compile("\"[\\w\"\\s+\\{\\}\\(\\)\\.]*(\\b)default(\\b)[\\@\\-\\w\"\\s+\\{\\}\\(\\)\\.]*\"");
			Matcher defaultTextMatcher = defaultTextPattern.matcher(fileAsString);
			while(defaultTextMatcher.find()){
				fileAsString = StringUtils.replaceOnce(fileAsString, defaultTextMatcher.group(), defaultTextMatcher.group().replace("default", ""));
			}
			
			// handling default stmt in annotation types
			fileAsString = fileAsString.replaceAll("(\\b)default(\\b)(\\s)*[\\@\\-\\w\"\\s+\\{\\}\\(\\)\\.]*(\\s)*;", ";");
			
			// handling static imports to support java 5 mode
			Pattern staticImportPattern = Pattern.compile("import static (.)*;");
			Matcher staticImportMatcher = staticImportPattern.matcher(fileAsString);
			while(staticImportMatcher.find()){
				String staticImport = staticImportMatcher.group();
				staticImport = staticImport.substring(0, staticImport.lastIndexOf(".")).replace("static ", "")+".*;";
				Pattern staticImportPattern2 = Pattern.compile("import(\\s)+"+staticImport.replace("import", "").trim());
				Matcher staticImportMatcher2 = staticImportPattern2.matcher(fileAsString);
				if(!staticImportMatcher2.find()){
					fileAsString = StringUtils.replaceOnce(fileAsString, staticImportMatcher.group(), staticImport);	
				}
				else{
					fileAsString = StringUtils.replaceOnce(fileAsString, staticImportMatcher.group(), "");	
				}
			}
			
			// handling standard Java 5 annotations
			if(fileAsString.contains("@Inherited")){
				fileAsString = fileAsString.replaceAll("@Inherited((\\s)*\\((\\s)*\\))?", "");
			}
			
			if(fileAsString.contains("@Override")){
				fileAsString = fileAsString.replaceAll("@Override((\\s)*\\((\\s)*\\))?", "");
			}
			
			if(fileAsString.contains("@Deprecated")){
				fileAsString = fileAsString.replaceAll("@Deprecated((\\s)*\\((\\s)*\\))?", "");
			}
			
			if(fileAsString.contains("@SuppressWarnings")){
				fileAsString = fileAsString.replaceAll("@SuppressWarnings((\\s)*\\(\"(.)*\"\\))?", "");
				fileAsString = fileAsString.replaceAll("@SuppressWarnings((\\s)*\\((\\s)*\\))?", "");
			}
			
			// handling dot dot dot in Java 5
			if(fileAsString.contains("...")){
				fileAsString = fileAsString.replace("...", "[]");
			}
			
			if(fileAsString.contains("<?>")){
				fileAsString = fileAsString.replace("<?>", "");
			}

			if(java8OrLater){
				if(fileAsString.contains("->")){
					fileAsString = fileAsString.replace("->", "=");
				}
				fileAsString = fileAsString.replaceAll("(\\s)*\\:(\\s)*\\:(\\s)*new", "");
			}
			
			// striping static blocks --- [[[hemr]]]
			fileAsString = stripStaticBlock(actualFileAsString, fileAsString);
			
		} // final generic treatment
		
		// for debugging purposes remove the following comment - [[[hemr]]]
//		if(file.getCanonicalPath().endsWith("*.java"))
//		System.err.println(fileAsString);
		return fileAsString;
	}

	private static String getFieldDeclAnnotationsAsJML(StringBuffer annotationArea, JavaField annotatedJavaField ) {
		StringBuffer fieldDeclAnnotations = new StringBuffer("");
		
		if(annotationArea.toString().contains("@SpecPublic")){
			fieldDeclAnnotations.append("spec_public ");
		}			

		if(annotationArea.toString().contains("@SpecProtected")){
			fieldDeclAnnotations.append("spec_protected ");
		}
		
		if(annotationArea.toString().contains("@NonNull") ||
				annotationArea.toString().contains("@org.aspectjml.lang.annotation.NonNull")){
			for (int k = 0; k < annotatedJavaField.getAnnotations().length; k++) {
				if(annotatedJavaField.getAnnotations()[k].toString().equals("@org.aspectjml.lang.annotation.NonNull()")){
					fieldDeclAnnotations.append("non_null ");
					break;
				}
			}
		}
		if(annotationArea.toString().contains("@Nullable")){
			fieldDeclAnnotations.append("nullable ");
		}
		if(annotationArea.toString().contains("@Model")){
			fieldDeclAnnotations.append("model ");
		}
		
		if(fieldDeclAnnotations.toString().equals("")){
			return "";
		}
		else {
			return "/*@ "+fieldDeclAnnotations.toString()+"@*/";
		}
	}
	
	private static String getJavaDocTagAsJMLClauseForTypeToShift(List<DocletTag> docletTagFieldsToShift){
		StringBuffer jmlClauses = new StringBuffer("");
		for (Iterator<DocletTag> iterator = docletTagFieldsToShift.iterator(); iterator.hasNext();) {
			DocletTag docletTag = iterator.next();
			String tagName = docletTag.getName();
			String tagValue = docletTag.getValue().replaceAll("\n", "").replaceAll("\r", "");
			if(tagName.equals("model_definition")){
				String clause = "/*@ "+" "+tagValue+"@*/";
				jmlClauses.append(clause);
			}
			else if(tagName.equals("model_method_definition")){
				String clause = "/*@ "+" "+tagValue+"@*/";
				jmlClauses.append(clause);
			}
			else if(tagName.equals("model_field_definition")){
				String clause = "/*@ "+" "+tagValue+"@*/";
				jmlClauses.append(clause);
			}
			else if(tagName.equals("invariant_definition")){
				String clause = "/*@ "+" "+tagValue+"@*/";
				jmlClauses.append(clause);
			}
			// improved javadoc syntax [[[hemr]]]
			else if(tagName.equals("inv") && Main.aRacOptions.javadocBasedJML()){
				int startIndex = tagValue.indexOf("[");
				int endIndex = tagValue.lastIndexOf("]");
				if(startIndex != -1 && endIndex != -1){
					String clause = "/*@ "+"public invariant "+tagValue.substring(startIndex+1, endIndex)+";@*/";
					jmlClauses.append(clause);
				}
				else{
					String clause = "/*@ "+"public invariant "+tagValue+";@*/";
					jmlClauses.append(clause);
				}
			}
			else if(tagName.equals("constraint_definition")){
				String clause = "/*@ "+" "+tagValue+"@*/";
				jmlClauses.append(clause);
			}
			else if(tagName.equals("initially_definition")){
				String clause = "/*@ "+" "+tagValue+"@*/";
				jmlClauses.append(clause);
			}
			else if(tagName.equals("represents_definition")){
				String clause = "/*@ "+" "+tagValue+"@*/";
				jmlClauses.append(clause);
			}
		}
		return jmlClauses.toString();
	}
	
	private static String getJavaDocTagAsJMLClauseForFieldToShift(List<DocletTag> docletTagFieldsToShift){
		StringBuffer jmlClauses = new StringBuffer("");
		for (Iterator<DocletTag> iterator = docletTagFieldsToShift.iterator(); iterator.hasNext();) {
			DocletTag docletTag = iterator.next();
			String tagName = docletTag.getName();
			String tagValue = docletTag.getValue().replaceAll("\n", "").replaceAll("\r", "");
			if(tagName.equals("in")){
				String clauseKind = "in";
				String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
				jmlClauses.append(clause);
			}
			// improved javadoc syntax [[[hemr]]]
			else if(tagName.equals("inv") && Main.aRacOptions.javadocBasedJML()){
				int startIndex = tagValue.indexOf("[");
				int endIndex = tagValue.lastIndexOf("]");
				if(startIndex != -1 && endIndex != -1){
					String clause = "/*@ "+"public invariant "+tagValue.substring(startIndex+1, endIndex)+";@*/";
					jmlClauses.append(clause);
				}
				else{
					String clause = "/*@ "+"public invariant "+tagValue+";@*/";
					jmlClauses.append(clause);
				}
			}
		}
		return jmlClauses.toString();
	}
	
	private static String getJavaDocTagAsJMLClause(List<DocletTag> docletTags, int fileLineNumber, MjcCommonOptions options){
		DocletTag foundTag = null;
		String result = "";
		for (Iterator<DocletTag> iterator = docletTags.iterator(); iterator.hasNext();) {
			DocletTag docletTag = iterator.next();
			if(docletTag.getLineNumber() == fileLineNumber){
				foundTag = docletTag;
			}
		}
		if(foundTag != null){
			String tagName = foundTag.getName();
			String tagValue = foundTag.getValue().replaceAll("\n", "").replaceAll("\r", "");
			// allowing informal (non-assertion) brackets --- [[[hemr]]]
			tagValue = tagValue.replace("'['", "(").replace("']'", ")").replace("\"[\"", "(").replace("\"]\"", ")");
			if(tagName.equals("nullable_by_default")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "nullable_by_default";
				String clause = "/*@ "+clauseKind+" "+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("non_null_by_default")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "non_null_by_default";
				String clause = "/*@ "+clauseKind+" "+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("spec_case_header")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clause = "/*@ "+" "+tagValue+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("helper")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "helper";
				String clause = "/*@ "+clauseKind+" "+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("pure")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "pure";
				String clause = "/*@ "+clauseKind+" "+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("also")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "also";
				String clause = "/*@ "+clauseKind+" "+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("spec_public")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "spec_public";
				String clause = "/*@ "+clauseKind+" "+"@*/";
				result = start+line+clause+end;
			}
			// improved javadoc syntax [[[hemr]]]
			else if((tagName.equals("doc_public") || tagName.equals("doc_open")) 
					&& ((JmlCommonOptions) options).javadocBasedJML()){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "spec_public";
				String clause = "/*@ "+clauseKind+" "+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("spec_protected")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "spec_protected";
				String clause = "/*@ "+clauseKind+" "+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("non_null")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "non_null";
				String clause = "/*@ "+clauseKind+" "+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("nullable")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "nullable";
				String clause = "/*@ "+clauseKind+" "+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("requires")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "requires";
				// improved javadoc syntax [[[hemr]]]
				if(((JmlCommonOptions) options).javadocBasedJML()){
					int startIndex = tagValue.indexOf("[");
					int endIndex = tagValue.lastIndexOf("]");
					if(startIndex != -1 && endIndex != -1){
						String clause = "/*@ "+clauseKind+" "+tagValue.substring(startIndex+1, endIndex)+";@*/";
						result = start+line+clause+end;
					}
//					else{
//						String clause = "/*@ "+clauseKind+" "+tagValue+";@*/";
//						result = start+line+clause+end;
//					}
				}
				else{
					String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
					result = start+line+clause+end;
				}
//				result = result.replace("#forall", "\\forall");
//				result = result.replace("#exists", "\\exists");
				result = result.replace("@forall", "\\forall");
				result = result.replace("@exists", "\\exists");
			}
			else if(tagName.equals("pre")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "pre";
				// improved javadoc syntax [[[hemr]]]
				if(((JmlCommonOptions) options).javadocBasedJML()){
					int startIndex = tagValue.indexOf("[");
					int endIndex = tagValue.lastIndexOf("]");
					if(startIndex != -1 && endIndex != -1){
						String clause = "/*@ "+clauseKind+" "+tagValue.substring(startIndex+1, endIndex)+";@*/";
						result = start+line+clause+end;
					}
//					else{
//						String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
//						result = start+line+clause+end;
//					}
				}
				else{
					String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
					result = start+line+clause+end;
				}
				result = result.replace("@forall", "\\forall");
				result = result.replace("@exists", "\\exists");
//				result = result.replace("#forall", "\\forall");
//				result = result.replace("#exists", "\\exists");
			}
			// improved javadoc syntax [[[hemr]]]
			else if(tagName.equals("param") && ((JmlCommonOptions) options).javadocBasedJML()){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "requires";
				int startIndex = tagValue.indexOf("[");
				int endIndex = tagValue.lastIndexOf("]");
				if(startIndex != -1 && endIndex != -1){
					String clause = "/*@ "+clauseKind+" "+tagValue.substring(startIndex+1, endIndex)+";@*/";
					result = start+line+clause+end;
				}
				result = result.replace("@forall", "\\forall");
				result = result.replace("@exists", "\\exists");
//				result = result.replace("#forall", "\\forall");
//				result = result.replace("#exists", "\\exists");
			}
			else if(tagName.equals("ensures")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "ensures";
				// improved javadoc syntax [[[hemr]]]
				if(((JmlCommonOptions) options).javadocBasedJML()){
					int startIndex = tagValue.indexOf("[");
					int endIndex = tagValue.lastIndexOf("]");
					if(startIndex != -1 && endIndex != -1){
						String clause = "/*@ "+clauseKind+" "+tagValue.substring(startIndex+1, endIndex).replace("\\return", "\\result").
								replace("@return", "\\result").replace("@old", "\\old").replace("@forall", "\\forall").replace("@exists", "\\exists")+";@*/";
						result = start+line+clause+end;
					}
				}
				else{
					String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
					result = start+line+clause+end;
				}
				
			}
			else if(tagName.equals("post")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "post";
				// improved javadoc syntax [[[hemr]]]
				if(((JmlCommonOptions) options).javadocBasedJML()){
					int startIndex = tagValue.indexOf("[");
					int endIndex = tagValue.lastIndexOf("]");
					if(startIndex != -1 && endIndex != -1){
						String clause = "/*@ "+clauseKind+" "+tagValue.substring(startIndex+1, endIndex).replace("\\return", "\\result").
								replace("@return", "\\result").replace("@old", "\\old").replace("@forall", "\\forall").replace("@exists", "\\exists")+";@*/";
						result = start+line+clause+end;
					}
				}
				else{
					String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
					result = start+line+clause+end;
				}
			}
			// improved javadoc syntax [[[hemr]]]
			else if(tagName.equals("return") && ((JmlCommonOptions) options).javadocBasedJML()){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "ensures";
				int startIndex = tagValue.indexOf("[");
				int endIndex = tagValue.lastIndexOf("]");
				if(startIndex != -1 && endIndex != -1){
					String clause = "/*@ "+clauseKind+" "+tagValue.substring(startIndex+1, endIndex).replace("\\return", "\\result").
							replace("@return", "\\result").replace("@old", "\\old").replace("@forall", "\\forall").replace("@exists", "\\exists")+";@*/";
					result = start+line+clause+end;
				}
			}
			else if(tagName.equals("assignable")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "assignable";
				String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("signals")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "signals";
				String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
				result = start+line+clause+end;
			}
			// improved javadoc syntax [[[hemr]]]
			else if((tagName.equals("throws") || tagName.equals("exception")) && ((JmlCommonOptions) options).javadocBasedJML()){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = null;
				clauseKind = "signals";
				int startIndex = tagValue.indexOf("[");
				int endIndex = tagValue.lastIndexOf("]");
				if(startIndex != -1 && endIndex != -1){
					String exHeader = tagValue.substring(0, startIndex);
					if(exHeader.trim().startsWith("(")){
						int startIndexLParen = exHeader.indexOf("(");
						int endIndexRParen = exHeader.indexOf(")");
						if(startIndexLParen != -1 && endIndexRParen != -1){
							exHeader = exHeader.substring(startIndexLParen, endIndexRParen+1);
						}
					}
					else{
						int startIndexBlankChar = exHeader.indexOf(" ");
						int startIndexTabChar = exHeader.indexOf("\t");
						if(startIndexBlankChar != -1 && startIndexTabChar != -1){
							if(startIndexBlankChar < startIndexTabChar){
								exHeader = "("+exHeader.substring(0, startIndexBlankChar)+")";
							}
							else {
								exHeader = "("+exHeader.substring(0, startIndexTabChar)+")";
							}
						}
						else {
							if(startIndexBlankChar != -1) {
								exHeader = "("+exHeader.substring(0, startIndexBlankChar)+")";
							}
							else if(startIndexTabChar != -1) {
								exHeader = "("+exHeader.substring(0, startIndexTabChar)+")";
							}
						}
					}
					// old... [[[hemr]]]
//					int startIndexLParen = exHeader.indexOf("(");
//					int endIndexRParen = exHeader.indexOf(")");
//					if(startIndexLParen != -1 && endIndexRParen != -1){
//						exHeader = exHeader.substring(startIndexLParen, endIndexRParen+1);
//					}
//					else{
//						int startIndexBlankChar = exHeader.indexOf(" ");
//						int startIndexTabChar = exHeader.indexOf("\t");
//						if(startIndexBlankChar != -1 && startIndexTabChar != -1){
//							if(startIndexBlankChar < startIndexTabChar){
//								exHeader = "("+exHeader.substring(0, startIndexBlankChar)+")";
//							}
//							else {
//								exHeader = "("+exHeader.substring(0, startIndexTabChar)+")";
//							}
//						}
//						else {
//							if(startIndexBlankChar != -1) {
//								exHeader = "("+exHeader.substring(0, startIndexBlankChar)+")";
//							}
//							else if(startIndexTabChar != -1) {
//								exHeader = "("+exHeader.substring(0, startIndexTabChar)+")";
//							}
//						}
//					}
					String assertion = tagValue.substring(startIndex+1, endIndex);
					String predExp = exHeader + assertion;
//					if(assertion.equals("\\nothing")){
//						clauseKind = "signals_only";
//					}
//					// improved javadoc syntax [[[hemr]]]
//					// adding the "\not_allowed" support for documentation
//					else if(assertion.equals("\\not_allowed") || assertion.equals("@not_allowed")){
//						clauseKind = "signals_only";
//						assertion = "\\nothing";
//						predExp = assertion;
//					}
//					else{
//						clauseKind = "signals";
//					}
					String clause = "/*@ "+clauseKind+" "+predExp+";@*/";
					result = start+line+clause+end;
//					result = result.replace("#old", "\\old");
//					result = result.replace("#forall", "\\forall");
//					result = result.replace("#exists", "\\exists");
					result = result.replace("@old", "\\old");
					result = result.replace("@forall", "\\forall");
					result = result.replace("@exists", "\\exists");
				}
				else{
					clauseKind = "signals";
					int startIndexBlankChar = tagValue.indexOf(" ");
					String assertion = "";
					if(startIndexBlankChar != -1) {
						assertion = tagValue.substring(0, startIndexBlankChar);
					}
					else {
						assertion = tagValue;
					}
					String clause = "/*@ "+clauseKind+" ("+assertion+") true;@*/";
					result = start+line+clause+end;
				}
			}
			// improved javadoc syntax [[[hemr]]]
			else if((tagName.equals("nothrows") || tagName.equals("noexception")) 
					&& ((JmlCommonOptions) options).javadocBasedJML()){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "signals_only";
				String clause = "/*@ "+clauseKind+" "+"\\nothing"+";@*/";
				result = start+line+clause+end;

			}
			// improved javadoc syntax [[[hemr]]]
			else if((tagName.equals("onlythrows") || tagName.equals("onlyexception")) 
					&& ((JmlCommonOptions) options).javadocBasedJML()){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "signals_only";
				int startIndexBlankChar = tagValue.indexOf(" ");
				String assertion = "";
				if(startIndexBlankChar != -1) {
					assertion = tagValue.substring(0, startIndexBlankChar);
				}
				else {
					assertion = tagValue;
				}
				String clause = "/*@ "+clauseKind+" "+assertion+";@*/";
				result = start+line+clause+end;

			}
			else if(tagName.equals("signals_only")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "signals_only";
				String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("callable")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "callable";
				String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("accessible")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "accessible";
				String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("duration")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "duration";
				String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("captures")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "captures";
				String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
				result = start+line+clause+end;
			}
			else if(tagName.equals("working_space")){
				String start = "*/";
				String end = "/**";
				String line = "/*"+foundTag.getLineNumber()+"*/";
				String clauseKind = "working_space";
				String clause = "/*@ "+clauseKind+" "+tagValue+"@*/";
				result = start+line+clause+end;
			}
		}
		return result;
	}
	
	private static String stripStaticBlock(String actualFileAsString, String fileAsString) throws IOException{
		List<com.github.antlrjavaparser.api.body.InitializerDeclaration> javaDeclInitializers = QDoxUtil.getAllDeclaredJavaInitializersInFile(actualFileAsString);
		
		BufferedReader buffer = null;
		StringBuffer staticBlock = null;
		for (Iterator iterator = javaDeclInitializers.iterator(); iterator.hasNext();) {
			com.github.antlrjavaparser.api.body.InitializerDeclaration initializerDeclaration = 
					(com.github.antlrjavaparser.api.body.InitializerDeclaration) iterator.next();
			int beginLine = initializerDeclaration.getBeginLine();
			int endLine = initializerDeclaration.getEndLine();
			
			String line = "";
			staticBlock = new StringBuffer("");
			buffer = new BufferedReader(new StringReader(fileAsString));
			line = buffer.readLine();
			int lineNumber = 1;
			while (line != null) {
				if(lineNumber == endLine){
					staticBlock.append(line);
					break;
				}
				if(lineNumber == beginLine){
					int staticBlockBegin = line.indexOf("static");
					if(staticBlockBegin != -1){
						staticBlock.append(line.substring(staticBlockBegin));
					}
					else{
						staticBlock.append(line);
					}
					staticBlock.append("\n");
				}
				else if(lineNumber > beginLine){
					staticBlock.append(line);
					staticBlock.append("\n");
				}
				line = buffer.readLine();
				lineNumber++;
			}
//			System.out.println("staticBlock = "+staticBlock.toString());
			// does it a static block?? --- [[[hemr]]]
			if(staticBlock.toString().contains("static") && staticBlock.toString().contains("{")){
				StringBuffer staticBlockSurrogate = new StringBuffer("");
				int newLineCount = QDoxUtil.getLineNumbersQtd(staticBlock.toString());
				staticBlockSurrogate.append("static {");
				staticBlockSurrogate.append(QDoxUtil.getNewLinesCaracter(newLineCount));
				staticBlockSurrogate.append("}");
//				System.out.println("staticBlockSurrogate = "+staticBlockSurrogate.toString());
				fileAsString = StringUtils.replaceOnce(fileAsString, staticBlock.toString(), staticBlockSurrogate.toString());
			}
			buffer.close();
		}		
		return fileAsString;
	}
	
	private static String changeFileLine(String fileAsString, int lineNumberTochange, String lineSurrogate) throws IOException{
		BufferedReader buffer = null;
		StringBuffer bufferedFile = null;
		String line = "";

		bufferedFile = new StringBuffer("");
		buffer = new BufferedReader(new StringReader(fileAsString));
		line = buffer.readLine();
		int lineNumber = 1;
		while (line != null) {
			if(lineNumber == lineNumberTochange){
				bufferedFile.append(lineSurrogate);
			}
			else{
				bufferedFile.append(line);
			}
			bufferedFile.append("\n");
			line = buffer.readLine();
			lineNumber ++;
		}
		buffer.close();
		return bufferedFile.toString();
	}
	
	private static String appendFileLine(String fileAsString, int lineNumberTochange, String lineAppend) throws IOException{
		BufferedReader buffer = null;
		StringBuffer bufferedFile = null;
		String line = "";

		bufferedFile = new StringBuffer("");
		buffer = new BufferedReader(new StringReader(fileAsString));
		line = buffer.readLine();
		int lineNumber = 1;
		while (line != null) {
			if(lineNumber == lineNumberTochange){
				bufferedFile.append(line+lineAppend);
			}
			else{
				bufferedFile.append(line);
			}
			bufferedFile.append("\n");
			line = buffer.readLine();
			lineNumber ++;
		}
		buffer.close();
		return bufferedFile.toString();
	}
	
	private static String appendAtBeginFileLine(String fileAsString, int lineNumberTochange, String lineAppend) throws IOException{
		BufferedReader buffer = null;
		StringBuffer bufferedFile = null;
		String line = "";

		bufferedFile = new StringBuffer("");
		buffer = new BufferedReader(new StringReader(fileAsString));
		line = buffer.readLine();
		int lineNumber = 1;
		while (line != null) {
			if(lineNumber == lineNumberTochange){
				bufferedFile.append(lineAppend+line);
			}
			else{
				bufferedFile.append(line);
			}
			bufferedFile.append("\n");
			line = buffer.readLine();
			lineNumber ++;
		}
		buffer.close();
		return bufferedFile.toString();
	}
	
	private static String getFieldJMLInAnnotationsProcessedAsStandardJML(
			JavaField annotatedJavaField) {
		StringBuffer inAnno = new StringBuffer("");
		//start
		for (int i = 0; i < annotatedJavaField.getAnnotations().length; i++) {
			Annotation currentAnnotation = annotatedJavaField.getAnnotations()[i];
			// handling single JML annotations
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.In")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "in_redundantly" : "in") : "in");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+clauseKind+" "+pred+"; @*/";
					inAnno.append(clause);
//					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
//			not needed due to repeatable annotations in java 8 --- [[[hemr]]]
			// handling composite JML annotations
//			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.InDefinitions")){
//				if(currentAnnotation.getPropertyMap().containsKey("value")){
//					@SuppressWarnings("unchecked")
//					ArrayList<Annotation> definitions = (ArrayList<Annotation>)currentAnnotation.getProperty("value").getParameterValue();
//					for (Iterator<Annotation> iterator2 = definitions
//							.iterator(); iterator2.hasNext();) {
//						Annotation currentDef =  iterator2.next();
//						if(currentDef.getNamedParameterMap().containsKey("value")){
//							String clauseKind = (currentDef.getNamedParameterMap().containsKey("redundantly")? 
//									(currentDef.getProperty("redundantly").toString().equals("true")? "in_redundantly" : "in") : "in");
//							String preLineNumber = "/*"+currentDef.getLineNumber()+"*/";
//							String pred = QDoxUtil.preparePred(currentDef.getProperty("value").toString());
//							String clause = "/*@ "+clauseKind+" "+pred+"; @*/";
//							annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
//						}
//					}
//				}
//			}
		}
		return inAnno.toString();
	}

	private static String getTypeAnnotationAreaCommentedProcessedWithJML(
			JavaClass annotatedJavaType, String annotationAreaCommentedStr) {
		//start
		for (int i = 0; i < annotatedJavaType.getAnnotations().length; i++) {
			Annotation currentAnnotation = annotatedJavaType.getAnnotations()[i];
			// handling single JML annotations
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Invariant")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String header = (currentAnnotation.getNamedParameterMap().containsKey("header")? QDoxUtil.preparePred(currentAnnotation.getProperty("header").toString())+" ":"");
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "invariant_redundantly" : "invariant") : "invariant");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+header+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Constraint")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String header = (currentAnnotation.getNamedParameterMap().containsKey("header")? QDoxUtil.preparePred(currentAnnotation.getProperty("header").toString())+" ":"");
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "constraint_redundantly" : "constraint") : "constraint");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+header+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Initially")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String header = (currentAnnotation.getNamedParameterMap().containsKey("header")? QDoxUtil.preparePred(currentAnnotation.getProperty("header").toString())+" ":"");
					String clauseKind = "initially";
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+header+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Represents")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String header = (currentAnnotation.getNamedParameterMap().containsKey("header")? QDoxUtil.preparePred(currentAnnotation.getProperty("header").toString())+" ":"");
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "represents_redundantly" : "represents") : "represents");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+header+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Model")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String header = "model ";
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+header+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.ModelMethod")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String header = "model ";
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+header+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.ModelField")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String header = "model ";
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+header+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			// handling composite JML annotations
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.InvariantDefinitions")){
				if(currentAnnotation.getPropertyMap().containsKey("value")){
					@SuppressWarnings("unchecked")
					ArrayList<Annotation> definitions = (ArrayList<Annotation>)currentAnnotation.getProperty("value").getParameterValue();
					for (Iterator<Annotation> iterator2 = definitions
							.iterator(); iterator2.hasNext();) {
						Annotation currentDef =  iterator2.next();
						if(currentDef.getNamedParameterMap().containsKey("value")){
							String header = (currentDef.getNamedParameterMap().containsKey("header")? QDoxUtil.preparePred(currentDef.getProperty("header").toString())+" ":"");
							String clauseKind = (currentDef.getNamedParameterMap().containsKey("redundantly")? 
									(currentDef.getProperty("redundantly").toString().equals("true")? "invariant_redundantly" : "invariant") : "invariant");
							String preLineNumber = "/*"+currentDef.getLineNumber()+"*/";
							String pred = QDoxUtil.preparePred(currentDef.getProperty("value").toString());
							String clause = "/*@ "+header+clauseKind+" "+pred+"; @*/";
							annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
						}
					}
				}
			}
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.ConstraintDefinitions")){
				if(currentAnnotation.getPropertyMap().containsKey("value")){
					@SuppressWarnings("unchecked")
					ArrayList<Annotation> definitions = (ArrayList<Annotation>)currentAnnotation.getProperty("value").getParameterValue();
					for (Iterator<Annotation> iterator2 = definitions
							.iterator(); iterator2.hasNext();) {
						Annotation currentDef =  iterator2.next();
						if(currentDef.getNamedParameterMap().containsKey("value")){
							String header = (currentDef.getNamedParameterMap().containsKey("header")? QDoxUtil.preparePred(currentDef.getProperty("header").toString())+" ":"");
							String clauseKind = (currentDef.getNamedParameterMap().containsKey("redundantly")? 
									(currentDef.getProperty("redundantly").toString().equals("true")? "constraint_redundantly" : "constraint") : "constraint");
							String preLineNumber = "/*"+currentDef.getLineNumber()+"*/";
							String pred = QDoxUtil.preparePred(currentDef.getProperty("value").toString());
							String clause = "/*@ "+header+clauseKind+" "+pred+"; @*/";
							annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
						}
					}
				}
			}
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.InitiallyDefinitions")){
				if(currentAnnotation.getPropertyMap().containsKey("value")){
					@SuppressWarnings("unchecked")
					ArrayList<Annotation> definitions = (ArrayList<Annotation>)currentAnnotation.getProperty("value").getParameterValue();
					for (Iterator<Annotation> iterator2 = definitions
							.iterator(); iterator2.hasNext();) {
						Annotation currentDef =  iterator2.next();
						if(currentDef.getNamedParameterMap().containsKey("value")){
							String header = (currentDef.getNamedParameterMap().containsKey("header")? QDoxUtil.preparePred(currentDef.getProperty("header").toString())+" ":"");
							String clauseKind = "initially";
							String preLineNumber = "/*"+currentDef.getLineNumber()+"*/";
							String pred = QDoxUtil.preparePred(currentDef.getProperty("value").toString());
							String clause = "/*@ "+header+clauseKind+" "+pred+"; @*/";
							annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
						}
					}
				}
			}
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.RepresentsDefinitions")){
				if(currentAnnotation.getPropertyMap().containsKey("value")){
					@SuppressWarnings("unchecked")
					ArrayList<Annotation> definitions = (ArrayList<Annotation>)currentAnnotation.getProperty("value").getParameterValue();
					for (Iterator<Annotation> iterator2 = definitions
							.iterator(); iterator2.hasNext();) {
						Annotation currentDef =  iterator2.next();
						if(currentDef.getNamedParameterMap().containsKey("value")){
							String header = (currentDef.getNamedParameterMap().containsKey("header")? QDoxUtil.preparePred(currentDef.getProperty("header").toString())+" ":"");
							String clauseKind = (currentDef.getNamedParameterMap().containsKey("redundantly")? 
									(currentDef.getProperty("redundantly").toString().equals("true")? "represents_redundantly" : "represents") : "represents");
							String preLineNumber = "/*"+currentDef.getLineNumber()+"*/";
							String pred = QDoxUtil.preparePred(currentDef.getProperty("value").toString());
							String clause = "/*@ "+header+clauseKind+" "+pred+"; @*/";
							annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
						}
					}
				}
			}
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.ModelDefinitions")){
				if(currentAnnotation.getPropertyMap().containsKey("value")){
					@SuppressWarnings("unchecked")
					ArrayList<Annotation> definitions = (ArrayList<Annotation>)currentAnnotation.getProperty("value").getParameterValue();
					for (Iterator<Annotation> iterator2 = definitions
							.iterator(); iterator2.hasNext();) {
						Annotation currentDef =  iterator2.next();
						if(currentDef.getNamedParameterMap().containsKey("value")){
							String header = "model ";
							String preLineNumber = "/*"+currentDef.getLineNumber()+"*/";
							String pred = QDoxUtil.preparePred(currentDef.getProperty("value").toString());
							String clause = "/*@ "+header+" "+pred+"; @*/";
							annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
						}
					}	
				}
				if(currentAnnotation.getPropertyMap().containsKey("fields")){
					@SuppressWarnings("unchecked")
					ArrayList<Annotation> definitions = (ArrayList<Annotation>)currentAnnotation.getProperty("fields").getParameterValue();
					for (Iterator<Annotation> iterator2 = definitions
							.iterator(); iterator2.hasNext();) {
						Annotation currentDef =  iterator2.next();
						if(currentDef.getNamedParameterMap().containsKey("value")){
							String header = "model ";
							String preLineNumber = "/*"+currentDef.getLineNumber()+"*/";
							String pred = QDoxUtil.preparePred(currentDef.getProperty("value").toString());
							String clause = "/*@ "+header+" "+pred+"; @*/";
							annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
						}
					}	
				}
				if(currentAnnotation.getPropertyMap().containsKey("methods")){
					@SuppressWarnings("unchecked")
					ArrayList<Annotation> definitions = (ArrayList<Annotation>)currentAnnotation.getProperty("methods").getParameterValue();
					for (Iterator<Annotation> iterator2 = definitions
							.iterator(); iterator2.hasNext();) {
						Annotation currentDef =  iterator2.next();
						if(currentDef.getNamedParameterMap().containsKey("value")){
							String header = "model ";
							String preLineNumber = "/*"+currentDef.getLineNumber()+"*/";
							String pred = QDoxUtil.preparePred(currentDef.getProperty("value").toString());
							String clause = "/*@ "+header+" "+pred+"; @*/";
							annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
						}
					}	
				}
			}
		}
		// end
		return annotationAreaCommentedStr;
	}

	// helper methods

	public static List<JavaMethod> getAllDeclaredJavaMethodsInFile(JavaDocBuilder qDoxFile){
		List<JavaMethod> result = new ArrayList<JavaMethod>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			JavaMethod [] javaDeclMethods = currentJavaDeclType.getMethods();
			for (int j = 0; j < javaDeclMethods.length; j++) {
				JavaMethod currentJavaDeclMethod = javaDeclMethods[j];
				result.add(currentJavaDeclMethod);
			}
		}
		Comparator<JavaMethod> COMPARE_BY_LINE_NUMBER = new Comparator<JavaMethod>() {
			public int compare(JavaMethod one, JavaMethod other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	public static List<JavaMethod> getAllDeclaredJavaMethodsInFile2(JavaClass [] javaDeclTypes){
		List<JavaMethod> result = new ArrayList<JavaMethod>();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			JavaMethod [] javaDeclMethods = currentJavaDeclType.getMethods();
			for (int j = 0; j < javaDeclMethods.length; j++) {
				JavaMethod currentJavaDeclMethod = javaDeclMethods[j];
				result.add(currentJavaDeclMethod);
			}
		}
		Comparator<JavaMethod> COMPARE_BY_LINE_NUMBER = new Comparator<JavaMethod>() {
			public int compare(JavaMethod one, JavaMethod other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static void getAllDeclaredAnnotatedITDJavaMethodsInFile(JavaDocBuilder qDoxFile){
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(isPrincipalEnclosingTypeDeclNotExposeOnly(currentJavaDeclType)){
				JavaMethod [] javaDeclMethods = currentJavaDeclType.getMethods();
				for (int j = 0; j < javaDeclMethods.length; j++) {
					JavaMethod currentJavaDeclMethod = javaDeclMethods[j];
					if(!currentJavaDeclMethod.getCodeBlock().contains("/*##defaultmethod##*/")){
						if(!currentJavaDeclMethod.getParentClass().isInner()) {
							AspectUtil.getInstance().appendItdMethDecl(currentJavaDeclMethod);	
						}
						else if(!isTypeDeclOpen(currentJavaDeclMethod.getParentClass())) {
							AspectUtil.getInstance().appendItdMethDecl(currentJavaDeclMethod);	
						}
					}
				}
			}
		}
	}
	
	private static void getAllDeclaredAnnotatedITDJavaClassesInFile(JavaDocBuilder qDoxFile){
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(currentJavaDeclType.isInner() 
					&& !isPrincipalEnclosingTypeDeclExposeOnly(currentJavaDeclType)){
//				System.out.println("--> "+currentJavaDeclType.getFullyQualifiedName());
				
//				[1 - isTypeDeclOpen]
				if(isTypeDeclOpen(currentJavaDeclType)) {
//					[1.1 - isEnclosingTypeDeclOpen]
					if(currentJavaDeclType.getParentClass() != null && isTypeDeclOpen(currentJavaDeclType.getParentClass())) {
						if(!currentJavaDeclType.getParentClass().isInner()) {
//							System.out.println("[included for ITD]: "+currentJavaDeclType.getFullyQualifiedName());
							AspectUtil.getInstance().appendItdClassDecl(currentJavaDeclType);
						}
					}
					else if(currentJavaDeclType.getParentClass() != null && !isTypeDeclOpen(currentJavaDeclType.getParentClass())) {
//						System.out.println("[included for ITD]: "+currentJavaDeclType.getFullyQualifiedName());
						AspectUtil.getInstance().appendItdClassDecl(currentJavaDeclType);
					}
				}
				
//				if(isPrincipalEnclosingTypeDeclNotExposeOnly(currentJavaDeclType)){
//					// ensuring one level of nesting...[[[hemr]]]
//					if(isTypeDeclPartialOrOpen(currentJavaDeclType.getParentClass()) 
//							&& isTypeDeclPartialOrOpen(currentJavaDeclType)){
//						AspectUtil.getInstance().appendItdClassDecl(currentJavaDeclType);
//					}
//				}	
			}
		}
	}
	
	public static boolean isFieldDeclPartial(JavaField javaField){
		boolean result = false;
		for (int i = 0; i < javaField.getAnnotations().length; i++) {
			Annotation currentAnno = javaField.getAnnotations()[i];
			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Partial")) ||
					(currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.ExposeOnly"))){
				result = true;
				break;
			}
//			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Partial"))){
//				result = true;
//				break;
//			}
		}
		return result;
	}
	
	public static boolean isFieldDeclDeclareErrorOrWarning(JavaField javaField){
		boolean result = false;
		for (int i = 0; i < javaField.getAnnotations().length; i++) {
			Annotation currentAnno = javaField.getAnnotations()[i];
			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.DeclareError"))||
					(currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.DeclareWarning"))){
				result = true;
				break;
			}
		}
		return result;
	}
	
//	public static boolean isFieldDeclDeclareError(JavaField javaField){
//		boolean result = false;
//		for (int i = 0; i < javaField.getAnnotations().length; i++) {
//			Annotation currentAnno = javaField.getAnnotations()[i];
//			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.DeclareError"))){
//				result = true;
//				break;
//			}
//		}
//		return result;
//	}
//	
//	public static boolean isFieldDeclDeclareWarning(JavaField javaField){
//		boolean result = false;
//		for (int i = 0; i < javaField.getAnnotations().length; i++) {
//			Annotation currentAnno = javaField.getAnnotations()[i];
//			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.DeclareWarning"))){
//				result = true;
//				break;
//			}
//		}
//		return result;
//	}
	
	public static JavaClass getTopOfTopOwner(JavaClass javaClass){
		if(javaClass.isInner()){
			while(javaClass.getParentClass() != null){
				javaClass = javaClass.getParentClass();
			}
		}
		return javaClass;
	}
	
//	public static boolean isTypeDeclWrongOpenSpec(JavaClass javaClass){
//		boolean result = false;
//		
//		for (int i = 0; i < javaClass.getAnnotations().length; i++) {
//			Annotation currentAnno = javaClass.getAnnotations()[i];
//			if(currentAnno.toString().startsWith("@OpenS")){
//				result = true;
//				break;
//			}
//			if(currentAnno.toString().startsWith("@Open")){
//				result = true;
//				break;
//			}
//			if(currentAnno.toString().startsWith("@Partial")){
//				result = true;
//				break;
//			}
//		}
//		return result;
//	}
	
	public static boolean isTypeDeclReadOnly(JavaClass javaClass){
		boolean result = false;
		
		for (int i = 0; i < javaClass.getAnnotations().length; i++) {
			Annotation currentAnno = javaClass.getAnnotations()[i];
			if(currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.ReadOnly")){
				result = true;
				break;
			}
		}
		return result;
	}
	
	private static boolean isTypeDeclPartialOrExposeOnly(JavaClass javaClass){
		boolean result = false;
		
		for (int i = 0; i < javaClass.getAnnotations().length; i++) {
			Annotation currentAnno = javaClass.getAnnotations()[i];
			if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Partial")) ||
					(currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.ExposeOnly"))){
				result = true;
				break;
			}
		}
		return result;
	}
	
	public static boolean isTypeDeclOpenSpecOrSpecsOnly(JavaClass javaClass){
		boolean result = false;
		
		for (int i = 0; i < javaClass.getAnnotations().length; i++) {
			Annotation currentAnno = javaClass.getAnnotations()[i];
			if(currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.OpenS")){
				result = true;
				break;
			}
			else if(currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.SpecsOnly")){
				result = true;
				break;
			}
		}
		return result;
	}
	
//	public static boolean isTypeDeclPartialOrOpen(JavaClass javaClass){
//		boolean result = false;
//		
//		for (int i = 0; i < javaClass.getAnnotations().length; i++) {
//			Annotation currentAnno = javaClass.getAnnotations()[i];
//			if(!currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.OpenS")){
//				if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Partial")) ||
//						(currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Open"))){
//					result = true;
//					break;
//				}
//			}
//			
//		}
//		return result;
//	}
	
	// New way to deal with openess in AspectJML --- [[[hemr]]]
	public static boolean isTypeDeclPartialOrOpen(JavaClass javaClass){
		boolean result = false;
		
		if(!isTypeDeclReadOnly(javaClass) && !isTypeDeclOpenSpecOrSpecsOnly(javaClass)){
			result = true;
			if(javaClass.isInner()){
				if(isTypeDeclPartialOrExposeOnly(javaClass)){
					result = false;
				}
			}
		}
		return result;
	}
	
	public static boolean isTypeDeclOpen(JavaClass javaClass){
		boolean result = false;
		
		if(!isTypeDeclReadOnly(javaClass) 
				&& !isTypeDeclOpenSpecOrSpecsOnly(javaClass)
				&& !isTypeDeclPartialOrExposeOnly(javaClass)){
			result = true;
		}
		return result;
	}
	
//	public static boolean isTypeDeclIndirectPartialOrOpen(JavaClass javaClass){
//		boolean result = false;
//		
//		if(javaClass.isInner()){
//			while(javaClass.getParentClass() != null){
//				javaClass = javaClass.getParentClass();
//			}
//		}
//		
//		for (int i = 0; i < javaClass.getAnnotations().length; i++) {
//			Annotation currentAnno = javaClass.getAnnotations()[i];
//			if(!currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.OpenS")){
//				if((currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Partial")) ||
//						(currentAnno.toString().startsWith("@org.aspectjml.lang.annotation.Open"))){
//					result = true;
//					break;
//				}
//			}
//		}
//		return result;
//	}
	
	// New way to deal with openess in AspectJML --- [[[hemr]]]
	public static boolean isPrincipalEnclosingTypeDeclNotExposeOnly(JavaClass javaClass){
		boolean result = false;
		
		if(javaClass.isInner()){
			while(javaClass.getParentClass() != null){
				javaClass = javaClass.getParentClass();
			}
		}
		
		if(!isTypeDeclReadOnly(javaClass) 
				&& !isTypeDeclPartialOrExposeOnly(javaClass) 
				&& !isTypeDeclOpenSpecOrSpecsOnly(javaClass)){
			result = true;
		}
		return result;
	}
	
	public static boolean isPrincipalEnclosingTypeDeclExposeOnly(JavaClass javaClass){
		return !isPrincipalEnclosingTypeDeclNotExposeOnly(javaClass);
	}
	
	private static List<JavaMethod> getAllDeclaredAnnotatedAdviceJavaMethodsInFile(JavaDocBuilder qDoxFile, boolean isAspectJMLFile){
		List<JavaMethod> result = new ArrayList<JavaMethod>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			boolean isPartialOrDR = true;
			if(isAspectJMLFile){
				if(currentJavaDeclType.isInner()){
					JavaClass topOwner = QDoxUtil.getTopOfTopOwner(currentJavaDeclType);
					for (int m = 0; m < topOwner.getAnnotations().length; m++) {
						if(topOwner.getAnnotations()[m].toString().startsWith("@org.aspectjml.lang.annotation.OpenS")){
							isPartialOrDR = false;
							break;
						}
						else if(topOwner.getAnnotations()[m].toString().startsWith("@org.aspectjml.lang.annotation.ReadOnly")){
							isPartialOrDR = false;
							break;
						}
					}
				}
				else{
					for (int m = 0; m < currentJavaDeclType.getAnnotations().length; m++) {
						if(currentJavaDeclType.getAnnotations()[m].toString().startsWith("@org.aspectjml.lang.annotation.OpenS")){
							isPartialOrDR = false;
							break;
						}
						else if(currentJavaDeclType.getAnnotations()[m].toString().startsWith("@org.aspectjml.lang.annotation.ReadOnly")){
							isPartialOrDR = false;
							break;
						}
					}
				}
			}
			
			JavaMethod [] javaDeclMethods = currentJavaDeclType.getMethods();
			if(isPartialOrDR && isAspectJMLFile){
				for (int j = 0; j < javaDeclMethods.length; j++) {
					JavaMethod currentJavaDeclMethod = javaDeclMethods[j];
					for (int l = 0; l < currentJavaDeclMethod.getAnnotations().length; l++) {
						if(currentJavaDeclMethod.getAnnotations()[l].toString().startsWith("@org.aspectjml.lang.annotation.Before")){
							result.add(currentJavaDeclMethod);
							break;
						}
						else if(currentJavaDeclMethod.getAnnotations()[l].toString().startsWith("@org.aspectjml.lang.annotation.After")){
							result.add(currentJavaDeclMethod);
							break;
						}
						else if(currentJavaDeclMethod.getAnnotations()[l].toString().startsWith("@org.aspectjml.lang.annotation.Around")){
							result.add(currentJavaDeclMethod);
							break;
						}
						else if(currentJavaDeclMethod.getAnnotations()[l].toString().startsWith("@org.aspectjml.lang.annotation.Revise")){
							result.add(currentJavaDeclMethod);
							break;
						}
						else if(currentJavaDeclMethod.getAnnotations()[l].toString().startsWith("@org.aspectjml.lang.annotation.Append")){
							result.add(currentJavaDeclMethod);
							break;
						}
						else if(currentJavaDeclMethod.getAnnotations()[l].toString().startsWith("@org.aspectjml.lang.annotation.Add")){
							result.add(currentJavaDeclMethod);
							break;
						}
					}
				}
			}
			else if(!isAspectJMLFile){
				for (int j = 0; j < javaDeclMethods.length; j++) {
					JavaMethod currentJavaDeclMethod = javaDeclMethods[j];
					if(!currentJavaDeclMethod.getSource().getFile().getAbsolutePath().endsWith(".jml")){
						for (int l = 0; l < currentJavaDeclMethod.getAnnotations().length; l++) {
							if(currentJavaDeclMethod.getAnnotations()[l].toString().startsWith("@org.aspectjml.lang.annotation.Before")){
								result.add(currentJavaDeclMethod);
								break;
							}
							else if(currentJavaDeclMethod.getAnnotations()[l].toString().startsWith("@org.aspectjml.lang.annotation.After")){
								result.add(currentJavaDeclMethod);
								break;
							}
							else if(currentJavaDeclMethod.getAnnotations()[l].toString().startsWith("@org.aspectjml.lang.annotation.Around")){
								result.add(currentJavaDeclMethod);
								break;
							}
						}
					}
				}
			}
		}
		Comparator<JavaMethod> COMPARE_BY_LINE_NUMBER = new Comparator<JavaMethod>() {
			public int compare(JavaMethod one, JavaMethod other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<JavaMethod> getAllDeclaredAnnotatedSuppliedJavaMethodsInFile(JavaDocBuilder qDoxFile){
		List<JavaMethod> result = new ArrayList<JavaMethod>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			boolean isPartialOrDR = true;
			if(isTypeDeclReadOnly(currentJavaDeclType)){
				isPartialOrDR = false;
			}
			
			// eventually, this should be confirmed... [[[hemr]]]
			if(isPartialOrDR){
				JavaMethod [] javaDeclMethods = currentJavaDeclType.getMethods();
				for (int j = 0; j < javaDeclMethods.length; j++) {
					JavaMethod currentJavaDeclMethod = javaDeclMethods[j];
					if(!currentJavaDeclMethod.getSource().getFile().getAbsolutePath().endsWith(".jml")){
						for (int l = 0; l < currentJavaDeclMethod.getAnnotations().length; l++) {
							if(currentJavaDeclMethod.getAnnotations()[l].toString().startsWith("@org.aspectjml.lang.annotation.Supply")){
								result.add(currentJavaDeclMethod);
								break;
							}
						}
					}
				}
			}
		}
		Comparator<JavaMethod> COMPARE_BY_LINE_NUMBER = new Comparator<JavaMethod>() {
			public int compare(JavaMethod one, JavaMethod other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<JavaMethod> getAllDeclaredAnnotatedDeclareSoftJavaMethodsInFile(JavaDocBuilder qDoxFile){
		List<JavaMethod> result = new ArrayList<JavaMethod>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			boolean isAspect = false;
			for (int m = 0; m < currentJavaDeclType.getAnnotations().length; m++) {
				if(currentJavaDeclType.getAnnotations()[m].toString().startsWith("@org.aspectj.lang.annotation.Aspect")){
					isAspect = true;
					break;
				}
			}
			if(!isAspect){
				JavaMethod [] javaDeclMethods = currentJavaDeclType.getMethods();
				for (int j = 0; j < javaDeclMethods.length; j++) {
					JavaMethod currentJavaDeclMethod = javaDeclMethods[j];
					for (int l = 0; l < currentJavaDeclMethod.getAnnotations().length; l++) {
						 if(currentJavaDeclMethod.getAnnotations()[l].toString().startsWith("@org.aspectjml.lang.annotation.DeclareSoft")){
							result.add(currentJavaDeclMethod);
							break;
						}
					}
				}
			}
		}
		Comparator<JavaMethod> COMPARE_BY_LINE_NUMBER = new Comparator<JavaMethod>() {
			public int compare(JavaMethod one, JavaMethod other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
//	public static List<String> getAllImplementAndExtendClauseAsDeclareParentsInFile(JavaDocBuilder qDoxFile){
//		List<String> result = new ArrayList<String>();
//		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();
//
//		for (int i = 0; i < javaDeclTypes.length; i++) {
//			JavaClass currentJavaDeclType = javaDeclTypes[i];
//			if(isTypeDeclOpen(currentJavaDeclType) || currentJavaDeclType.isInner()){
//				// handling super-class
//				if (currentJavaDeclType.getSuperJavaClass() != null)
//				if(!currentJavaDeclType.getSuperJavaClass().getFullyQualifiedName().equals("java.lang.Object")){
//					result.add("declare parents: "+currentJavaDeclType.getFullyQualifiedName().replace("$",".")+
//							" extends "+currentJavaDeclType.getSuperJavaClass().getFullyQualifiedName().replace("$",".")+";");
//				}
//				// handling implemented interfaces
//				JavaClass [] interfaces = currentJavaDeclType.getImplementedInterfaces();
//				for (int j = 0; j < interfaces.length; j++) {
//					result.add("declare parents: "+currentJavaDeclType.getFullyQualifiedName().replace("$",".")+
//							" implements "+interfaces[j].getFullyQualifiedName().replace("$",".")+";");
//				}
//			}
//		}
//		
//		return result;
//	}
	
	private static List<com.github.antlrjavaparser.api.body.BodyDeclaration> getAllDeclaredAnnotatedJavaMethodsInFile(String fileAsString) throws com.github.antlrjavaparser.ParseException, UnsupportedEncodingException, IOException{
		com.github.antlrjavaparser.api.CompilationUnit cu = com.github.antlrjavaparser.JavaParser.parse(new ByteArrayInputStream(fileAsString.getBytes("UTF-8")));
		List<com.github.antlrjavaparser.api.body.BodyDeclaration> result = new ArrayList<com.github.antlrjavaparser.api.body.BodyDeclaration>();
		
		for (Iterator iterator = cu.getTypes().iterator(); iterator.hasNext();) {
			com.github.antlrjavaparser.api.body.TypeDeclaration type = (com.github.antlrjavaparser.api.body.TypeDeclaration) iterator.next();
			if(type.getMembers() != null){
				for (Iterator iterator2 = type.getMembers().iterator(); iterator2
						.hasNext();) {
					com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
					if(member instanceof com.github.antlrjavaparser.api.body.ConstructorDeclaration){
						if(member.getAnnotations().size() > 0){
							result.add(member);
						}
					}
					else if(member instanceof com.github.antlrjavaparser.api.body.MethodDeclaration){
						if(member.getAnnotations().size() > 0){
							result.add(member);
						}
					}
					else if(member instanceof com.github.antlrjavaparser.api.body.AnnotationMemberDeclaration){
						if(member.getAnnotations().size() > 0){
							result.add(member);
						}
					}
					else if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
						QDoxUtil.getAllDeclaredAnnotatedJavaMethodsInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
					}
					else {
						continue;
					}
				}
			}
		}	
		Comparator<com.github.antlrjavaparser.api.body.BodyDeclaration> COMPARE_BY_LINE_NUMBER = new Comparator<com.github.antlrjavaparser.api.body.BodyDeclaration>() {
			public int compare(com.github.antlrjavaparser.api.body.BodyDeclaration one, com.github.antlrjavaparser.api.body.BodyDeclaration other) {
				return Integer.valueOf(one.getBeginLine()).compareTo(other.getBeginLine());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<com.github.antlrjavaparser.api.body.BodyDeclaration> getAllDeclaredJavaMethodsInFile(String fileAsString) throws com.github.antlrjavaparser.ParseException, UnsupportedEncodingException, IOException{
		com.github.antlrjavaparser.api.CompilationUnit cu = com.github.antlrjavaparser.JavaParser.parse(new ByteArrayInputStream(fileAsString.getBytes("UTF-8")));
		List<com.github.antlrjavaparser.api.body.BodyDeclaration> result = new ArrayList<com.github.antlrjavaparser.api.body.BodyDeclaration>();
		
		for (Iterator iterator = cu.getTypes().iterator(); iterator.hasNext();) {
			com.github.antlrjavaparser.api.body.TypeDeclaration type = (com.github.antlrjavaparser.api.body.TypeDeclaration) iterator.next();
			if(type.getMembers() != null){
				for (Iterator iterator2 = type.getMembers().iterator(); iterator2
						.hasNext();) {
					com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
					if(member instanceof com.github.antlrjavaparser.api.body.ConstructorDeclaration){
						result.add(member);
					}
					else if(member instanceof com.github.antlrjavaparser.api.body.MethodDeclaration){
						result.add(member);
					}
					else if(member instanceof com.github.antlrjavaparser.api.body.AnnotationMemberDeclaration){
						result.add(member);
					}
					else if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
						QDoxUtil.getAllDeclaredJavaMethodsInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
					}
					else {
						continue;
					}
				}
			}
		}	
		Comparator<com.github.antlrjavaparser.api.body.BodyDeclaration> COMPARE_BY_LINE_NUMBER = new Comparator<com.github.antlrjavaparser.api.body.BodyDeclaration>() {
			public int compare(com.github.antlrjavaparser.api.body.BodyDeclaration one, com.github.antlrjavaparser.api.body.BodyDeclaration other) {
				return Integer.valueOf(one.getBeginLine()).compareTo(other.getBeginLine());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static void getAllDeclaredAnnotatedJavaMethodsInFileHelper(com.github.antlrjavaparser.api.body.TypeDeclaration type, 
			List<com.github.antlrjavaparser.api.body.BodyDeclaration> result){
		if(type.getMembers() != null){
			for (Iterator iterator2 = type.getMembers().iterator(); iterator2
					.hasNext();) {
				com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
				if(member instanceof com.github.antlrjavaparser.api.body.ConstructorDeclaration){
					if(member.getAnnotations().size() > 0){
						result.add(member);
					}
				}
				else if(member instanceof com.github.antlrjavaparser.api.body.MethodDeclaration){
					if(member.getAnnotations().size() > 0){
						result.add(member);
					}
				}
				else if(member instanceof com.github.antlrjavaparser.api.body.AnnotationMemberDeclaration){
					if(member.getAnnotations().size() > 0){
						result.add(member);
					}
				}
				else if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
					QDoxUtil.getAllDeclaredAnnotatedJavaMethodsInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
				}
				else {
					continue;
				}
			}
		}
	}
	
	private static void getAllDeclaredJavaMethodsInFileHelper(com.github.antlrjavaparser.api.body.TypeDeclaration type, 
			List<com.github.antlrjavaparser.api.body.BodyDeclaration> result){
		if(type.getMembers() != null){
			for (Iterator iterator2 = type.getMembers().iterator(); iterator2
					.hasNext();) {
				com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
				if(member instanceof com.github.antlrjavaparser.api.body.ConstructorDeclaration){
					result.add(member);
				}
				else if(member instanceof com.github.antlrjavaparser.api.body.MethodDeclaration){
					result.add(member);
				}
				else if(member instanceof com.github.antlrjavaparser.api.body.AnnotationMemberDeclaration){
					result.add(member);
				}
				else if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
					QDoxUtil.getAllDeclaredJavaMethodsInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
				}
				else {
					continue;
				}
			}
		}
	}
	
	private static List<com.github.antlrjavaparser.api.body.TypeDeclaration> getAllDeclaredJavaTypesInFile(String fileAsString) throws com.github.antlrjavaparser.ParseException, UnsupportedEncodingException, IOException{
		List<com.github.antlrjavaparser.api.body.TypeDeclaration> result = new ArrayList<com.github.antlrjavaparser.api.body.TypeDeclaration>();
		com.github.antlrjavaparser.api.CompilationUnit cu = com.github.antlrjavaparser.JavaParser.parse(new ByteArrayInputStream(fileAsString.getBytes("UTF-8")));
		for (Iterator iterator = cu.getTypes().iterator(); iterator.hasNext();) {
			com.github.antlrjavaparser.api.body.TypeDeclaration type = (com.github.antlrjavaparser.api.body.TypeDeclaration) iterator.next();
			result.add(type);
			QDoxUtil.getAllDeclaredJavaTypesInFileHelper(type, result);
		}	
		Comparator<com.github.antlrjavaparser.api.body.TypeDeclaration> COMPARE_BY_LINE_NUMBER = new Comparator<com.github.antlrjavaparser.api.body.TypeDeclaration>() {
			public int compare(com.github.antlrjavaparser.api.body.TypeDeclaration one, com.github.antlrjavaparser.api.body.TypeDeclaration other) {
				return Integer.valueOf(one.getBeginLine()).compareTo(other.getBeginLine());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<com.github.antlrjavaparser.api.body.TypeDeclaration> getAllDeclaredAnnotatedJavaTypesInFile(String fileAsString) throws com.github.antlrjavaparser.ParseException, UnsupportedEncodingException, IOException{
		List<com.github.antlrjavaparser.api.body.TypeDeclaration> result = new ArrayList<com.github.antlrjavaparser.api.body.TypeDeclaration>();
		com.github.antlrjavaparser.api.CompilationUnit cu = com.github.antlrjavaparser.JavaParser.parse(new ByteArrayInputStream(fileAsString.getBytes("UTF-8")));
		for (Iterator iterator = cu.getTypes().iterator(); iterator.hasNext();) {
			com.github.antlrjavaparser.api.body.TypeDeclaration type = (com.github.antlrjavaparser.api.body.TypeDeclaration) iterator.next();
			if(type.getAnnotations() != null && type.getAnnotations().size() > 0){
				result.add(type);
			}
			QDoxUtil.getAllDeclaredAnnotatedJavaTypesInFileHelper(type, result);
		}	
		Comparator<com.github.antlrjavaparser.api.body.TypeDeclaration> COMPARE_BY_LINE_NUMBER = new Comparator<com.github.antlrjavaparser.api.body.TypeDeclaration>() {
			public int compare(com.github.antlrjavaparser.api.body.TypeDeclaration one, com.github.antlrjavaparser.api.body.TypeDeclaration other) {
				return Integer.valueOf(one.getBeginLine()).compareTo(other.getBeginLine());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static void getAllDeclaredJavaTypesInFileHelper(com.github.antlrjavaparser.api.body.TypeDeclaration type, 
			List<com.github.antlrjavaparser.api.body.TypeDeclaration> result){
		if(type.getMembers() != null){
			for (Iterator iterator2 = type.getMembers().iterator(); iterator2
					.hasNext();) {
				com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
				if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
					result.add((com.github.antlrjavaparser.api.body.TypeDeclaration)member);
					QDoxUtil.getAllDeclaredJavaTypesInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
				}
				else {
					continue;
				}
			}
		}
	}
	
	private static void getAllDeclaredAnnotatedJavaTypesInFileHelper(com.github.antlrjavaparser.api.body.TypeDeclaration type, 
			List<com.github.antlrjavaparser.api.body.TypeDeclaration> result){
		if(type.getMembers() != null){
			for (Iterator iterator2 = type.getMembers().iterator(); iterator2
					.hasNext();) {
				com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
				if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
					if(member.getAnnotations() != null && member.getAnnotations().size() > 0){
						result.add((com.github.antlrjavaparser.api.body.TypeDeclaration)member);
					}
					QDoxUtil.getAllDeclaredAnnotatedJavaTypesInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
				}
				else {
					continue;
				}
			}
		}
	}
	
//	private static List<JavaClass> getAllDeclaredJavaTypesInFile(JavaDocBuilder qDoxFile){
//		List<JavaClass> result = new ArrayList<JavaClass>();
//		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();
//
//		for (int i = 0; i < javaDeclTypes.length; i++) {
//			JavaClass currentJavaDeclType = javaDeclTypes[i];
//			result.add(currentJavaDeclType);
//		}
//		Comparator<JavaClass> COMPARE_BY_LINE_NUMBER = new Comparator<JavaClass>() {
//			public int compare(JavaClass one, JavaClass other) {
//				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
//			}
//		};
//		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
//		return result;
//	}
	
	public static void processJavaTypesWithRightFullQualifiedNames(JavaDocBuilder qDoxFile){
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(!currentJavaDeclType.getFullyQualifiedName().contains(".")){
				String imports [] = currentJavaDeclType.getSource().getImports();
				for (int j = 0; j < imports.length; j++) {
					String currentImport = imports[j];
					if(currentImport.endsWith(currentJavaDeclType.getFullyQualifiedName())){
						String pn = currentImport.substring(0, currentImport.lastIndexOf("."));
						currentJavaDeclType.setName(pn+"."+currentJavaDeclType.getFullyQualifiedName());
						break;
					}
				}
			}
		}
	}
	
	private static List<JavaClass> getAllDeclaredAnnotatedJavaTypesInFile(JavaDocBuilder qDoxFile){
		List<JavaClass> result = new ArrayList<JavaClass>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(currentJavaDeclType.getAnnotations().length > 0){
				result.add(currentJavaDeclType);
			}
		}
		Comparator<JavaClass> COMPARE_BY_LINE_NUMBER = new Comparator<JavaClass>() {
			public int compare(JavaClass one, JavaClass other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<JavaClass> getAllDeclaredJavaTypesInFile(JavaDocBuilder qDoxFile){
		List<JavaClass> result = new ArrayList<JavaClass>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			result.add(currentJavaDeclType);
			if(!AspectUtil.getInstance().getAllJavaTypeDeclarations().contains(currentJavaDeclType)){
				AspectUtil.getInstance().appendAllJavaTypeDecl(currentJavaDeclType);
			}
		}
		Comparator<JavaClass> COMPARE_BY_LINE_NUMBER = new Comparator<JavaClass>() {
			public int compare(JavaClass one, JavaClass other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<JavaClass> getAllDeclaredJavaTypesWithJavaDocTagsInFile(JavaDocBuilder qDoxFile){
		List<JavaClass> result = new ArrayList<JavaClass>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(currentJavaDeclType.getTags().length > 0){
				result.add(currentJavaDeclType);
			}
		}
		Comparator<JavaClass> COMPARE_BY_LINE_NUMBER = new Comparator<JavaClass>() {
			public int compare(JavaClass one, JavaClass other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}

	private static List<JavaField> getAllDeclaredJavaFieldsInFile(JavaDocBuilder qDoxFile){
		List<JavaField> result = new ArrayList<JavaField>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			JavaField [] javaDeclFields = currentJavaDeclType.getFields();
			for (int j = 0; j < javaDeclFields.length; j++) {
				JavaField currentJavaDeclField = javaDeclFields[j];
				result.add(currentJavaDeclField);
			}
		}
		Comparator<JavaField> COMPARE_BY_LINE_NUMBER = new Comparator<JavaField>() {
			public int compare(JavaField one, JavaField other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<JavaField> getAllDeclaredAnnotatedJavaFieldsInFile(JavaDocBuilder qDoxFile){
		List<JavaField> result = new ArrayList<JavaField>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			JavaField [] javaDeclFields = currentJavaDeclType.getFields();
			for (int j = 0; j < javaDeclFields.length; j++) {
				JavaField currentJavaDeclField = javaDeclFields[j];
				if(currentJavaDeclField.getAnnotations().length > 0){
					result.add(currentJavaDeclField);
				}
			}
		}
		Comparator<JavaField> COMPARE_BY_LINE_NUMBER = new Comparator<JavaField>() {
			public int compare(JavaField one, JavaField other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static void getAllDeclaredAnnotatedITDJavaFieldsInFile(JavaDocBuilder qDoxFile){
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(isPrincipalEnclosingTypeDeclNotExposeOnly(currentJavaDeclType)){
				JavaField [] javaDeclFields = currentJavaDeclType.getFields();
				for (int j2 = 0; j2 < javaDeclFields.length; j2++) {
					JavaField currentJavaDeclField = javaDeclFields[j2];
					if(!currentJavaDeclField.getParentClass().isInner()) {
						if(!isFieldDeclPartial(currentJavaDeclField)) {
							AspectUtil.getInstance().appendItdFieldDecl(currentJavaDeclField);
						}
					}
					else if(!isTypeDeclOpen(currentJavaDeclField.getParentClass())) {
						if(!isFieldDeclPartial(currentJavaDeclField)) {
							AspectUtil.getInstance().appendItdFieldDecl(currentJavaDeclField);
						}
					}
				}
			}
		}
	}
	
	private static void getAllDeclaredJavaFieldsInJMLFile(JavaDocBuilder qDoxFile){
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			JavaField [] javaDeclFields = currentJavaDeclType.getFields();
			for (int j2 = 0; j2 < javaDeclFields.length; j2++) {
				JavaField currentJavaDeclField = javaDeclFields[j2];
				if(currentJavaDeclField.getSource().getFile().getAbsolutePath().endsWith(".jml")){
					AspectUtil.getInstance().appendJMLFileFieldDecl(currentJavaDeclField);
				}
			}
		}
	}
	
	private static List<com.github.antlrjavaparser.api.body.InitializerDeclaration> getAllDeclaredJavaInitializersInFile(String fileAsString) 
			throws com.github.antlrjavaparser.ParseException, UnsupportedEncodingException, IOException{
		com.github.antlrjavaparser.api.CompilationUnit cu = com.github.antlrjavaparser.JavaParser.parse(new ByteArrayInputStream(fileAsString.getBytes("UTF-8")));
		List<com.github.antlrjavaparser.api.body.InitializerDeclaration> result = new ArrayList<com.github.antlrjavaparser.api.body.InitializerDeclaration>();
		for (Iterator iterator = cu.getTypes().iterator(); iterator.hasNext();) {
			com.github.antlrjavaparser.api.body.TypeDeclaration type = (com.github.antlrjavaparser.api.body.TypeDeclaration) iterator.next();
			if(type.getMembers() != null){
				for (Iterator iterator2 = type.getMembers().iterator(); iterator2
						.hasNext();) {
					com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
					if(member instanceof com.github.antlrjavaparser.api.body.InitializerDeclaration){
						result.add((com.github.antlrjavaparser.api.body.InitializerDeclaration)member);
					}
					else if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
						QDoxUtil.getAllDeclaredJavaInitializersInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
					}
					else {
						continue;
					}
				}
			}
		}	
		Comparator<com.github.antlrjavaparser.api.body.InitializerDeclaration> COMPARE_BY_LINE_NUMBER = new Comparator<com.github.antlrjavaparser.api.body.InitializerDeclaration>() {
			public int compare(com.github.antlrjavaparser.api.body.InitializerDeclaration one, com.github.antlrjavaparser.api.body.InitializerDeclaration other) {
				return Integer.valueOf(one.getBeginLine()).compareTo(other.getBeginLine());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<com.github.antlrjavaparser.api.body.EnumDeclaration> getAllDeclaredJavaEnumConstantsInFile(String fileAsString) 
			throws com.github.antlrjavaparser.ParseException, UnsupportedEncodingException, IOException{
		com.github.antlrjavaparser.api.CompilationUnit cu = com.github.antlrjavaparser.JavaParser.parse(new ByteArrayInputStream(fileAsString.getBytes("UTF-8")));
		List<com.github.antlrjavaparser.api.body.EnumDeclaration> result = new ArrayList<com.github.antlrjavaparser.api.body.EnumDeclaration>();
		for (Iterator iterator = cu.getTypes().iterator(); iterator.hasNext();) {
			com.github.antlrjavaparser.api.body.TypeDeclaration type = (com.github.antlrjavaparser.api.body.TypeDeclaration) iterator.next();
			if(type instanceof com.github.antlrjavaparser.api.body.EnumDeclaration){
				result.add((com.github.antlrjavaparser.api.body.EnumDeclaration)type);
			}
			if(type.getMembers() != null){
				for (Iterator iterator2 = type.getMembers().iterator(); iterator2
						.hasNext();) {
					com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
					if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
						QDoxUtil.getAllDeclaredJavaEnumConstantsInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
					}
					else {
						continue;
					}
				}
			}
		}	
		Comparator<com.github.antlrjavaparser.api.body.EnumDeclaration> COMPARE_BY_LINE_NUMBER = new Comparator<com.github.antlrjavaparser.api.body.EnumDeclaration>() {
			public int compare(com.github.antlrjavaparser.api.body.EnumDeclaration one, com.github.antlrjavaparser.api.body.EnumDeclaration other) {
				return Integer.valueOf(one.getBeginLine()).compareTo(other.getBeginLine());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<com.github.antlrjavaparser.api.body.FieldDeclaration> getAllDeclaredJavaFieldsInFile(String fileAsString) 
			throws com.github.antlrjavaparser.ParseException, UnsupportedEncodingException, IOException{
		com.github.antlrjavaparser.api.CompilationUnit cu = com.github.antlrjavaparser.JavaParser.parse(new ByteArrayInputStream(fileAsString.getBytes("UTF-8")));
		List<com.github.antlrjavaparser.api.body.FieldDeclaration> result = new ArrayList<com.github.antlrjavaparser.api.body.FieldDeclaration>();
		for (Iterator iterator = cu.getTypes().iterator(); iterator.hasNext();) {
			com.github.antlrjavaparser.api.body.TypeDeclaration type = (com.github.antlrjavaparser.api.body.TypeDeclaration) iterator.next();
			if(type.getMembers() != null){
				for (Iterator iterator2 = type.getMembers().iterator(); iterator2
						.hasNext();) {
					com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
					if(member instanceof com.github.antlrjavaparser.api.body.FieldDeclaration){
						result.add((com.github.antlrjavaparser.api.body.FieldDeclaration)member);
					}
					else if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
						QDoxUtil.getAllDeclaredJavaFieldsInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
					}
					else {
						continue;
					}
				}
			}
		}	
		Comparator<com.github.antlrjavaparser.api.body.FieldDeclaration> COMPARE_BY_LINE_NUMBER = new Comparator<com.github.antlrjavaparser.api.body.FieldDeclaration>() {
			public int compare(com.github.antlrjavaparser.api.body.FieldDeclaration one, com.github.antlrjavaparser.api.body.FieldDeclaration other) {
				return Integer.valueOf(one.getBeginLine()).compareTo(other.getBeginLine());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<com.github.antlrjavaparser.api.body.FieldDeclaration> getAllDeclaredAnnotatedJavaFieldsInFile(String fileAsString) 
			throws com.github.antlrjavaparser.ParseException, UnsupportedEncodingException, IOException{
		com.github.antlrjavaparser.api.CompilationUnit cu = com.github.antlrjavaparser.JavaParser.parse(new ByteArrayInputStream(fileAsString.getBytes("UTF-8")));
		List<com.github.antlrjavaparser.api.body.FieldDeclaration> result = new ArrayList<com.github.antlrjavaparser.api.body.FieldDeclaration>();
		for (Iterator iterator = cu.getTypes().iterator(); iterator.hasNext();) {
			com.github.antlrjavaparser.api.body.TypeDeclaration type = (com.github.antlrjavaparser.api.body.TypeDeclaration) iterator.next();
			if(type.getMembers() != null){
				for (Iterator iterator2 = type.getMembers().iterator(); iterator2
						.hasNext();) {
					com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
					if(member instanceof com.github.antlrjavaparser.api.body.FieldDeclaration){
						if(((com.github.antlrjavaparser.api.body.FieldDeclaration)member).getAnnotations().size() > 0){
							result.add((com.github.antlrjavaparser.api.body.FieldDeclaration)member);
						}
					}
					else if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
						QDoxUtil.getAllDeclaredAnnotatedJavaFieldsInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
					}
					else {
						continue;
					}
				}
			}
		}	
		Comparator<com.github.antlrjavaparser.api.body.FieldDeclaration> COMPARE_BY_LINE_NUMBER = new Comparator<com.github.antlrjavaparser.api.body.FieldDeclaration>() {
			public int compare(com.github.antlrjavaparser.api.body.FieldDeclaration one, com.github.antlrjavaparser.api.body.FieldDeclaration other) {
				return Integer.valueOf(one.getBeginLine()).compareTo(other.getBeginLine());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static void getAllDeclaredJavaInitializersInFileHelper(com.github.antlrjavaparser.api.body.TypeDeclaration type, 
			List<com.github.antlrjavaparser.api.body.InitializerDeclaration> result){
		if(type.getMembers() != null){
			for (Iterator iterator2 = type.getMembers().iterator(); iterator2
					.hasNext();) {
				com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
				if(member instanceof com.github.antlrjavaparser.api.body.InitializerDeclaration){
					result.add((com.github.antlrjavaparser.api.body.InitializerDeclaration)member);
				}
				else if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
					QDoxUtil.getAllDeclaredJavaInitializersInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
				}
				else {
					continue;
				}
			}
		}
	}
	
	private static void getAllDeclaredJavaEnumConstantsInFileHelper(com.github.antlrjavaparser.api.body.TypeDeclaration type, 
			List<com.github.antlrjavaparser.api.body.EnumDeclaration> result){
		if(type instanceof com.github.antlrjavaparser.api.body.EnumDeclaration){
			result.add((com.github.antlrjavaparser.api.body.EnumDeclaration)type);
		}
		if(type.getMembers() != null){
			for (Iterator iterator2 = type.getMembers().iterator(); iterator2
					.hasNext();) {
				com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
				 if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
					QDoxUtil.getAllDeclaredJavaEnumConstantsInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
				}
				else {
					continue;
				}
			}
		}
	}
	
	private static void getAllDeclaredJavaFieldsInFileHelper(com.github.antlrjavaparser.api.body.TypeDeclaration type, 
			List<com.github.antlrjavaparser.api.body.FieldDeclaration> result){
		if(type.getMembers() != null){
			for (Iterator iterator2 = type.getMembers().iterator(); iterator2
					.hasNext();) {
				com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
				if(member instanceof com.github.antlrjavaparser.api.body.FieldDeclaration){
					result.add((com.github.antlrjavaparser.api.body.FieldDeclaration)member);
				}
				else if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
					QDoxUtil.getAllDeclaredJavaFieldsInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
				}
				else {
					continue;
				}
			}
		}
	}
	
	private static void getAllDeclaredAnnotatedJavaFieldsInFileHelper(com.github.antlrjavaparser.api.body.TypeDeclaration type, 
			List<com.github.antlrjavaparser.api.body.FieldDeclaration> result){
		if(type.getMembers() != null){
			for (Iterator iterator2 = type.getMembers().iterator(); iterator2
					.hasNext();) {
				com.github.antlrjavaparser.api.body.BodyDeclaration member = (com.github.antlrjavaparser.api.body.BodyDeclaration) iterator2.next();
				if(member instanceof com.github.antlrjavaparser.api.body.FieldDeclaration){
					if(((com.github.antlrjavaparser.api.body.FieldDeclaration)member).getAnnotations().size() > 0){
						result.add((com.github.antlrjavaparser.api.body.FieldDeclaration)member);
					}
				}
				else if (member instanceof com.github.antlrjavaparser.api.body.TypeDeclaration){
					QDoxUtil.getAllDeclaredAnnotatedJavaFieldsInFileHelper((com.github.antlrjavaparser.api.body.TypeDeclaration)member, result);
				}
				else {
					continue;
				}
			}
		}
	}
	
	private static List<JavaField> getAllDeclaredJavaFieldsWithJavaDocTagsInFile(JavaDocBuilder qDoxFile){
		List<JavaField> result = new ArrayList<JavaField>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			JavaField [] javaDeclFields = currentJavaDeclType.getFields();
			for (int j = 0; j < javaDeclFields.length; j++) {
				JavaField currentJavaDeclField = javaDeclFields[j];
				if(currentJavaDeclField.getTags().length > 0){
					result.add(currentJavaDeclField);
				}
			}
		}
		Comparator<JavaField> COMPARE_BY_LINE_NUMBER = new Comparator<JavaField>() {
			public int compare(JavaField one, JavaField other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<JavaMethod> getAllDeclaredAnnotatedJavaMethodsInFile(JavaDocBuilder qDoxFile){
		List<JavaMethod> result = new ArrayList<JavaMethod>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			JavaMethod [] javaDeclMethods = currentJavaDeclType.getMethods();
			for (int j = 0; j < javaDeclMethods.length; j++) {
				JavaMethod currentJavaDeclMethod = javaDeclMethods[j];
				if(currentJavaDeclMethod.getAnnotations().length > 0){
					result.add(currentJavaDeclMethod);
				}
			}
		}
		Comparator<JavaMethod> COMPARE_BY_LINE_NUMBER = new Comparator<JavaMethod>() {
			public int compare(JavaMethod one, JavaMethod other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<DocletTag> getAllJavaDocTagsInAJavaTypedDecl(JavaClass type){
		List<DocletTag> result = new ArrayList<DocletTag>();
		if(type.getTags().length > 0){
			for (int i = 0; i < type.getTags().length; i++) {
				result.add(type.getTags()[i]);
			}
		}
		Comparator<DocletTag> COMPARE_BY_LINE_NUMBER = new Comparator<DocletTag>() {
			public int compare(DocletTag one, DocletTag other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<DocletTag> getAllJavaDocTagsInAJavaFieldDecl(JavaField field){
		List<DocletTag> result = new ArrayList<DocletTag>();
		if(field.getTags().length > 0){
			for (int i = 0; i < field.getTags().length; i++) {
				result.add(field.getTags()[i]);
			}
		}
		Comparator<DocletTag> COMPARE_BY_LINE_NUMBER = new Comparator<DocletTag>() {
			public int compare(DocletTag one, DocletTag other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}
	
	private static List<DocletTag> getAllJavaDocTagsInFile(JavaDocBuilder qDoxFile){
		List<DocletTag> result = new ArrayList<DocletTag>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			
			// type javadoc tags
			if(currentJavaDeclType.getTags().length > 0){
				for (int k = 0; k < currentJavaDeclType.getTags().length; k++) {
					result.add(currentJavaDeclType.getTags()[k]);
				}
			}
			// field javadoc tags
			JavaField [] javaDeclFields = currentJavaDeclType.getFields();
			for (int j = 0; j < javaDeclFields.length; j++) {
				JavaField currentJavaDeclField = javaDeclFields[j];
				if(currentJavaDeclField.getTags().length > 0){
					for (int k = 0; k < currentJavaDeclField.getTags().length; k++) {
						result.add(currentJavaDeclField.getTags()[k]);
					}
				}
			}
			// method javadoc tags
			JavaMethod [] javaDeclMethods = currentJavaDeclType.getMethods();
			for (int j = 0; j < javaDeclMethods.length; j++) {
				JavaMethod currentJavaDeclMethod = javaDeclMethods[j];
				if(currentJavaDeclMethod.getTags().length > 0){
					for (int k = 0; k < currentJavaDeclMethod.getTags().length; k++) {
						result.add(currentJavaDeclMethod.getTags()[k]);
					}
				}
			}
		}
		Comparator<DocletTag> COMPARE_BY_LINE_NUMBER = new Comparator<DocletTag>() {
			public int compare(DocletTag one, DocletTag other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}

	private static List<JavaField> getADeclaredJavaEnumerationFieldsInEnumType(JavaClass enumType){
		List<JavaField> result = new ArrayList<JavaField>();

		JavaField [] javaDeclFields = enumType.getFields();
		for (int j = 0; j < javaDeclFields.length; j++) {
			JavaField currentJavaDeclField = javaDeclFields[j];
			if(enumType.getLineNumber() == currentJavaDeclField.getLineNumber()){
				result.add(currentJavaDeclField);
			}
		}
		Comparator<JavaField> COMPARE_BY_LINE_NUMBER = new Comparator<JavaField>() {
			public int compare(JavaField one, JavaField other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}

	private static String getADeclaredJavaEnumerationFieldsInEnumTypeAsErasure(JavaClass enumType){
		List<JavaField> javaEnumFields = new ArrayList<JavaField>();
		StringBuffer code = new StringBuffer("");

		JavaField [] javaDeclFields = enumType.getFields();
		for (int j = 0; j < javaDeclFields.length; j++) {
			JavaField currentJavaDeclField = javaDeclFields[j];
			if(enumType.getLineNumber() == currentJavaDeclField.getLineNumber()){
				javaEnumFields.add(currentJavaDeclField);
			}
		}
		Comparator<JavaField> COMPARE_BY_LINE_NUMBER = new Comparator<JavaField>() {
			public int compare(JavaField one, JavaField other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(javaEnumFields, COMPARE_BY_LINE_NUMBER);

		for (Iterator<JavaField> iterator = javaEnumFields.iterator(); iterator.hasNext();) {
			JavaField javaEnumField = iterator.next();
			code.append("public /*@ nullable @*/static final "+javaEnumField.getType().getFullyQualifiedName()+" "+javaEnumField.getName()+" = null;");
		}

		return code.toString();
	}

	private static List<JavaClass> getAllDeclaredJavaEnumTypesInFile(JavaDocBuilder qDoxFile){
		List<JavaClass> result = new ArrayList<JavaClass>();
		JavaClass [] javaDeclTypes  = qDoxFile.getClasses();

		for (int i = 0; i < javaDeclTypes.length; i++) {
			JavaClass currentJavaDeclType = javaDeclTypes[i];
			if(currentJavaDeclType.isEnum()){
				result.add(currentJavaDeclType);
			}
		}
		Comparator<JavaClass> COMPARE_BY_LINE_NUMBER = new Comparator<JavaClass>() {
			public int compare(JavaClass one, JavaClass other) {
				return Integer.valueOf(one.getLineNumber()).compareTo(other.getLineNumber());
			}
		};
		Collections.sort(result, COMPARE_BY_LINE_NUMBER);
		return result;
	}

	private static HashMap<String, String> getGenericAndErasureTypeParametersForField(JavaClass clazz, JavaField field){
		HashMap<String, String> result = new HashMap<String, String>();
		// adding the current type generic type parameters
//		System.out.println("--> "+clazz.getFullyQualifiedName());
//		System.out.println("field = "+field.getName());
		if(clazz.getTypeParameters().length > 0){
			TypeVariable [] typeVariables = clazz.getTypeParameters();
			for (int i = 0; i < typeVariables.length; i++) {
				TypeVariable currentTypeVariable = typeVariables[i];
				String genericType = currentTypeVariable.getName();
				String erasureType = currentTypeVariable.getValue();
				if(erasureType.equals("")){
					erasureType = "java.lang.Object";
				}
//				System.out.println("genericType<added> = "+genericType);
//				System.out.println("erasureType<added> = "+erasureType);
				result.put(genericType, erasureType);
			}
		}
		if(clazz.isInner()){
			while(clazz.getParentClass() != null){
				clazz = clazz.getParentClass();
			}
			if(clazz.getTypeParameters().length > 0){
				TypeVariable [] typeVariables = clazz.getTypeParameters();
				for (int i = 0; i < typeVariables.length; i++) {
					TypeVariable currentTypeVariable = typeVariables[i];
					String genericType = currentTypeVariable.getName();
					String erasureType = currentTypeVariable.getValue();
					if(erasureType.equals("")){
						erasureType = "java.lang.Object";
					}
//					System.out.println("genericType<added> = "+genericType);
//					System.out.println("erasureType<added> = "+erasureType);
					result.put(genericType, erasureType);
				}
			}
		}
		// adding the current method generic type parameters
		String genericType = field.getType().getFullyQualifiedName();
		String erasureType = field.getType().getValue();
		if(erasureType.equals("")){
			erasureType = "java.lang.Object";
		}
		else if(!erasureType.equals(genericType)){
			result.put(genericType, erasureType);
		}
		return result;
	}
	
	private static HashMap<String, String> getGenericAndErasureTypeParametersForMethod(JavaClass clazz, JavaMethod meth){
		HashMap<String, String> result = new HashMap<String, String>();
		// adding the current type generic type parameters
		if(clazz.getTypeParameters().length > 0){
			TypeVariable [] typeVariables = clazz.getTypeParameters();
			for (int i = 0; i < typeVariables.length; i++) {
				TypeVariable currentTypeVariable = typeVariables[i];
				String genericType = currentTypeVariable.getName();
				String erasureType = currentTypeVariable.getValue();
				if(erasureType.equals("")){
					erasureType = "java.lang.Object";
				}
				result.put(genericType, erasureType);
			}
		}
		if(clazz.isInner()){
			while(clazz.getParentClass() != null){
				clazz = clazz.getParentClass();
			}
			if(clazz.getTypeParameters().length > 0){
				TypeVariable [] typeVariables = clazz.getTypeParameters();
				for (int i = 0; i < typeVariables.length; i++) {
					TypeVariable currentTypeVariable = typeVariables[i];
					String genericType = currentTypeVariable.getName();
					String erasureType = currentTypeVariable.getValue();
					if(erasureType.equals("")){
						erasureType = "java.lang.Object";
					}
					result.put(genericType, erasureType);
				}
			}
		}
		// adding the current method generic type parameters
		if(meth.getTypeParameters().length > 0){
			TypeVariable [] typeVariables = meth.getTypeParameters();
			for (int i = 0; i < typeVariables.length; i++) {
				TypeVariable currentTypeVariable = typeVariables[i];
				String genericType = currentTypeVariable.getName();
				String erasureType = currentTypeVariable.getValue();
				if(erasureType.equals("")){
					erasureType = "java.lang.Object";
				}
				if(!erasureType.equals(genericType)){
					result.put(genericType, erasureType);
				}
			}
		}
		// adding some inner types that crashes compilation... [[[hemr]]]
		if(meth.getParameters().length > 0){
			result.put("java.util.Map.Entry", "Entry");
			result.put("Map.Entry", "Entry");
		}
		return result;
	}

	private static String stripMethBodies(
			String fileAsString, List<JavaMethod> javaDeclMeths,
			List<String> fileMeths) {
		int indexMeth = 0;
		for (Iterator<String> iterator = fileMeths.iterator(); iterator.hasNext();) {
			String returnTypeDefaultValue = "";
			String currentMeth = iterator.next();
			String currentMethStrippedBody = "";
			currentMethStrippedBody = currentMeth;
			JavaMethod javaMeth = javaDeclMeths.get(indexMeth);
			
			if(((javaMeth.getReturnType() != null) && !(javaMeth.isConstructor()))){
				if(!QDoxUtil.defaultValue(javaMeth.getReturnType().getFullyQualifiedName()).equals("")){
					returnTypeDefaultValue = "return "+QDoxUtil.defaultValue(javaMeth.getReturnType().getFullyQualifiedName())+";";
				}
			}
			if(!currentMeth.equals("")){
				int openBraceIndex = currentMeth.indexOf('{');
				if(openBraceIndex != -1){
					String tmp = "";
					StringBuffer linesToAppend = new StringBuffer("");
					tmp = currentMeth.substring(openBraceIndex+1);
					int newLineCount = QDoxUtil.getLineNumbersQtd(tmp);
					linesToAppend.append("{");
					linesToAppend.append(QDoxUtil.getNewLinesCaracter(newLineCount));
					linesToAppend.append(returnTypeDefaultValue);
					linesToAppend.append("}");
					String body = currentMeth.substring(openBraceIndex, currentMeth.lastIndexOf("}")+1);
					currentMethStrippedBody = currentMeth.replace(body, linesToAppend.toString());
					fileAsString = StringUtils.replaceOnce(fileAsString, currentMeth, currentMethStrippedBody);
				}
			}
			indexMeth++;
		}
		return fileAsString;
	}
	
	public static boolean balancedDiamond(String s) {
		Stack<Character> stack  = new Stack<Character>();
		for(int i = 0; i < s.length(); i++) {
			char c = s.charAt(i);
			if(c == '<') {
				stack.push(c);
			}else if(c == '>') {
				if(stack.isEmpty()) return false;
				if(stack.pop() != '<') return false;
			}
		}
		return stack.isEmpty();
	}
	
	public static boolean balancedParenthensies(String s) {
		Stack<Character> stack  = new Stack<Character>();
		for(int i = 0; i < s.length(); i++) {
			char c = s.charAt(i);
			if(c == '[' || c == '(' || c == '{' ) {
				stack.push(c);
			}else if(c == ']') {
				if(stack.isEmpty()) return false;
				if(stack.pop() != '[') return false;

			}else if(c == ')') {
				if(stack.isEmpty()) return false;
				if(stack.pop() != '(') return false;

			}else if(c == '}') {
				if(stack.isEmpty()) return false;
				if(stack.pop() != '{') return false;
			}

		}
		return stack.isEmpty();
	}
	
	public static int getAnnotationDeclPerLine(String line){
		int result = 0;
		for (int i = 0; i < line.toCharArray().length; i++) {
			if(line.toCharArray()[i] == '@'){
				result++;
			}
		}
		return result;
	}

	private static String removeMethParameterAnnotations(String currentMeth){
		// processing @NonNull or @Nullable annotations as standard JML annotation comments --- [[[hemr]]]
		Pattern jmlNullityPattern = Pattern.compile("(@NonNull(\\b)(\\s)*(\\((\\s)*\\))?)|(@Nullable(\\b)(\\s)*(\\((\\s)*\\))?)");
		Matcher jmlNullityMatcher = jmlNullityPattern.matcher(currentMeth);
		while(jmlNullityMatcher.find()){
			if(jmlNullityMatcher.group().contains("@NonNull")){
				currentMeth = currentMeth.replace(jmlNullityMatcher.group(), "/*@ non_null @*/");
			}
			else{
				currentMeth = currentMeth.replace(jmlNullityMatcher.group(), "/*@ nullable @*/");
			}
			
		}
		// masking @ within standar JML comments to avoid removal like an annotation --- [[[hemr]]]
		currentMeth = currentMeth.replace("/*@", "/*#").replace("@*/", "#*/");
//		System.out.println("check: "+currentMeth);
		char [] currentMethChars = null;
		Pattern jmlAnnoPattern = Pattern.compile("@[\\w]+");
		Matcher jmlAnnoMatcher = jmlAnnoPattern.matcher(currentMeth);
		while(jmlAnnoMatcher.find()){
			currentMethChars = currentMeth.toCharArray();
			int startAnno = currentMeth.indexOf(jmlAnnoMatcher.group());
			// mount @anno
			StringBuffer anno = new StringBuffer(jmlAnnoMatcher.group());
			for (int i = (startAnno+jmlAnnoMatcher.group().length()); i < currentMethChars.length; i++) {
				if(Character.isLetter(currentMethChars[i])){
					anno.append(currentMethChars[i]);
					break;
				}
				else{
					anno.append(currentMethChars[i]);
				}
			}
			if(anno.toString().contains("(")){
				boolean canCheckBalance = false;
				anno = new StringBuffer(jmlAnnoMatcher.group());
				for (int i = (startAnno+jmlAnnoMatcher.group().length()); i < currentMethChars.length; i++) {
					if(canCheckBalance && balancedParenthensies(anno.toString())){
//						System.out.println("match! -> "+anno.toString());
						break;
					}
					if(Character.isLetter(currentMethChars[i]) && !canCheckBalance){
						canCheckBalance = true;
						if(canCheckBalance && balancedParenthensies(anno.toString())){
//							System.out.println("match! -> "+anno.toString());
							break;
						}
						anno.append(currentMethChars[i]);
					}
					else{
						anno.append(currentMethChars[i]);
					}
					
				}
//				System.out.println("extracted anno: "+anno.toString());
//				System.out.println("currentMeth<before> = "+currentMeth);
				currentMeth = StringUtils.remove(currentMeth, anno.toString());
//				System.out.println("currentMeth<after> = "+currentMeth);
				
			}
			else{
//				System.out.println("extracted anno: "+anno.toString());
				anno = new StringBuffer(jmlAnnoMatcher.group());
				currentMeth = StringUtils.remove(currentMeth, anno.toString());
			}
		}
		jmlAnnoMatcher.reset();
		// reestablishing the standar JML comments, if any --- [[[hemr]]]
		currentMeth = currentMeth.replace("/*#", "/*@").replace("#*/", "@*/");
//		System.out.println("final meth: "+currentMeth);
		return currentMeth;
	}
	
	private static String getFileMethDeclsProcessed(
			String fileAsString, List<JavaMethod> javaDeclMeths,
			List<String> fileMeths, File file) throws PositionedError {
		int indexMeth = 0;
		for (Iterator<String> iterator = fileMeths.iterator(); iterator.hasNext();) {
			String currentMethErasure = "";
			String currentMethErasureTmp = "";
			String currentMeth = iterator.next();
			JavaMethod javaMeth = javaDeclMeths.get(indexMeth);
			
//			two versions of parameter annotation removal
//			if(javaMeth.getAnnotations().length > 0){
				// processing method annotations including parameter ones
//				String currentMethAnnoProcessed = QDoxUtil.removeNonJMLAnnotationParameters(currentMeth);
//			    removing method parameters --- [[[hemr]]]
				String currentMethAnnoProcessed = QDoxUtil.removeMethParameterAnnotations(currentMeth);
				fileAsString = StringUtils.replaceOnce(fileAsString, currentMeth, currentMethAnnoProcessed);
				currentMeth = currentMethAnnoProcessed;
//			}
			currentMethErasureTmp = currentMeth;
			JavaClass onwerType = javaMeth.getParentClass();
			HashMap<String, String> genericAndErasureTypeParameters = QDoxUtil.getGenericAndErasureTypeParametersForMethod(onwerType, javaMeth);
			Set<String> genericTypes = genericAndErasureTypeParameters.keySet();

//			System.out.println("Meth = "+currentMeth);
			for (Iterator<String> iterator2 = genericTypes.iterator(); iterator2
					.hasNext();) {
				String genericType = iterator2.next();
				String erasureType = genericAndErasureTypeParameters.get(genericType);
				currentMethErasureTmp = currentMethErasureTmp.replaceAll("(\\b)"+genericType+"(\\b)", erasureType);
				currentMethErasure = currentMethErasureTmp;
			}
//			System.out.println("currentMethErasure = "+currentMethErasure);
			indexMeth++;
			
			if(!currentMethErasure.equals("")){
				// means that we have generic type parameters 
				if(currentMethErasure.contains("<") && currentMethErasure.contains(">")){
					while(currentMethErasure.contains("<")){
						char [] currentMethErasureChars = currentMethErasure.toCharArray();
						StringBuffer currentMethErasureSb = new StringBuffer(""); 
						boolean canAdd = false;
						for (int i = 0; i < currentMethErasureChars.length; i++) {
							if(currentMethErasureChars[i] == '<'){
								canAdd = true;
							}
							if(canAdd){
								currentMethErasureSb.append(currentMethErasureChars[i]);
							}
							if(currentMethErasureSb.length() > 0 && balancedDiamond(currentMethErasureSb.toString())){
								break;
							}
						}
//						System.out.println("currentMethErasureSb.toString() = "+currentMethErasureSb.toString());
//						System.out.println("currentMethErasure<before> = "+currentMethErasure);
						currentMethErasure = currentMethErasure.replace(currentMethErasureSb.toString(), "");
//						System.out.println("currentMethErasure<after> = "+currentMethErasure);
					}
				}
				// handling crosscutting specs
				JavaClass ownerType = javaMeth.getParentClass();
				QDoxUtil.handlingCrosscutSpecs(file, javaMeth, currentMethErasure, ownerType);
				fileAsString = StringUtils.replaceOnce(fileAsString, currentMeth, currentMethErasure);
			}
			else{
				String currentMethTmp = currentMeth;
				// means that we have generic type parameters 
				if(currentMethTmp.contains("<") && currentMethTmp.contains(">")){
					while(currentMethTmp.contains("<")){
						char [] currentMethChars = currentMethTmp.toCharArray();
						StringBuffer currentMethSb = new StringBuffer(""); 
						boolean canAdd = false;
						for (int i = 0; i < currentMethChars.length; i++) {
							if(currentMethChars[i] == '<'){
								canAdd = true;
							}
							if(canAdd){
								currentMethSb.append(currentMethChars[i]);
							}
							if(currentMethSb.length() > 0 && balancedDiamond(currentMethSb.toString())){
								break;
							}
						}
						currentMethTmp = currentMethTmp.replace(currentMethSb.toString(), "");
					}
					// handling crosscutting specs
					JavaClass ownerType = javaMeth.getParentClass();
					QDoxUtil.handlingCrosscutSpecs(file, javaMeth, currentMeth, ownerType);
					fileAsString = StringUtils.replaceOnce(fileAsString, currentMeth, currentMethTmp);
				}
				else {
					// means that we have a method in a .jml file
					// handling crosscutting specs
					JavaClass ownerType = javaMeth.getParentClass();
					QDoxUtil.handlingCrosscutSpecs(file, javaMeth, currentMeth, ownerType);
				}
			}
		}
		return fileAsString;
	}
	
	public static String getSignatureErasure(String signature, JavaMethod javaMeth){
		JavaClass onwerType = javaMeth.getParentClass();
		HashMap<String, String> genericAndErasureTypeParameters = QDoxUtil.getGenericAndErasureTypeParametersForMethod(onwerType, javaMeth);
		Set<String> genericTypes = genericAndErasureTypeParameters.keySet();
		for (Iterator<String> iterator2 = genericTypes.iterator(); iterator2
				.hasNext();) {
			String genericType = iterator2.next();
			String erasureType = genericAndErasureTypeParameters.get(genericType);
			signature = signature.replaceAll("(\\b)"+genericType+"(\\b)", erasureType);
		}
		return signature;
	}

	private static void handlingCrosscutSpecs(File file, JavaMethod javaMeth,
			String currentMeth, JavaClass ownerType) throws PositionedError {
		boolean isAspect = false;
		for (int i = 0; i < ownerType.getAnnotations().length; i++) {
			if(ownerType.getAnnotations()[i].toString().startsWith("@org.aspectj.lang.annotation.Aspect")){
				isAspect = true;
				break;
			}
		}
		
		 if(isAspect){
			// means that we can have advice and pointcuts but is not a part of the crosscutting contract specification
			if(Main.aRacOptions != null){
				if(!Main.aRacOptions.noContractChecking()){
//					AspectUtil.getInstance().appendAspectDecl(ownerType);
					JavaDocBuilder qDoxJavaMethodParserForCrosscutSpecs = new JavaDocBuilder();
					qDoxJavaMethodParserForCrosscutSpecs.addSource(new StringReader("public class JMethodParser{\n"+currentMeth+"\n}"));
					// adding Java imports
					for (int i = 0; i < ownerType.getSource().getImports().length; i++) {
						qDoxJavaMethodParserForCrosscutSpecs.getSources()[0].addImport(ownerType.getSource().getImports()[i]);
					}
					JavaClass parsedClassTmp = qDoxJavaMethodParserForCrosscutSpecs.getClassByName("JMethodParser");
					if(parsedClassTmp.getMethods().length == 1){
						JavaMethod parsedMethTmp = parsedClassTmp.getMethods()[0];
						parsedMethTmp.setAnnotations(javaMeth.getAnnotations());
						parsedMethTmp.setParentClass(ownerType);
						// handling advice specs --- [[[hemr]
						boolean isToInclude = false;
						for (int i = 0; i < parsedMethTmp.getAnnotations().length; i++) {
							if(parsedMethTmp.getAnnotations()[i].toString().startsWith("@org.aspectj.lang.annotation.A")){
								isToInclude = true;
							}
							if(parsedMethTmp.getAnnotations()[i].toString().startsWith("@org.aspectj.lang.annotation.Before")){
								isToInclude = true;
							}
						}
						if(isToInclude){
							AspectUtil.getInstance().appendAspectAdvice(parsedMethTmp);
						}
						// handling crosscutting contract specs for advice --- [[[hemr]
						isToInclude = false;
						boolean isPC = false;
						for (int i = 0; i < parsedMethTmp.getAnnotations().length; i++) {
							if(parsedMethTmp.getAnnotations()[i].toString().startsWith("@org.aspectj.lang.annotation.Pointcut")){
								isPC = true;
							}
						}
						isToInclude = isPC;
						if(isToInclude){
							AspectUtil.getInstance().appendCrosscutSpecMeth(parsedMethTmp);
						}
					}
				}
			}
		}
		else {
			// is to handle crosscut specs?
			if(Main.aRacOptions != null){
				if(!Main.aRacOptions.nocrosscutting()){
					JavaDocBuilder qDoxJavaMethodParserForCrosscutSpecs = new JavaDocBuilder();
					try {
						qDoxJavaMethodParserForCrosscutSpecs.addSource(new StringReader("public class JMethodParser{\n"+currentMeth+"\n}"));
					} catch (java.lang.ArrayIndexOutOfBoundsException e) {
						// ignore for now...
					}
					catch(com.thoughtworks.qdox.parser.ParseException e){
						try {
							if(currentMeth.trim().startsWith("*")){
								currentMeth = "/** \n"+currentMeth.trim();
							}
							else if(currentMeth.trim().startsWith("@*/")) {
								currentMeth = currentMeth.replace("@*/", "");
							}
							if (currentMeth.trim().endsWith("}") && !currentMeth.trim().contains("{")) {
								currentMeth = currentMeth.replace("}", "");
							}
							qDoxJavaMethodParserForCrosscutSpecs.addSource(new StringReader("public class JMethodParser{\n"+currentMeth+"\n}"));
						} catch (Exception e1) {
//							System.out.println("owner = "+javaMeth.getParentClass().getName());
//							System.out.println("file = "+file.getAbsolutePath());
//							System.out.println("currentMeth = "+currentMeth);
							e1.printStackTrace();
						}
					}
					catch (Exception e) {
//						System.out.println("owner = "+javaMeth.getParentClass().getName());
//						System.out.println("file = "+file.getAbsolutePath());
//						System.out.println("currentMeth = "+currentMeth);
						e.printStackTrace();
					}
					// adding Java imports
					for (int i = 0; i < ownerType.getSource().getImports().length; i++) {
						try {
							qDoxJavaMethodParserForCrosscutSpecs.getSources()[0].addImport(ownerType.getSource().getImports()[i]);
						} catch (Exception e) {
//							System.out.println(currentMeth);
//							e.printStackTrace();
						}
					}
					JavaClass parsedClassTmp = qDoxJavaMethodParserForCrosscutSpecs.getClassByName("JMethodParser");
					if(parsedClassTmp.getMethods().length == 1){
						JavaMethod parsedMethTmp = parsedClassTmp.getMethods()[0];
						parsedMethTmp.setAnnotations(javaMeth.getAnnotations());
						parsedMethTmp.setParentClass(ownerType);
						boolean isToInclude = false;
						boolean isPC = false;
						for (int i = 0; i < parsedMethTmp.getAnnotations().length; i++) {
							if(parsedMethTmp.getAnnotations()[i].toString().startsWith("@org.aspectj.lang.annotation.Pointcut")){
								isPC = true;
								break;
							}
						}
						isToInclude = isPC;
						if(isToInclude){
							AspectUtil.getInstance().appendCrosscutSpecMeth(parsedMethTmp);
						}
					}
				}
			}
		}
	}
	
	public static boolean isXCS(JavaMethod javaMeth){
		boolean isXCS = false;
		boolean isPC = false;
		boolean isAroundAdv = false;
		boolean result = false;
		if(javaMeth != null){
			for (int i = 0; i < javaMeth.getAnnotations().length; i++) {
				if((javaMeth.getAnnotations()[i].toString().startsWith("@org.aspectj.lang.annotation.Pointcut"))){
					isPC = true;
				}
				if((javaMeth.getAnnotations()[i].toString().startsWith("@org.aspectjml.lang.annotation.Around"))){
					isAroundAdv = true;
				}
			}
			
			for (int i = 0; i < javaMeth.getAnnotations().length; i++) {
				if((javaMeth.getAnnotations()[i].toString().startsWith("@org.aspectjml.lang.annotation.XCS"))){
					// only an around advice in AspectJML can be used to implement/specify crosscutting contracts --- [[[hemr]]]
					if(isAroundAdv){
						isXCS = true;
					}
					break;
				}
			}
			result = (isPC || isXCS);
		}
		return result;
	}
	
	public static boolean isAnnotatedWithXCS(JavaMethod javaMeth){
		boolean isXCS = false;
		if(javaMeth != null){
			for (int i = 0; i < javaMeth.getAnnotations().length; i++) {
				if((javaMeth.getAnnotations()[i].toString().startsWith("@org.aspectjml.lang.annotation.XCS"))){
					isXCS = true;
					break;
				}
			}
		}
		return isXCS;
	}
	
	public static void processAllInnersFromJML(JmlTypeDeclaration typeDecl, ArrayList innersResult){
		ArrayList inners = typeDecl.inners();
		if(inners.size() > 0){
			for (Iterator iterator = inners.iterator(); iterator.hasNext();) {
				JmlTypeDeclaration currentInnerType = (JmlTypeDeclaration) iterator.next();
				if(!innersResult.contains(currentInnerType)){
					innersResult.add(currentInnerType);
				}
				if(currentInnerType.inners() != null &&  currentInnerType.inners().size() > 0){
					processAllInnersFromJML(currentInnerType, innersResult);
				}
			}
		}
	}
	
	public static ArrayList getAllInnersFromJML(JmlTypeDeclaration typeDecl){
		ArrayList inners = typeDecl.inners();
		ArrayList innersResult = new ArrayList();
		if(inners.size() > 0){
			for (Iterator iterator = inners.iterator(); iterator.hasNext();) {
				JmlTypeDeclaration currentInnerType = (JmlTypeDeclaration) iterator.next();
				if(!innersResult.contains(currentInnerType)){
					innersResult.add(currentInnerType);
				}
				if(currentInnerType.inners() != null &&  currentInnerType.inners().size() > 0){
					processAllInnersFromJML(currentInnerType, innersResult);
				}
			}
		}
		return inners;
	}
	
	public static String getJMLModelMethodBodyFromFile(File file, int lineNumberStart) throws IOException {
		String jmlModelMethodBody = "";
		StringBuffer bufferedFile = new StringBuffer("");
		BufferedReader buffer = new BufferedReader(new FileReader( file ));
		String line = buffer.readLine();
		int fileLineNumber = 1;
		boolean canStart = false;
		while (line != null) {
			if(lineNumberStart == fileLineNumber){ canStart = true;}
			if(canStart){
				bufferedFile.append(processJMLModelMethodLine(line));
				bufferedFile.append("\n");
				JavaDocBuilder qDoxJavaMethodParser = new JavaDocBuilder();
				try {
					qDoxJavaMethodParser.addSource(new StringReader("public class JMethodParser{\n"+bufferedFile.toString()+"\n}"));
					if(qDoxJavaMethodParser.getClassByName("JMethodParser").getMethods().length == 1){
						// if match the wanted JML model method
						jmlModelMethodBody = "{\n"+ qDoxJavaMethodParser.getSources()[0].getClasses()[0].getMethods()[0].getSourceCode()+" \n}";
						break;
					}
				} catch (com.thoughtworks.qdox.parser.ParseException e) {
					// continue if it is not a valid method or constructor... [[[hemr]]]
				}
			}
			fileLineNumber++;
			line = buffer.readLine();
		}
		buffer.close();
		return jmlModelMethodBody;
	}
	
	private static String processJMLModelMethodLine(String line) {
		String lineProcessed = line.replace("/*@", "").replace("@*/", "").replace("//@", "")
				.replace("\\readonly", ""); // general rule
		// this is usually for the first line
		if((lineProcessed.contains("model") || lineProcessed.contains("pure")) 
				&& lineProcessed.contains("(")) {
			lineProcessed = lineProcessed.replaceFirst("model", "").replaceFirst("pure", "");
		}
		if(lineProcessed.contains("@")) {
			lineProcessed = lineProcessed.replaceFirst("@", "");
		}
		return lineProcessed;
	}
	
	private static boolean canAdd(String lineNumber, List<String> lineNumbers){
		boolean result = true;
		for (Iterator iterator = lineNumbers.iterator(); iterator.hasNext();) {
			String lineNumberTmp = (String) iterator.next();
			if(lineNumberTmp.equals(lineNumber)){
				result = false;
				break;
			}
			
		}
		return result;
	}
	
	private static List<String> getEnumConstantLineNumbersPerEnumDecl(com.github.antlrjavaparser.api.body.EnumDeclaration enumType){
		List<String> lineNumbers = new ArrayList<String>();
		for (Iterator iterator = enumType.getEntries().iterator(); iterator.hasNext();) {
			com.github.antlrjavaparser.api.body.EnumConstantDeclaration enumConstDecl = (com.github.antlrjavaparser.api.body.EnumConstantDeclaration) iterator.next();
			if(canAdd(enumConstDecl.getBeginLine()+"", lineNumbers)){
				lineNumbers.add(enumConstDecl.getBeginLine()+"");
			}
			if(canAdd(enumConstDecl.getEndLine()+"", lineNumbers)){
				lineNumbers.add(enumConstDecl.getEndLine()+"");
			}
		}
		return lineNumbers;
	}

	private static String getFileEnumTypeErasureProcessingAsString(
			StringBuffer bufferedFile, String fileAsString,
			List<JavaClass> javaDeclEnumTypes) throws IOException {
		BufferedReader buffer = null;
		String line = "";
		List<com.github.antlrjavaparser.api.body.EnumDeclaration> javaDeclEnumTypes2 = QDoxUtil.getAllDeclaredJavaEnumConstantsInFile(fileAsString);
		fileAsString = fileAsString.replace("enum ", "final class ");

		//		handling enumeration fields
		for (int i = 0; i < javaDeclEnumTypes.size(); i++) {
			JavaClass enumType = javaDeclEnumTypes.get(i);
			com.github.antlrjavaparser.api.body.EnumDeclaration enumType2 = javaDeclEnumTypes2.get(i);
			List<JavaField> enumTypeFields = QDoxUtil.getADeclaredJavaEnumerationFieldsInEnumType(enumType);
			if(enumTypeFields.size() > 0){
				String enumFieldsConverted = QDoxUtil.getADeclaredJavaEnumerationFieldsInEnumTypeAsErasure(enumType);
				List<String> listEnumConstLineNumbers = getEnumConstantLineNumbersPerEnumDecl(enumType2);
				buffer = new BufferedReader(new StringReader(fileAsString));
				bufferedFile.delete(0, (bufferedFile.length()-1)); // resetting buffer
				line = buffer.readLine();
				int lineStart = enumType2.getBeginLine();
				int lineEnd = enumType2.getEndLine();
				int lineNumber = 1;
				String lastLine = "";
				String innerEnumType = "";
				while (line != null) {
					if(listEnumConstLineNumbers.get(listEnumConstLineNumbers.size() - 1).equals(lineNumber+"")){
						if((listEnumConstLineNumbers.size() == 1) && line.contains("class") && 
								line.contains("{") && line.contains("}")){
							int lb = line.indexOf('{');
							int rb = line.lastIndexOf('}');
							String lineProcessed = line.substring((lb+1), rb);
							lastLine = "/* "+lineProcessed+" */#";
//							fileAsString = StringUtils.replaceOnce(fileAsString, lineProcessed, lastLine);
							if(enumType.isInner()) {
								innerEnumType = " final class "+ enumType.getName()+"{}";
							}
							fileAsString = changeFileLine(fileAsString, lineNumber, innerEnumType+lastLine);
						}
						else{
							if(line.endsWith("};") || line.endsWith("},")){
								lastLine = line+" */#";
							}
							else{
								lastLine = "/* "+line+" */#";
							}
//							fileAsString = StringUtils.replaceOnce(fileAsString, line, lastLine);
							fileAsString = changeFileLine(fileAsString, lineNumber, lastLine);
						}
					}
					else if(listEnumConstLineNumbers.contains(lineNumber+"")){
						String lineProcessed = null;
						if(line.contains("{")){
							lineProcessed = "/* "+line;
						}
						else if(line.endsWith("};") || line.endsWith("},")){
							lineProcessed = line+" */";
						}
						else{
							if(line.endsWith("*/")){
								lineProcessed = "/* "+line;
							}
							else{
								lineProcessed = "/* "+line+" */";
							}
						}
//						fileAsString = StringUtils.replaceOnce(fileAsString, line, lineProcessed);
						fileAsString = changeFileLine(fileAsString, lineNumber, lineProcessed);
					}
					if(lineNumber >= lineStart && lineNumber <= lineEnd){
						if(line.contains("{@inheritDoc}")){
							fileAsString = changeFileLine(fileAsString, lineNumber, "");
						}
					}
					line = buffer.readLine();
					lineNumber++;
				}
				buffer.close();
				fileAsString = StringUtils.replaceOnce(fileAsString, lastLine, lastLine.replace("#", enumFieldsConverted));
				// issue about the internal use of values() --- [[[hemr]]]
				String implicityMeth = "public static /*@ pure @*/ "+enumType.getName()+"[] values() {return null;}";
				fileAsString = QDoxUtil.appendAtBeginFileLine(fileAsString, lineEnd, implicityMeth);
			}
		}
		return fileAsString;
	}
	
	private static String handleConstructorDeclInJMLFileMode(StringBuffer bufferedFile, String fileAsString,
			List<JavaMethod> javaDeclMeths) throws IOException{
		
		BufferedReader buffer;
		String line;
		for (Iterator<JavaMethod> iterator = javaDeclMeths.iterator(); iterator
				.hasNext();) {
			JavaMethod javaMethod = iterator.next();
			if(!javaMethod.isConstructor()){
				continue;
			}
			int lineNumber = javaMethod.getLineNumber();
			buffer = new BufferedReader(new StringReader(fileAsString));
			line = buffer.readLine();
			int fileLineNumber = 1;
			boolean canStart = false;
			while (line != null) {
				if(lineNumber == fileLineNumber){ canStart = true;}
				if(canStart){
					bufferedFile.append(line);
					bufferedFile.append("\n");
					JavaDocBuilder qDoxJavaMethodParser = new JavaDocBuilder();
					try {
						qDoxJavaMethodParser.addSource(new StringReader("public class JMethodParser{\n"+bufferedFile.toString()+"\n}"));
						if(qDoxJavaMethodParser.getClassByName("JMethodParser").getMethods().length == 1){
							// if match the wanted contructor
							if(line.contains(";")){
								String lineProcessed = line.replace(";", "");
								fileAsString = StringUtils.replaceOnce(fileAsString, bufferedFile.toString(),bufferedFile.toString().replace(line, lineProcessed)+"{}");
							}
							bufferedFile.delete(0, (bufferedFile.length()-1));
							break;
						}
					} catch (com.thoughtworks.qdox.parser.ParseException e) {
						// continue if it is not a valid method or constructor... [[[hemr]]]
					}
				}
				line = buffer.readLine();
				fileLineNumber++;
			}
		}
		
		return fileAsString;
	}
	
	private static List<String> getAllJavaMethodDeclLexicallyInFile(StringBuffer bufferedFile, String fileAsString,
			List<JavaMethod> javaDeclMeths, List<com.github.antlrjavaparser.api.body.BodyDeclaration> members) throws IOException {
		List<String> fileMeths = new ArrayList<String>();
		for (int i = 0; i < javaDeclMeths.size(); i++) {
			com.github.antlrjavaparser.api.body.BodyDeclaration member = members.get(i);
			JavaMethod currentJavaMethod = javaDeclMeths.get(i);
			int lineNumberStart = 0;
			int lineNumberEnd = 0;
			if(member.getAnnotations().size() > 0){
				if((member.getAnnotations().get(member.getAnnotations().size()-1).getEndLine() + 1) < member.getBeginLine()){
					lineNumberStart = member.getBeginLine();
				}
				else{
					lineNumberStart = member.getAnnotations().get(member.getAnnotations().size()-1).getEndLine() + 1;
//					lineNumberStart = member.getAnnotations().get(member.getAnnotations().size()-1).getEndLine();
				}
				if(member.getBeginComments() != null){
					lineNumberStart += member.getBeginComments().size();
				}
				lineNumberEnd = member.getEndLine();
			}
			else{
				lineNumberStart = member.getBeginLine();
				lineNumberEnd = member.getEndLine();
			}
			
//			System.out.println("javaMethod<name> = "+javaDeclMeths.get(i).getName());
//			System.out.println("javaMethod = "+javaDeclMeths.get(i).getLineNumber());
//			System.out.println("lineNumberStart = "+lineNumberStart); 
//			System.out.println("lineNumberEnd = "+lineNumberEnd); 
			
			BufferedReader buffer = null;
			String line = "";
			String meth = "";
			bufferedFile = new StringBuffer("");
			buffer = new BufferedReader(new StringReader(fileAsString));
			line = buffer.readLine();
			int fileLineNumber = 1;
			while (line != null) {
				if((fileLineNumber >= lineNumberStart) && (fileLineNumber <= lineNumberEnd)){
					bufferedFile.append(line);
					bufferedFile.append("\n");
				}
				line = buffer.readLine();
				fileLineNumber++;
			}
			buffer.close();
			meth = bufferedFile.toString();
			JavaDocBuilder qDoxJavaMethodITD = new JavaDocBuilder();
			try {
				qDoxJavaMethodITD.addSource(new StringReader("public class JMethodParser{\n"+meth+"\n}"));
				if(!currentJavaMethod.isConstructor() && 
						qDoxJavaMethodITD.getSources()[0].getClasses()[0].getMethods()[0].getReturnType() == null	){
					lineNumberStart--;
					buffer = null;
					line = "";
					meth = "";
					bufferedFile = new StringBuffer("");
					buffer = new BufferedReader(new StringReader(fileAsString));
					line = buffer.readLine();
					fileLineNumber = 1;
					while (line != null) {
						if((fileLineNumber >= lineNumberStart) && (fileLineNumber <= lineNumberEnd)){
							bufferedFile.append(line);
							bufferedFile.append("\n");
						}
						line = buffer.readLine();
						fileLineNumber++;
					}
					buffer.close();
					meth = bufferedFile.toString();
				}
				fileMeths.add(meth);
			} catch (Exception e) {
				lineNumberStart--;
				buffer = null;
				line = "";
				meth = "";
				bufferedFile = new StringBuffer("");
				buffer = new BufferedReader(new StringReader(fileAsString));
				line = buffer.readLine();
				fileLineNumber = 1;
				while (line != null) {
					if((fileLineNumber >= lineNumberStart) && (fileLineNumber <= lineNumberEnd)){
						bufferedFile.append(line);
						bufferedFile.append("\n");
					}
					line = buffer.readLine();
					fileLineNumber++;
				}
				buffer.close();
				meth = bufferedFile.toString();
				fileMeths.add(meth);
			}
//			System.out.println("***********");
//			System.out.println(meth);
//			System.out.println("***********");
			QDoxUtil.resolveOriginalITDMethDeclWithNameProcessed(currentJavaMethod, meth);
			
			// saving all methods processed in File
			AspectUtil.getInstance().appendAllJavaMethDecl(currentJavaMethod); 
			
		}
		
		return fileMeths;
	}
	
	private static void resolveOriginalITDTypeDeclWithNameProcessed(JavaClass itdJavaClass, String type){
		if(itdJavaClass.getParentClass() != null){
			String oldName = itdJavaClass.getName();
			String newName = itdJavaClass.getParentClass().getFullyQualifiedName().replace("$", ".")+"."+itdJavaClass.getName();
			if(type.contains("*/")){
				type = type.replaceAll("/\\*(.)*\\*/", "");
			}
			type = type.replaceFirst("(\\b)"+oldName+"(\\b)", newName.replace("$", "\\$"));
			
			if(itdJavaClass.getOriginalClassDecl() == null){
				itdJavaClass.setOriginalClassDecl(type);
			}
		}
	}
	
	private static void resolveOriginalITDFieldDeclWithNameProcessed(JavaField itdJavaField, String field){
		String oldName = itdJavaField.getName();
		String newName = itdJavaField.getParentClass().getFullyQualifiedName().replace("$", ".")+"."+itdJavaField.getName();
		if(field.contains("*/")){
			field = field.replaceAll("/\\*(.)*\\*/", "");
		}
		field = field.replaceFirst("(\\b)"+oldName+"(\\b)", newName.replace("$", "\\$"));
		if(field.contains("=")){
			field = field.replaceFirst("\\=[\\s]+\\;", ";");
		}
		if(itdJavaField.getOriginalFieldDecl() == null){
			itdJavaField.setOriginalFieldDecl(field);
		}
	}
	
	private static void resolveOriginalITDMethDeclWithNameProcessed(JavaMethod itdJavaMethod, String meth){
		String abstractt = "";
		String oldName = itdJavaMethod.getName();
		String newName = "";
		
		// resolving ITD newName
		if(itdJavaMethod.isConstructor()){
			if(AspectUtil.getInstance().isAspectJMLAppendLikeMethod(itdJavaMethod)){
				if(AspectUtil.getInstance().isAspectJMLAppendBefore(itdJavaMethod)){
					newName = "void "+itdJavaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+".new$AppendBefore";
				}
				else if(AspectUtil.getInstance().isAspectJMLAppendAfter(itdJavaMethod)){
					newName = "void "+itdJavaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+".new$AppendAfter";
				}
				else if(AspectUtil.getInstance().isAspectJMLAppendAfterReturning(itdJavaMethod)){
					newName = "void "+itdJavaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+".new$AppendAfterReturning";
				}
				else if(AspectUtil.getInstance().isAspectJMLAppendAfterThrowing(itdJavaMethod)){
					newName = "void "+itdJavaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+".new$AppendAfterThrowing";
				}
				newName += "$"+itdJavaMethod.getSource().getFile().getName().replace(".", "_");
			}
			else if(AspectUtil.getInstance().isAspectJMLReviseMethod(itdJavaMethod)){
				newName = "void "+itdJavaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+".new$Revise";
				newName += "$"+itdJavaMethod.getSource().getFile().getName().replace(".", "_");
			}
			else{
				newName = itdJavaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+".new";
			}
		}
		else{
			if(AspectUtil.getInstance().isAspectJMLAppendLikeMethod(itdJavaMethod)){
				if(AspectUtil.getInstance().isAspectJMLAppendBefore(itdJavaMethod)){
					newName = itdJavaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+"."+itdJavaMethod.getName()+"$AppendBefore";
				}
				else if(AspectUtil.getInstance().isAspectJMLAppendAfter(itdJavaMethod)){
					newName = itdJavaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+"."+itdJavaMethod.getName()+"$AppendAfter";
				}
				else if(AspectUtil.getInstance().isAspectJMLAppendAfterReturning(itdJavaMethod)){
					newName = itdJavaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+"."+itdJavaMethod.getName()+"$AppendAfterReturning";
				}
				else if(AspectUtil.getInstance().isAspectJMLAppendAfterThrowing(itdJavaMethod)){
					newName = itdJavaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+"."+itdJavaMethod.getName()+"$AppendAfterThrowing";
				}
				newName += "$"+itdJavaMethod.getSource().getFile().getName().replace(".", "_");
			}
			else if(AspectUtil.getInstance().isAspectJMLReviseMethod(itdJavaMethod)){
				newName = itdJavaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+"."+itdJavaMethod.getName()+"$Revise";
				newName += "$"+itdJavaMethod.getSource().getFile().getName().replace(".", "_");
			}
			else{
				newName = itdJavaMethod.getParentClass().getFullyQualifiedName().replace("$", ".")+"."+itdJavaMethod.getName();
			}

		}
		
		// adding abstract if needed
		if(itdJavaMethod.getCodeBlock().charAt(itdJavaMethod.getCodeBlock().length()-2) == ';'){
			if(itdJavaMethod.getParentClass().isInterface()){
				if(!itdJavaMethod.isAbstract()){
					abstractt = "abstract ";
				}	
			}
		}
		if(meth.contains("*/")){
			meth = meth.replaceAll("/\\*(.)*\\*/", "");
		}
		meth = abstractt+meth.replaceFirst("(\\b)"+oldName+"(\\b)", newName.replace("$", "\\$"));
//		meth = QDoxUtil.getITDMethNameProcessed(meth, oldName, newName, abstractt);
		if(AspectUtil.getInstance().isAspectJMLAppendLikeMethod(itdJavaMethod)){
			String methPattern = "";
			if(meth.contains("return")){
				methPattern = meth.replaceAll("//(.)*(\\b)return(\\b)", "").replaceAll("/\\*(.)*(\\b)return(\\b)(.)*\\*/", "");
			}
			Pattern pattern = Pattern.compile("(\\b)return(\\b)");
			Matcher matcher = pattern.matcher(methPattern);
			if(!matcher.find()){
				String returnTypeDefaultValue = "";
				if(((itdJavaMethod.getReturnType() != null) && !(itdJavaMethod.isConstructor()))){
					if(!QDoxUtil.defaultValue(itdJavaMethod.getReturnType().getFullyQualifiedName()).equals("")){
						returnTypeDefaultValue = "return "+QDoxUtil.defaultValue(itdJavaMethod.getReturnType().getFullyQualifiedName())+";";
						meth = meth.replace("}", "$AJML$2016$");
						int index = meth.lastIndexOf("$AJML$2016$");
						meth = meth.substring(0, index);
						meth = meth.replace("$AJML$2016$", "}");
						meth += returnTypeDefaultValue + "}\n";
					}
				}
			}
		}
		
		// resolving new ITD name
		if(itdJavaMethod.getCodeBlock().charAt(itdJavaMethod.getCodeBlock().length()-2) == ';'){
			if(itdJavaMethod.getParentClass().isInterface()){
				if(!itdJavaMethod.isAbstract()){
					if(itdJavaMethod.getOriginalMethodDecl() == null){
						itdJavaMethod.setOriginalMethodDecl(meth);
					}
				}	
				else if(itdJavaMethod.isAbstract()){
					if(itdJavaMethod.getOriginalMethodDecl() == null){
						itdJavaMethod.setOriginalMethodDecl(meth);
					}
				}
			}
		}
		else{
			if(itdJavaMethod.getOriginalMethodDecl() == null){
				itdJavaMethod.setOriginalMethodDecl(meth);
			}
		}
	}

	private static String getMethodAnnotationAreaCommentedProcessedWithJML(
			JavaMethod annotatedJavaMethod, StringBuffer annotationAreaCommented) {
		String annotationAreaCommentedStr = annotationAreaCommented.toString();
		Annotation[] methAnnotation = annotatedJavaMethod.getAnnotations();

		for (int i = 0; i < methAnnotation.length; i++) {
			Annotation currentAnnotation = methAnnotation[i];
			// handling single JML annotations
			if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Requires")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "requires_redundantly" : "requires") : "requires");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			else if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Assignable")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "assignable_redundantly" : "assignable") : "assignable");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			else if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Ensures")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "ensures_redundantly" : "ensures") : "ensures");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			else if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Signals")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "signals_redundantly" : "signals") : "signals");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			else if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.SignalsOnly")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "signals_only_redundantly" : "signals_only") : "signals_only");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			else if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Callable")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "callable_redundantly" : "callable") : "callable");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			else if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Accessible")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "accessible_redundantly" : "accessible") : "accessible");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			else if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Duration")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "duration_redundantly" : "duration") : "duration");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			else if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Captures")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "captures_redundantly" : "captures") : "captures");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			else if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.WorkingSpace")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String clauseKind = (currentAnnotation.getNamedParameterMap().containsKey("redundantly")? 
							(currentAnnotation.getProperty("redundantly").toString().equals("true")? "working_space_redundantly" : "working_space") : "working_space");
					String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					String clause = "/*@ "+clauseKind+" "+pred+"; @*/";
					annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
				}
			}
			else if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.Also")){
				String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
				String clause = "/*@ also @*/";
				annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
			}
			else if(currentAnnotation.getType().getFullyQualifiedName().equals("org.aspectjml.lang.annotation.SpecCase")){
				if(currentAnnotation.getNamedParameterMap().containsKey("value")){
					String pred = QDoxUtil.preparePred(currentAnnotation.getProperty("value").toString());
					if(!pred.equals("")){
						String preLineNumber = "/*"+currentAnnotation.getLineNumber()+"*/";
						String clause = "/*@ "+pred+" @*/";
						annotationAreaCommentedStr = annotationAreaCommentedStr.replace(preLineNumber, clause+preLineNumber);
					}
				}
			}
		}
		return annotationAreaCommentedStr;
	}
	
	private static int getLineNumbersQtd(String tmp) {
		int newLineCount = 0;
		
		BufferedReader buffer = new BufferedReader(new StringReader(tmp));
		String line = null;
		try {
			line = buffer.readLine();
			while (line != null) {
				line = buffer.readLine();
				newLineCount++;
			}
			buffer.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return newLineCount-1;
	}
	
	private static String getNewLinesCaracter(int numberOfNewLines){
		StringBuffer result = null;
		result = new StringBuffer("");
		for (int i = 0; i < numberOfNewLines; i++) {
			result.append("\n");
		}
		return result.toString();
	}

	private static String preparePred(String clause){
		clause = clause.replace("\\\"", "#");
		Pattern pattern = Pattern.compile("(\"[^\"]*\")*");
		Matcher patterhMatcher = pattern.matcher(clause);
		StringBuffer clauseParsed = new StringBuffer("");
		while(patterhMatcher.find()){
			clauseParsed.append(patterhMatcher.group());
		}
//		return clauseParsed.toString().replace("\"", "").replace("#dq", "\"").replace("#sq", "\'").replace("#", "\\");
		return clauseParsed.toString().replace("\"", "").replace("#", "\"").replace("@", "\\");
	}

	public static String defaultValue(String type) {
		String result = "";

		if(!type.equals("void")){
			if(type.equals("boolean")){
				result = "false";
			}
			else if(type.equals("byte")){
				result = "(byte) 0";
			}
			else if(type.equals("short")){
				result = "(short) 0";
			}
			else if(type.equals("int")){
				result = "0";
			}
			else if(type.equals("long")){
				result = "0L";
			}
			else if(type.equals("float")){
				result = "0.0f";
			}
			else if(type.equals("double")){
				result = "0.0d";
			}
			else if(type.equals("char")){
				result = "(char) 0";
			}
			else{
				result = "null";
			}
		}
		return result;
	}
	
	private static String getMapEntryImport(boolean java8OrLater, String line){
		if(java8OrLater){
			if(line.contains("import") && line.contains("java.util.Map")){
				Pattern pattern = Pattern.compile("(\\b)import(\\b)");
				Matcher matcher = pattern.matcher(line);
				if(matcher.find()){
					line += "import java.util.Map.Entry;";
				}
			}
		}
		return line;
	}
}
