/*
 * Copyright (C) 2008-2013 Universidade Federal de Pernambuco and 
 * University of Central Florida
 *
 * This file is part of AJML
 *
 * AJML is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * AJML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AJML; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: ReflectUtil.java,v 1.0 2013/05/17 11:37:06 henriquerebelo Exp $
 */

package org.aspectjml.util;

import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;

public class ReflectUtil {
	
	public static boolean isAssignableFrom(String advisingType, String advisedType, String urlClassLoader){
		boolean result = false;

		try {
			File file = new File(urlClassLoader);
			//convert the file to URL format
			URL url = file.toURI().toURL();
			URL[] urls = new URL[]{url};
			ClassLoader cl = new URLClassLoader(urls);
			result = cl.loadClass(advisingType).isAssignableFrom(cl.loadClass(advisedType));
		} catch (ClassNotFoundException e) {
//			e.printStackTrace();
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return result;
	}
	
	public static String getPackageNameFromClassLoader(String advisingType, String urlClassLoader){
		String result = "";
		
		try {
			File file = new File(urlClassLoader);
			//convert the file to URL format
			URL url = file.toURI().toURL();
			URL[] urls = new URL[]{url};
			ClassLoader cl = new URLClassLoader(urls);
			result = cl.loadClass(advisingType).getPackage().getName();      
		} catch (ClassNotFoundException e) {
//			e.printStackTrace();
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return result;
	}
	
	public static boolean hasExcludedAdvisingAnnotationInEnclosingType(String advisedType, String urlClassLoader){
		boolean result = false;
		
		try {
			File file = new File(urlClassLoader);
			//convert the file to URL format
			URL url = file.toURI().toURL();
			URL[] urls = new URL[]{url};
			ClassLoader cl = new URLClassLoader(urls);
			for (int i = 0; i < cl.loadClass(advisedType).getAnnotations().length; i++) {
				if(cl.loadClass(advisedType).getAnnotations()[i].toString().
						equals("@org.aspectjml.lang.annotation.ExcludeAdvising()")){
					result = true;
					break;
				}
				else if(cl.loadClass(advisedType).getAnnotations()[i].toString().
						equals("@org.aspectjml.lang.annotation.Sealed()")){
					result = true;
					break;
				}
			}
		} catch (ClassNotFoundException e) {
//			e.printStackTrace();
		} catch (MalformedURLException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return result;
	}
	
	// java.lang.reflect Constructor Util - [[[hemr]]]
	public static String getConstructorSignatureWithLongTypeNames ( Constructor<?> init ) {
		return getSignature(init, true, false);
	}

	public static String getConstructorSignatureWithShortTypeNames ( Constructor<?> init ) {
		return getSignature(init, false, false);
	}
	
	public static String getConstructorSignatureWithLongTypeNames2 ( Constructor<?> init ) {
		return getSignature(init, true, true);
	}

	public static String getConstructorSignatureWithShortTypeNames2 ( Constructor<?> init ) {
		return getSignature(init, false, true);
	}

    // java.lang.reflect Method Util - [[[hemr]]]
	public static String getMethodSignatureWithLongTypeNames ( Method method ) {
		return getSignature(method, true);
	}

	public static String getMethodSignatureWithShortTypeNames ( Method method ) {
		return getSignature(method, false);
	}

	// helper methods - [[[hemr]]]
	private static String getSignature ( Constructor<?> init , boolean longTypeNames, boolean includeAdviceNew) {
		if(includeAdviceNew) {
			return init.getName()+".new"+ "("+ parametersAsString(init, longTypeNames) + ")";
		}
		else {
			return init.getName()+ "("+ parametersAsString(init, longTypeNames) + ")";
		}
		
	}
	
	private static String parametersAsString ( Constructor<?> init, boolean longTypeNames ) {
		String result = "";
		Class<?>[] parameterTypes = init.getParameterTypes();
		if ( parameterTypes.length > 0 ) {
			StringBuilder paramString = new StringBuilder();
			paramString.append(longTypeNames ? parameterTypes[0].getName()
					: parameterTypes[0].getSimpleName());
			for ( int i = 1 ; i < parameterTypes.length ; i++ ) {
				paramString.append(",").append(
						longTypeNames  ? parameterTypes[i].getName()
								: parameterTypes[i].getSimpleName());
			}
			result = paramString.toString();
		}
		return result;
	}	
	
	private static String getSignature ( Method method, boolean longTypeNames ) {
		return method.getName() + "("+ parametersAsString(method, longTypeNames) + ")";
	}

	private static String parametersAsString ( Method method, boolean longTypeNames ) {
		String result = "";
		Class<?>[] parameterTypes = method.getParameterTypes();
		if ( parameterTypes.length > 0 ) {
			StringBuilder paramString = new StringBuilder();
			paramString.append(longTypeNames ? parameterTypes[0].getName()
					: parameterTypes[0].getSimpleName());
			for ( int i = 1 ; i < parameterTypes.length ; i++ ) {
				paramString.append(",").append(
						longTypeNames  ? parameterTypes[i].getName()
								: parameterTypes[i].getSimpleName());
			}
			result = paramString.toString();
		}
		return result;
	}	
}
